/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#include "generaldef.h"

#include <time.h>


#include "plane.h"
#include "hpc.h"
#include "gviewer.h"
#include "gbool.h"
#include "pipeline3d.h"
#include "svg3d.h"

bool HPC_USE_SYMBOLIC_MKPOL=true;
real HPC_TOLERANCE=((real)1e-6);
real HPC_TOLERANCE_BOOL=((real)0.5f*1e-3f);
bool HPC_USE_INTERNAL_FACES=true;
bool HPC_USE_PARTITION_PLANES=false;


/////////////////////////////////////////////////////////////////////////////////////////
void Hpc::dropSharedGeometry()
{
	g=0;

	for (int i=0;i<num();i++)
		get(i)->dropSharedGeometry();
}


/////////////////////////////////////////////////////////////////////////////////////////
static void reducedForm(Hpc* dst,const Hpc* src,const real* vacc,const real* hacc,bool bStrict,bool bShareGeometry,char* _name=0)
{
	// todo: copy properties
	// todo: memory deallocation for names

#if 1
	char* thisname=(char*)((Hpc*)src)->getProperty("name");
	char name[2048]; name[0]=0;
	if (_name) strcpy(name,_name);

	if (thisname)
	{
		if (name[0]) strcat(name,".");
		strcat(name,thisname);
	}
#endif

	const int spacedim=dst->spacedim;
	size_t matsize=(spacedim+1)*(spacedim+1);

	//actual matrix
	real* V=(src->vmat || vacc)?new real[matsize]:0;
	real* H=(src->hmat || hacc)?new real[matsize]:0;

	if (src->vmat)
	{
		tmat<real>::copy(spacedim+1,V,src->spacedim+1,src->vmat);
		tmat<real>::copy(spacedim+1,H,src->spacedim+1,src->hmat);

		if (vacc)
		{
			tmat<real>::product(spacedim+1,V,vacc,V);
			tmat<real>::product(spacedim+1,H,H,hacc);
		}
	}
	else if (vacc)
	{
		memcpy(V,vacc,sizeof(real)*matsize);
		memcpy(H,hacc,sizeof(real)*matsize);
	}

	if (src->g)
	{
		Hpc* sub=0;

		if (bStrict)
		{
			ggraph* g=src->g->clone();
			g->embed(spacedim);
			if (V) g->transform(V,H);
			sub=new Hpc(g);
		}
		else if (bShareGeometry)
		{
			ggraph* g=src->g; //share the same geometry
			sub=new Hpc(spacedim,V,H);
			sub->g=g;
			sub->pointdim=g->pointdim;
		}
		else if (V && src->g->pointdim==spacedim) 
		{
			ggraph* g=src->g->clone();
			g->transform(V,H);
			sub=new Hpc(g);
		}
		else
		{
			ggraph* g=src->g->clone();
			sub=new Hpc(spacedim,V,H);
			sub->g=g;
			sub->pointdim=g->pointdim;
		}

#if 1
		{
			if (name) 
			{
				char* tmp=new char[strlen(name)+1];
				strcpy(tmp,name);
				sub->addProperty("name",tmp);
			}
		}
#endif

		dst->add(sub);
		sub->release();
	}

	for (int i=0;i<src->num();i++)
		reducedForm(dst,src->get(i),V,H,bStrict,bShareGeometry,name);

	//do not use anymore
	if (V) delete [] V;
	if (H) delete [] H;
}


/////////////////////////////////////////////////////////////////////////////////////////

Hpc* Hpc::reducedForm(bool bStrict,bool bShareGeometry) const
{
	Hpc* ret=new Hpc(spacedim);
	::reducedForm(ret,this,0,0,bStrict,bShareGeometry);
	return ret;
}


/////////////////////////////////////////////////////////////////////////////////////////
void Hpc::limits(real* m,real* M) 
{
	bool bStartup,bApplySpacedim;
	real limit,maxdelta;
	Hpc* tmp=reducedForm(false,true); //!bStrict,bShareGeometry

	//not thread safe
	static key points[MAX_CELLS_PER_LEVEL];

	for (int n=0;n<=spacedim;n++) 
	{
		m[n]=+1e18f;
		M[n]=-1e18f;
	}
	
	for (int i=0;i<tmp->num();i++)
	{
		Hpc* sub=tmp->get(i);
		ggraph* g=sub->g;
		checkpoint(!g->vmat && !g->hmat && g->pointdim==g->spacedim);
		g->setSpacedim(sub->spacedim,sub->vmat,sub->hmat);
		key npoints=g->findCells(0,0,points); 
		g->getBoundingBox(npoints,points,m,M,limit,maxdelta,bStartup=false,bApplySpacedim=true);
		g->setSpacedim(g->pointdim,0,0);
	}

	tmp->dropSharedGeometry();
	tmp->release();
}

/////////////////////////////////////////////////////////////////////////////////////////

Hpc* Hpc::power(Hpc* _arg1,Hpc* _arg2)
{	
	Hpc* arg1=_arg1->reducedForm(false,true); //!bStrict,bShareGeometry
	Hpc* arg2=_arg2->reducedForm(false,true); //!bStrict,bShareGeometry 

	Hpc* ret=new Hpc;
	
	for (int I=0;I<arg1->num();I++)
	for (int J=0;J<arg2->num();J++)
	{
		Hpc *sub1=arg1->get(I);ggraph* g1=sub1->g;
		Hpc *sub2=arg2->get(J);ggraph* g2=sub2->g;

		ggraph* g=ggraph::power(
			g1,sub1->spacedim,sub1->vmat,sub1->hmat,
			g2,sub2->spacedim,sub2->vmat,sub2->hmat,
			HPC_TOLERANCE,HPC_USE_INTERNAL_FACES,HPC_USE_PARTITION_PLANES);

		Hpc* sub=new Hpc(g);
		ret->add(sub);
		sub->release();
	}

#ifdef _DEBUG
	ret->check();
#endif

	arg1->dropSharedGeometry();arg1->release(); 
	arg2->dropSharedGeometry();arg2->release(); 

	return ret;
}


/////////////////////////////////////////////////////////////////////////////////////////
/* 
from n-arguments take only the set which correspond to max-dimension
scale/translate to force all object be in the range -1,+1
returns the matrices to remap object to real position
*/
/////////////////////////////////////////////////////////////////////////////////////////

bsp_solver* Hpc::getSolver(int op,int n,Hpc** _v,real* vmat,real* hmat)
{
	int i,spacedim=0;

	real mL[(MAX_GEOMETRIC_DIMENSION+1)],ML[MAX_GEOMETRIC_DIMENSION+1];
	real mT[(MAX_GEOMETRIC_DIMENSION+1)],MT[MAX_GEOMETRIC_DIMENSION+1];

	for (i=0;i<n;i++) 
		spacedim=max2(spacedim,_v[i]->spacedim);

	//more robust scaling globally to unit box
	int ngood=0;
	for (i=0;i<n;i++) 
	{
		//only regularized operations!
		if (_v[i]->pointdim!=spacedim) continue;
	 
		if (!ngood)
		{
			_v[i]->limits(mT,MT);
		}
		else
		{
			_v[i]->limits(mL,ML);

			for (int j=0;j<=spacedim;j++)
			{
				mT[j]=min2(mT[j],mL[j]);
				MT[j]=max2(MT[j],ML[j]);
			}
		}

		++ngood;
	}

	if (!ngood) 
		return 0;

	tmat<real>::identity(spacedim+1,vmat);
	tmat<real>::identity(spacedim+1,hmat);

	for (i=1;i<=spacedim;i++)
	{
		//not full dimension
		if (!(MT[i]-mT[i])) 
			return 0;

		real t=(real)(-0.5*(mT[i]+MT[i]));
		real s=(real)2.0f/(MT[i]-mT[i]);

		vmat[i*(spacedim+1)]=+t*s;vmat[i*(spacedim+1)+i]=1*s;
		hmat[i*(spacedim+1)]=-t  ;hmat[i*(spacedim+1)+i]=1/s;
	}

	checkpoint(tmat<real>::isInvert(spacedim+1,vmat,hmat));

	Hpc** v=new Hpc*[n];

	for (i=0;i<n;i++) 
	{
		v[i]=0;

		if (_v[i]->pointdim!=spacedim) 
			continue;

		Hpc* tmp=Hpc::transform(spacedim,vmat,hmat)->add(_v[i]);
		v[i]=tmp->reducedForm(false,false); //!bStrict,!bShareGeometry (I need the clone of geometry)
		tmp->release();
	}

	//I'm in the range -1,+1. seems too picky to set HPC_TOLERANCE, so I've used HPC_TOLERANCE_BOOL
	GBool b(spacedim, HPC_TOLERANCE_BOOL,HPC_USE_INTERNAL_FACES,HPC_USE_PARTITION_PLANES);

	key Up4=b.BoolOp(op);

	for (i=0;i<n;i++) 
	{
		if (v[i]->pointdim!=spacedim) 
			continue;
		
		key Or3=b.BoolOp(bsp_bool::BOOL_OR,3);

		for (int u=0;u<v[i]->num();u++)
		{
			Hpc* sub=v[i]->get(u);
			ggraph* g=sub->g;

			if (g->pointdim==spacedim) 
			{
				checkpoint(!sub->vmat && !sub->hmat);
				b.addArg(b.add(g),Or3);
			}
		}
		b.addArg(Or3,Up4);
	}

	b.setRoot(Up4);
	b.visit();

	for (i=0;i<n;i++)
	{
		if (v[i]->pointdim!=spacedim)  continue;
		v[i]->release();
	}

	delete [] v;
	
	float limit=1.1f; //unit box plus delta
	bsp_solver* s=new bsp_solver(&b,HPC_TOLERANCE,0,limit,false);
	s->visit();
	return s;
}


/////////////////////////////////////////////////////////////////////////////////////////
Hpc* Hpc::booleanOp(int op,int n,Hpc** _v)
{
	//not thread safe
	static real vmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
	static real hmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];

	bsp_solver* solver=getSolver(op,n,_v,vmat,hmat);

	if (!solver)
		return 0;

	//solver->print();

	int spacedim=solver->getDim();
	ggraph* g=solver->getGraph()->clone();
	Hpc* tmp=new Hpc(g);
	Hpc* ret=Hpc::transform(spacedim,hmat,vmat)->add(tmp);
	tmp->release();
	delete solver;
	return ret;
}



/////////////////////////////////////////////////////////////////////////////////////////
int Hpc::exportSvg(char* outfilename,char* type,int width,int height,
						 bool back_to_front,
						 double vrp[3],double vuv[3],double vpn[3],double cop[3],
						 double front,double back,
						 double range_u[2],double range_v[2],
						 double range_x[2],double range_y[2])
{
	//not thread safe
	static real vmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
	static real hmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];

	Hpc* _v[]={this};
	bsp_solver* solver=getSolver(bsp_bool::BOOL_OR,1,_v,vmat,hmat);

	if (!solver)
		return 0;

	int spacedim=solver->getDim();

	//valid only in 3 dimension
	if (spacedim!=3)
	{
		delete solver;
		return 0;
	}

	real Tv[16]={1,0,0,0,  0,1,0,0,  0,0,1,0,  0,0,0,1};
	real observer[4]={1,0,0,0};
	makePipeline3d(Tv,&observer[1],type,vrp,vuv,vpn,cop,front,back,range_u,range_v,range_x,range_y);

	//homogeneous coordinates are different
	int permutation[MAX_GEOMETRIC_DIMENSION+1];
	int matdim=spacedim+1;
	permutation[0]=matdim-1;
	for (int I=1;I<matdim;I++) permutation[I]=I-1;
	tmat<real>::swap_rows(matdim,Tv,permutation);
	tmat<real>::swap_cols(matdim,Tv,permutation);
	tmat<real>::product(matdim,Tv,Tv,hmat); //must apply the transformation matrix of getSolver

	svg3d(solver,outfilename,width,height, back_to_front,observer,Tv);

	delete solver;
	return 1;
}




/////////////////////////////////////////////////////////////////////////////////////////

Hpc* Hpc::skeleton(int level) 
{
	Hpc* tmp=reducedForm(false,true); //!bStrict,bShareGeometry
	Hpc* ret=new Hpc;

	for (int i=0;i<tmp->num();i++)
	{
		Hpc     *pc=tmp->get(i);
		ggraph  *g=pc->g;

		//copy also the eventually transformation
		Hpc* sub=new Hpc(pc->spacedim,pc->vmat,pc->hmat); 

		if (g->pointdim==level) 
		{
			sub->g=g->clone();
			sub->pointdim=g->pointdim;
		}
		else
		{
			int numg;
			ggraph** v=g->skeleton(level,numg);
			for (int j=0;j<numg;j++) 
			{
				Hpc* tmp=new Hpc(v[j]);
				sub->add(tmp);
				tmp->release();
			}	
			delete [] v;
		}

		ret->add(sub);
		sub->release();
	}
	
	tmp->dropSharedGeometry();
	tmp->release(); 
	return ret;
}


/////////////////////////////////////////////////////////////////////////////////////////

Hpc* Hpc::join(int n,Hpc** _pcs) 
{
	real tmp[MAX_GEOMETRIC_DIMENSION+1];
	int spacedim=0,i;

	Hpc** pcs=new Hpc*[n];

	for (i=0;i<n;i++) 
	{
		pcs[i]=_pcs[i]->reducedForm(false,true); //!bStrict,bShareGeometry
		spacedim=max2(spacedim,pcs[i]->spacedim);
	}

	//not thread safe
	static key points[MAX_CELLS_PER_LEVEL];
	ggraph gpoints(spacedim);	
	key npoints=0;

	for (i=0;i<n;i++) 
	{
		for (int j=0;j<pcs[i]->num();j++)
		{
			Hpc* sub=pcs[i]->get(j);
			ggraph* g=sub->g;

			for (ggraph::list_iterator it=g->each(0);!it.end();it++)
			{
				real* vg=g->getGeom(*it);
				memset(tmp,0,sizeof(tmp));
				memcpy(tmp,vg,sizeof(real)*(g->pointdim+1));
				if (sub->vmat) productv<real,real>(sub->spacedim+1,tmp,sub->vmat,tmp);
				points[npoints++]=gpoints.addVertex(tmp);
			}
		}
	}

	for (i=0;i<n;i++) 
	{
		pcs[i]->dropSharedGeometry();
		pcs[i]->release(); 
	}

	delete [] pcs;

	ggraph* g=gpoints.mkpol(npoints,points,HPC_USE_SYMBOLIC_MKPOL,HPC_TOLERANCE);

	Hpc* ret=new Hpc(g);
	return ret;
}



Hpc* Hpc::mkpol(int pointdim,int npoints,const real* points,int nhulls,const int** hulls)
{
	//base case
	if (!npoints) 
		return 0;

	real Tmp[MAX_GEOMETRIC_DIMENSION+1];Tmp[0]=1;
	std::map<int,key> vmap;
	Hpc* ret=new Hpc(pointdim);
	ggraph gpoints(pointdim);
	const real* p=points;

	for (int i=0;i<npoints;i++,p+=pointdim)
	{
		memcpy(&Tmp[1],p,sizeof(real)*pointdim);
		vmap[i]=gpoints.addVertex(Tmp);
	}

	//not thread safe
	static key Id[MAX_CELLS_PER_LEVEL];
	
	for (int H=0;H<nhulls;H++)
	{
		const int* v=hulls[H];
		int nv=*v++;

		if (nv)
		{
			for (int i=0;i<nv;i++) Id[i]=vmap[*v++];
			ggraph* g=gpoints.mkpol(nv,Id,HPC_USE_SYMBOLIC_MKPOL,HPC_TOLERANCE);
			Hpc* sub=new Hpc(g);
			ret->add(sub);
			sub->release();
		}
	}

	return ret;
}


/////////////////////////////////////////////////////////////////////////////////////////

void Hpc::ukpol(int& dim,int& nnode,real*& node,int& nhulls,int**& hulls,bool bFace)
{
	ggraph::list_iterator it;
	int i,Lvl,NA=0,NH=0;

	nnode=0;nhulls=0;
	node =0;hulls =0;
	dim=this->spacedim;

	//for bFace dim>=2! xge does not store faces 
	if (bFace && this->spacedim<=1)
		{dim=0;return;}

	Hpc* tmp=reducedForm(true,false); //BStrict,!bShareGeometry

	for (i=0;i<tmp->num();i++) 
	{
		ggraph* g=tmp->get(i)->g;
		if (bFace && g->each(dim).end()) continue;//not full dimension
		if (bFace && dim>1) g->fixPlaneOrientation(); //make the normals point towards "outside"
		nnode+=g->getNCells(bFace?(dim-1):0); //plane or points
		for(Lvl=dim;!bFace && Lvl && g->each(Lvl).end(); Lvl--) ;
		nhulls+=g->getNCells(Lvl);
	}

	if (!nnode || !nhulls)
	{
		tmp->release();
		dim=0;
		return;
	}

	//not thread safe
	static key zerocells[MAX_CELLS_PER_LEVEL];
	key nzerocells;

	node  =new real[nnode*(bFace?(dim+1):(dim))];
	hulls =new int*[nhulls];

	real*  ptrnode  =node;
	int**  ptrhulls =hulls;
	
	for (i=0;i<tmp->num();i++)
	{
		std::map<key,int> mymap;
		ggraph* g=tmp->get(i)->g;
		if (bFace && g->each(dim).end()) continue;

		for (it=g->each(bFace?(dim-1):0);!it.end();it++)
		{
			real* Coords=g->getGeom(*it);
			for (int j=bFace?0:1;j<=dim;j++) *ptrnode++=Coords[j];
			mymap[*it]=NA++;
		}

		for (Lvl=dim;!bFace && Lvl && g->each(Lvl).end(); Lvl--) ;

		for (it=g->each(Lvl);!it.end();it++,NH++)
		{
			if (bFace)
			{
				ptrhulls[NH]=new int[g->getNDw(*it)+1];
				int* ptr=ptrhulls[NH];*ptr++=g->getNDw(*it);
				for (ggraph::iterator jt=g->goDw(*it);!jt.end();jt++) *ptr++=mymap[*jt];
			}
			else
			{
				key C=*it;
				nzerocells=g->findCells(0,*it,zerocells);
				ptrhulls[NH]=new int[nzerocells+1];
				int* ptr=ptrhulls[NH];*ptr++=nzerocells;
				for (int k=0;k<nzerocells;k++) *ptr++=(int)mymap[zerocells[k]];
			}
		}
	}

	tmp->release();
	assert(NA==nnode );
	assert(NH==nhulls);
}


/////////////////////////////////////////////////////////////////////////////////////////
void Hpc::print(int nrec,FILE* file) const
{
	char spaces[1024];
	char fspaces[16];
	sprintf(fspaces,"%%%ds",nrec*3);
	sprintf(spaces,fspaces," ");

	fprintf(file,"%sspace [0x%x] dim(%d) pointdim(%d) ",spaces,(int)this,(int)spacedim,(int)this->pointdim);

	for (std::map< std::string,void*>::const_iterator it=prop.begin();it!=prop.end();it++)
		fprintf(file,"%s%s(%s)",spaces,it->first.c_str(),"...");

	fprintf(file,"\n");

	if (vmat)
	{
		int i,j;

		fprintf(file,"%s((",spaces);
		for (i=0;i<spacedim+1;++i)
		{
			if (i) fprintf(file,",");
			fprintf(file,"(");
			for (j=0;j<spacedim+1;++j) fprintf(file,"%s%.1f ",j?",":"",(dec64)vmat[i*(spacedim+1)+j]);
			fprintf(file,")");
		}
		fprintf(file,"))\n");

		fprintf(file,"%shmat((",spaces);
		for (i=0;i<spacedim+1;++i)
		{
			if (i) fprintf(file,",");
			fprintf(file,"(");
			for (j=0;j<spacedim+1;++j) fprintf(file,"%s%.1f ",j?",":"",(dec64)hmat[i*(spacedim+1)+j]);
			fprintf(file,")");
		}
		fprintf(file,"))\n");
	}

	if (g)
		g->print(spaces,file);

	for (int i=0;i<pols.size();i++)
		pols[i]->print(nrec+1,file);		

}


/////////////////////////////////////////////////////////////////////////////////////////

void Hpc::check()
{
	checkpoint((!vmat && !hmat) || (vmat && hmat));

	if (vmat)
	{
		real tmp[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
		tmat<real>::product(spacedim+1,tmp,vmat,hmat);

		for (int i=0;i<=spacedim;i++)
		{
			for (int j=0;j<=spacedim;j++)
			{
				real val=tmp[i*(spacedim+1)+j];
				checkpoint((i!=j && fabs(val)<1e-3) || (i==j && fabs(1-val)<1e-3));
			}
		}
	}
	if (g)  g->check();
	for (int i=0;i<num();i++) get(i)->check();
}


/////////////////////////////////////////////////////////////////////////////////////////

class HpcDraw: public Draw
{
	const Hpc* src;

	void display(const Hpc* src);

public:
	
	//constructor
	HpcDraw(const Hpc* src)
	{
		this->src=src;
	}

	//setbbox
	void setbbox()
	{
		real m[MAX_GEOMETRIC_DIMENSION+1],M[MAX_GEOMETRIC_DIMENSION+1];
		((Hpc*)src)->limits(m,M);

		float maxsize=max3(M[1]-m[1],M[2]-m[2],M[3]-m[3]);

		if (maxsize>0 && maxsize<((real)1e6))
		{
			float b[6];
			
			int i;
			for (i=0;i<3;i++)
			{
				float size=M[i+1]-m[i+1];
				b[i*2+0]=(size>0 && size<((real)1e6))?m[i+1]:0;
				b[i*2+1]=(size>0 && size<((real)1e6))?M[i+1]:1;
			}

			for (i=0;i<3;i++)
			{
				float size=b[i*2+1]-b[i*2+0];
				b[i*2+0]-=size*0.5;
				b[i*2+1]+=size*0.5;
			}

			getPanel()->setbbox(b);

			if (src->spacedim<3 || !(M[3]-m[3]))
			{
				getPanel()->resetMatrix();
				getPanel()->setNewRotation(0);
			}
		}
	}

	//display
	virtual void display()
		{display(src);}

};


/////////////////////////////////////////////////////////////////////////////////////////

class HpcViewer:public XgeViewer
{
	const Hpc* src;

public:

	inline HpcViewer(const Hpc* src)
	{
		this->src=src;

		HpcDraw draw(src);
		getPanel()->add(&draw);
		draw.setbbox();
		run();
	}

	virtual void keydown(unsigned char key, GLint x, GLint y)
	{
		//export in povray
		if (key=='y' || key=='Y')
		{
			src->povray();
			return;
		}

		XgeViewer::keydown(key,x,y);
	}
};

/////////////////////////////////////////////////////////////////////////////////////////
void Hpc::view() const
{
	HpcViewer win(this);
}


/////////////////////////////////////////////////////////////////////////////////////////
void HpcDraw::display(const Hpc* src) 
{
	XgeViewer* win=(XgeViewer*)this->getPanel()->getWindow();
	int spacedim=src->spacedim;
	real* vmat=src->vmat;
	ggraph* g=src->g;
	
	real   expand=win->getExpand()*0.1;
	bool   bball=win->getBBall();

	 glMatrixMode(GL_MODELVIEW);

	if (spacedim>3)
		return;

	if (vmat)
	{
		glPushMatrix();

		switch(spacedim)
		{
		case 0:{
				float aux[16]=
				{
					      1,0,0,     0,  
						   0,1,0,     0,  
							0,0,1,     0,  
							0,0,0,vmat[0]
				};
				glMultMatrixf(aux);break;
			}
		case 1:{
				float aux[16]=
				{
					vmat[3],0,0,vmat[1], 
					      0,1,0,      0, 
					      0,0,1, 		0, 
					vmat[2],0,0,vmat[0], 

				};
				glMultMatrixf(aux);break;
			}
		case 2:
			{
				float aux[16]=
				{
					vmat[4],vmat[7],0,vmat[1],
					vmat[5],vmat[8],0,vmat[2],
					      0,      0,1,      0,
					vmat[3],vmat[6],0,vmat[0],

				};
				glMultMatrixf(aux);break;
			}
		case 3:
			{

				float aux[16]={
					vmat[ 5],vmat[ 9],vmat[13],vmat[ 1],
					vmat[ 6],vmat[10],vmat[14],vmat[ 2],
					vmat[ 7],vmat[11],vmat[15],vmat[ 3],
					vmat[ 4],vmat[ 8],vmat[12],vmat[ 0]
				};

				glMultMatrixf(aux);break;
			}
		}
	}

	if (g)
	{
		int viewcelllevel=g->pointdim;
		while (viewcelllevel && g->each(viewcelllevel).end()) --viewcelllevel;
		const float colors[9]={1.0f,1.0f,0.0f, 0.1f,0.1f,0.1f, 0.5f,0.5f,0.5f};
		const float* C=colors+(viewcelllevel % 3)*3;
		XgeDraw::display(g,viewcelllevel,C[0],C[1],C[2],expand,expand,expand,false,8,3,bball);
	}

	for (int i=0;i<src->num();i++) display(src->get(i));
	if (vmat) glPopMatrix();
}


//-------------------------------------------------------------------------------------------
bool Hpc::svg(double width,const char* filename) const
{
	if (this->spacedim!=2)
		return false;

	FILE* file=fopen(filename,"wt");
	if (!file)
		return false;

	Hpc* tmp=reducedForm(true,false);//bStrict,!bShareGeometry

	real m[2],M[2];
	tmp->limits(m,M);
	double factor=width/(M[1]-m[1]);
	tmp=Hpc::scale(2,2,factor)->add(Hpc::scale(2,1,factor)->add(tmp));

	fprintf(file,"<?xml version='h1.0' encoding='iso-8859-1'?>\n");
	fprintf(file,"<svg>\n");
	fprintf(file,"<g style='fill:none;stroke:black'>\n");

	for (int i=0;i<tmp->num();i++)
	{
		Hpc* sub=tmp->get(i);
		ggraph* g=sub->g;

		if (!sub->vmat)
		{
			for (ggraph::list_iterator it=g->each(2);!it.end();it++)
			{
				key F=*it;
				key prev0=g->getFirstDwNode(g->getLastDwNode(F)),p1;
				key prev1=g->getLastDwNode (g->getLastDwNode(F)),p2;
				
				fprintf(file,"<path d=' M ");

				for (graph::iterator i=g->goDw(F);!i.end();i++,prev0=p1,prev1=p2)
				{
					p1=g->getFirstDwNode(*i);
					p2=g->getLastDwNode (*i);
					if (p1!=prev0 && p1!=prev1) Swap(key,p1,p2);
					real* g1=g->getGeom(p1);
					fprintf(file,"%f %f %s ",g1[1],g1[2],i.next().end()?"z":"L");
				}
				
				fprintf(file,"'/>\n");
			}
		}
	}

	fprintf(file,"</g>\n");
	fprintf(file,"</svg>\n");
	fclose(file);

	tmp->release();

	return true;
}

//-------------------------------------------------------------------------------------------
bool Hpc::vrml(const char* filename) const
{
	if (this->spacedim>3)
		return false;

	FILE* file=fopen(filename,"wt");

	if (!file)
		return false;

	fprintf(file,"#VRML V2.0 utf8                   \n");
	fprintf(file,"Group {                           \n");
	fprintf(file," children [                       \n");
	fprintf(file," Background  {                    \n");
	fprintf(file,"  skyColor    .3 .7 .9            \n");
	fprintf(file,"  groundColor .3 .7 .9            \n");
	fprintf(file," }                                \n");
	fprintf(file," SpotLight {                      \n");
	fprintf(file,"  location     0 1 0              \n");
	fprintf(file,"  direction    0 -1 0             \n");
	fprintf(file,"  intensity    1                  \n");
	fprintf(file,"  cutOffAngle  1.9                \n");
	fprintf(file," }                                \n");
	fprintf(file," DirectionalLight {               \n");
	fprintf(file,"  direction    0 0 -1             \n");
	fprintf(file,"  intensity    1.0                \n");
	fprintf(file," }                                \n");

	Hpc* aux=this->clone();
	aux=aux->embed(3);
	Hpc* tmp=aux->reducedForm(true,false); //bStrict,!bShareGeometry
	aux->release();
	checkpoint(!tmp->vmat);
 
	for (int i=0;i<tmp->num();i++)
	{
		Hpc* sub=tmp->get(i);
		ggraph* g=sub->g;
		checkpoint(!sub->vmat && g->pointdim==3);
		
		fprintf(file,"Shape {");

		fprintf(file,"\tappearance Appearance {\n");
		fprintf(file,"\tmaterial Material {\n");
		fprintf(file,"\tdiffuseColor 0.8 0.8 0.8\n");
		fprintf(file,"\tambientIntensity 0.2\n");
		fprintf(file,"\tspecularColor 0 0 0\n");
		fprintf(file,"\temissiveColor 0 0 0\n");
		fprintf(file,"\tshininess 0.2\n");
		fprintf(file,"\ttransparency 0\n");
		fprintf(file,"}}\n");

		fprintf(file,"geometry ");

		int Export;
		     if (!g->each(2).end()) {Export=2;fprintf(file,"IndexedFaceSet");}
      else if (!g->each(1).end()) {Export=1;fprintf(file,"IndexedLineSet");}
		else                        {Export=0;fprintf(file,"PointSet"      );}

		fprintf(file,"{ coord Coordinate { \npoint[\n");

		int nv=0;
		ggraph::list_iterator it;
		for (it=g->each(0);!it.end();it++,nv++)
		{
			key V=*it;g->nodeTmp(V)=nv;
			real* vg=g->getGeom(V);
			fprintf(file,"\t%s%e,%e,%e\n",(nv?",":" "),vg[1],vg[2],vg[3]);
		}

		fprintf(file,"]}\n");

		if (Export==1)
		{
			fprintf(file,"coordIndex[\n");
			int nl=0;
			for(it=g->each(1);!it.end();it++,nl++)
			{
				key L=*it,V0=g->getFirstDwNode(L),V1=g->getLastDwNode(L);
				fprintf(file,"%s%d,%d,-1\n",(nl?",":" "),(int)g->nodeTmp(V0),(int)g->nodeTmp(V1));
			}
			fprintf(file,"]\n");
		}
		else if (Export==2)
		{
			fprintf(file,"coordIndex[\n");
			int nf=0,nv=0;

			for(it=g->each(2);!it.end();it++,nf++)
			{
				int acc[2048],nacc=0;
				key F=*it,E=g->getLastDwNode(F),prev0=g->getFirstDwNode(E),prev1=g->getLastDwNode(E),p1,p2;
				for (graph::iterator jt=g->goDw(F);!jt.end();jt++,prev0=p1,prev1=p2,nv++)
				{
					p1=g->getFirstDwNode(*jt),p2=g->getLastDwNode (*jt);
					if (p1!=prev0 && p1!=prev1) Swap(key,p1,p2);
					acc[nacc++]=g->nodeTmp(p1);	
				}
				checkpoint(nacc<=2048);

				//direct and reverse to avoid face culling
				int u;
				for (u=0;u<nacc;u++) fprintf(file,"\t%s%d",(nf || u?",":" "),acc[u       ]);
				fprintf(file,",-1\n"); 
				for (u=0;u<nacc;u++) fprintf(file,"\t,%d",acc[nacc-u-1]);
				fprintf(file,",-1\n"); 
			}
			fprintf(file,"]\n");
		}
		fprintf(file,"}}\n");
	}

	fprintf(file," ]\n");
	fprintf(file,"}\n");
	fprintf(file,"#VRML 2.0 File End\n");
	fclose(file);

	tmp->release();
	return true;
}

//-------------------------------------------------------------------------------------------

bool Hpc::povray(const char* filename) const
{
	FILE* file=fopen(filename,"wt");

	if (!file)
		return false;

	Hpc* tmp=reducedForm(false,true); //!bStrict,bShareGeometry

	float target[3]={0,0,0};
	float eye[3]={1000,1000,1000};

	char* header=
		"#include \"colors.inc\" \n"
		"#include \"shapes.inc\" \n"
		"#include \"textures.inc\" \n"
		"#include \"metals.inc\" \n"
		"#include \"skies.inc\" \n"
		"global_settings {assumed_gamma 1 max_trace_level 5 radiosity  {pretrace_start 0.08 pretrace_end   0.02 count 50 nearest_count 5 error_bound 0.15 recursion_limit 1 low_error_factor .5 gray_threshold 0.0 minimum_reuse 0.015 brightness 1 adc_bailout 0.01/2}} \n"
		"camera {location <%e,%e,%e> right(4/(3*(1.000000)))*y look_at<%e,%e,%e> angle 36 } \n"
		"light_source {<120,30,220> color rgb <1.0, 0.9, 0.8>*2.4 } \n"
		"sky_sphere {pigment{color red 0.6 green 0.6 blue 1.0 }} \n"
		" \n\n\n";

	fprintf(file,header,eye[0],eye[1],eye[2],target[0],target[1],target[2]);

	std::map<std::string,FILE*> filemap;
	std::map<ggraph*,std::string> graphmap;
	fprintf(file,"#include \"geometry.inc\"\n");

	FILE* graphfile=fopen("geometry.inc","wt");

	for (int i=0;i<tmp->num();i++)
	{
		Hpc* sub=tmp->get(i);
		ggraph* g=sub->g;

		// check good dimension
		if (sub->spacedim!=2 && sub->spacedim!=3)
			continue;

		if (g->pointdim!=2 && g->pointdim!=3)
			continue;
		
		if (g->each(2).end()) 
			continue;

		// name of the current structure
		char name[1024];

		if ((char*)sub->getProperty("name"))
			sprintf(name,"%s",(char*)sub->getProperty("name"));
		else
			sprintf(name,"unknown");

		for (int J=0;J<strlen(name);J++)
			if (name[J]=='.') name[J]='_';
		
		// share the geometry
		if (graphmap.find(g)==graphmap.end())
		{
			char graphname[1024];
			sprintf(graphname,"graph_%04d",i);
			graphmap[g]=std::string(graphname);

			fprintf(graphfile,"#declare %s=object{mesh{\n",graphname);

			for(ggraph::list_iterator it=g->each(2);!it.end();it++)
			{
				key F=*it,E=g->getLastDwNode(F);

				key prev0=g->getFirstDwNode(E),p1;
				key prev1=g->getLastDwNode (E),p2;

				struct{float x,y,z;}
				v0,v1,v2,v;

				int nv=0;
				for (graph::iterator jt=g->goDw(F);!jt.end();jt++,prev0=p1,prev1=p2,nv++)
				{
					p1=g->getFirstDwNode(*jt);
					p2=g->getLastDwNode (*jt);
					if (p1!=prev0 && p1!=prev1) Swap(key,p1,p2);
					real* g1=g->getGeom(p1);
					v.x=g1[1];v.y=g1[2];v.z=(g->pointdim==3)?(g1[3]):(0);

					if      (!nv)   v0=v;
					else if (nv==1) v1=v;
					else            v2=v;

					if (nv>=2)
					{
						fprintf(graphfile,"triangle{<%e,%e,%e>,<%e,%e,%e>,<%e,%e,%e>}\n",v0.x,v0.y,v0.z,v1.x,v1.y,v1.z,v2.x,v2.y,v2.z);
						v1=v2;
					}
				}
			}

			fprintf(graphfile,"}}\n\n");
		}

		// all same structure in files
		if (filemap.find(name)==filemap.end())
		{
			char subfilename[1024];
			sprintf(subfilename,"%s.inc",name,i);
			filemap[name]=fopen(subfilename,"wt");

			fprintf(file,"#declare texture_%s = texture {pigment {color rgb  <0.6,0.6,0.6>} finish {ambient 0.3 diffuse 0.4 specular 0.35}}  ; \n",name);
			fprintf(file,"#include \"%s.inc\"\n",name);
		}

		FILE* subfile=filemap[name];
		fprintf(subfile,"\n\n//**** %s **** \n",name);
		fprintf(subfile,"object{ %s \n",graphmap[g].c_str());

		if (sub->vmat)
		{
			if (sub->spacedim==2)
			{
				fprintf(subfile," matrix <%e,%e,%e,  %e,%e,%e,  %e,%e,%e,  %e,%e,%e> ",
					sub->vmat[ 4],sub->vmat[ 7],0.0f,
					sub->vmat[ 5],sub->vmat[ 8],0.0f,
					         0.0f,         0.0f,1.0f,
					sub->vmat[ 3],sub->vmat[ 6],0.0f);	
			}
			else
			{
				fprintf(subfile," matrix <%e,%e,%e,  %e,%e,%e,  %e,%e,%e,  %e,%e,%e> ",
					sub->vmat[ 5],sub->vmat[ 9],sub->vmat[13],
					sub->vmat[ 6],sub->vmat[10],sub->vmat[14],
					sub->vmat[ 7],sub->vmat[11],sub->vmat[15],
					sub->vmat[ 4],sub->vmat[ 8],sub->vmat[12]);
			}
		}

		fprintf(subfile," texture{texture_%s} }\n",name);

	}

	for (std::map<std::string,FILE*>::iterator it=filemap.begin();it!=filemap.end();it++)
		fclose(it->second);

	fclose(graphfile);

	tmp->dropSharedGeometry();
	tmp->release();

	//	fprintf(file,"plane {<0,1,0>, %e pigment {color rgb<0.2,0.2,0.2>} normal {bumps 0.1 sine_wave scale 100} finish {reflection 0.2}} \n",-maxy-1);

	fflush(file);
	fclose(file);
	tmp->release();
	return true;
}
