/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _hpc_H__
#define _hpc_H__

#include "generaldef.h"

#include <string>
#include <vector>
#include <map>

#include <assert.h>
#include <string.h>

#include "ggraph.h"
#include "memcache.h"
#include "mat.h"
#include "bsp.h"
#include "viewer.h"


// CONFIGURATION 


extern bool HPC_USE_SYMBOLIC_MKPOL;
extern real HPC_TOLERANCE;
extern real HPC_TOLERANCE_BOOL;
extern bool HPC_USE_INTERNAL_FACES;
extern bool HPC_USE_PARTITION_PLANES;




////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

class Hpc
{
protected:

	//destructor (call release to deallocate memory automatically)
	inline ~Hpc()
	{
		if (this->vmat) delete [] this->vmat;
		if (this->hmat) delete [] this->hmat;
		if (this->g   ) delete    this->g   ;
	}

	//check
	void check();

	//increment/decrement an hpc (decremet could invoke deallocation)
	void incrRefCount();
	void decrRefCount();

	//modify
	static Hpc* modify(Hpc* src);

public:

	// *************** DATA ********************

	int refcount;

	//properties
	std::map<std::string, void*> prop;

	//geometry!
	ggraph *g;

	//sub-pols
	std::vector<Hpc*> pols;

	uint8		spacedim;
	uint8    pointdim;
	real		*vmat,*hmat;

	// *****************************************


	//constructor
	Hpc(int spacedim=0,const real* vmat=0,const real* hmat=0);

	//constructor from geometry (spacedim==pointdim=g->pointdim)
	Hpc(ggraph* g);

	//destructor
	void release();

	//clone
	Hpc* clone() const;

	//add a property
	Hpc* addProperty(const char* name,void* prop);

	//access the property
	void* getProperty(const char* pname);

	//add a sub-pol
	Hpc* add(Hpc* src);

	//access the sub-pol
	int        num()     const;
	Hpc*       get(int i);
	const Hpc* get(int i) const;

	//create a dim-cube
	static Hpc* cube(int dim);

	//create a dim-simplex
	static Hpc* simplex(int dim);

	//embed
	Hpc* embed(int NewSpaceDim);

	//print
	void print(int nrec=0,FILE* file=stdout) const;

	//opengl view
	void view() const;

	//reduce
	Hpc* reducedForm(bool bStrict,bool bShareGeometry=false) const;

	//to call after a reducedForm(false,true)
	void dropSharedGeometry();

	//geometric transformations
	static Hpc* scale     (int spacedim,int ref,real factor);
	static Hpc* translate (int spacedim,int ref,real factor);
	static Hpc* rotate    (int spacedim,int i,int j,real angle);
	static Hpc* shear     (int spacedim,int ref,const real* factor);
	static Hpc* transform (int spacedim,const real* vmat,const real* hmat=0);

	//get limits
	void limits(real* m,real* M);

	//join
	static Hpc* join(int n,Hpc** pcs);

	//mkpol
	static Hpc* mkpol(int pointdim,int npoints,const real* points,int nhulls,const int** hulls);

	//ukpol
	void ukpol(int& pointdim,int& npoints,real*& points,int& nhulls,int**& hulls,bool bFace);

	//boolean operation

	static bsp_solver*  getSolver(int op,int n,Hpc** _v,real* vmat,real* hmat);

	static Hpc* booleanOp(int op,int n,Hpc** v);
	static Hpc* Or			(int n,Hpc** v);
	static Hpc* And		(int n,Hpc** v);
	static Hpc* Diff		(int n,Hpc** v);
	static Hpc* Xor		(int n,Hpc** v);

	// return 0 on error, 1 on success
	int exportSvg(char* outfilename,char* type,int width,int height,
					  bool back_to_front,
					  double vrp[3],double vuv[3],double vpn[3],double cop[3],
					  double front,double back,
					  double range_u[2],double range_v[2],
					  double range_x[2],double range_y[2]);

	//skeleton
	Hpc* skeleton(int n);

	//power
	static Hpc* power(Hpc* p1,Hpc* p2);

	//vrml
	bool vrml(const char* filename="untitled.wrl") const;

	//svg
	bool svg(double width,const char* filename="untitled.svg") const;

	//povray
	bool povray(const char* filename="untitled.pov") const;

}; //end class



/////////////////////////////////////////////////////////////////////////
// INLINES 
/////////////////////////////////////////////////////////////////////////

inline Hpc*				Hpc::get(int i)			{return pols[i];}
inline const Hpc*		Hpc::get(int i) const	{return pols[i];}
inline int				Hpc::num() const			{return pols.size();}
inline Hpc*				Hpc::cube(int dim)		{return new Hpc(ggraph::cuboid(dim));}
inline Hpc*				Hpc::simplex(int dim)	{return new Hpc(ggraph::simplex(dim));}



inline Hpc::Hpc(int _spacedim,const real* vmat,const real* hmat) 
	: spacedim(_spacedim),pointdim(0),vmat(0),hmat(0),g(0),refcount(1)
{
	if (vmat || hmat)
	{
		checkpoint(vmat && hmat);
		int matsize=(spacedim+1)*(spacedim+1);
		this->vmat=new real[matsize];memcpy(this->vmat,vmat,sizeof(real)*matsize);
		this->hmat=new real[matsize];memcpy(this->hmat,hmat,sizeof(real)*matsize);
	}
}

inline Hpc::Hpc(ggraph* _g) 
	: spacedim(_g?_g->pointdim:0),pointdim(_g?_g->pointdim:0),vmat(0),hmat(0),g(_g),refcount(1)
{
	if (g)
	{
		this->spacedim=g->spacedim;g->spacedim=g->pointdim;
		this->vmat=g->vmat;g->vmat=0;
		this->hmat=g->hmat;g->hmat=0;
	}
}


inline void Hpc::incrRefCount()
{
	++refcount;

	for (int i=0;i<num();i++)
		get(i)->incrRefCount();
}

inline Hpc* Hpc::clone() const
{
	checkpoint(refcount);
	((Hpc*)this)->incrRefCount();
	return ((Hpc*)this);
}

inline void Hpc::decrRefCount()
{
	for (int i=0;i<num();i++) 
		get(i)->decrRefCount();
	
	if (!--refcount)
		delete this; //automatic release of memory
}

inline void Hpc::release()
{
	decrRefCount();
}

inline void* Hpc::getProperty(const char* name) 
{
	std::map<std::string, void*>::iterator it;
	return ((it=prop.find(std::string(name)))==prop.end())?0:(*it).second;
}


inline Hpc* Hpc::modify(Hpc* src)
{
	//if there is only one reference it can be safely modified by the only owner
	if (src->refcount==1)
		return src;

	//otherwise create a private copy
	Hpc* ret=new Hpc;

	ret->spacedim	 = src->spacedim;
	ret->pointdim   = src->pointdim;
	ret->prop		 = src->prop;
	ret->vmat		 = ret->hmat = 0;
	ret->g			 = 0;
	ret->pols       = src->pols; //refcount for childs remains unchanged

	int matsize=(src->spacedim+1)*(src->spacedim+1);
	if (src->vmat) {ret->vmat=new real[matsize];memcpy(ret->vmat,src->vmat,sizeof(real)*matsize);}
	if (src->hmat) {ret->hmat=new real[matsize];memcpy(ret->hmat,src->hmat,sizeof(real)*matsize);}
	if (src->g   ) {ret->g=src->g->clone();}

	return ret;

}

inline Hpc* Hpc::addProperty(const char* name,void* prop)
{
	Hpc* ret=modify(this);
	ret->prop[std::string(name)]=prop;
	return ret;
}


inline Hpc* Hpc::embed(int newspacedim)
{
	if (newspacedim<=this->spacedim) 
		return this;

	Hpc* ret=modify(this);

	if (ret->vmat)
	{
		real* vmat=new real[(newspacedim+1)*(newspacedim+1)];
		real* hmat=new real[(newspacedim+1)*(newspacedim+1)];
		tmat<real>::copy(newspacedim+1,vmat,ret->spacedim+1,ret->vmat);
		tmat<real>::copy(newspacedim+1,hmat,ret->spacedim+1,ret->hmat);
		Swap(real*,vmat,ret->vmat);
		Swap(real*,hmat,ret->hmat);
		delete [] vmat;
		delete [] hmat;
	}

	ret->spacedim=newspacedim;
	return ret;
}


inline Hpc* Hpc::add(Hpc* src)
{	
	Hpc* ret=modify(this);
	ret=ret->embed(src->spacedim);
	ret->pointdim=max2(ret->pointdim,src->pointdim);
	src->incrRefCount();
	ret->pols.push_back(src);
	return ret;
}

// *********** TRANSFORMATIONS***************************

inline Hpc* Hpc::scale(int spacedim,int ref,real factor)
{
	assert(ref>0); //x->1 y->2 z->3 etc...
	
	Hpc* ret=new Hpc(spacedim);
	if (factor==1) return ret;
	ret->vmat=new real[(spacedim+1)*(spacedim+1)];
	ret->hmat=new real[(spacedim+1)*(spacedim+1)];
	tmat<real>::scale(spacedim+1,ref,factor,ret->vmat,ret->hmat);
	return ret;
}

inline Hpc* Hpc::translate(int spacedim,int ref,real factor)
{
	assert(ref>0); //x->1 y->2 z->3 etc...
	Hpc* ret=new Hpc(spacedim);
	if (!factor) return ret;
	ret->vmat=new real[(spacedim+1)*(spacedim+1)];
	ret->hmat=new real[(spacedim+1)*(spacedim+1)];
	tmat<real>::translate(spacedim+1,ref,factor,ret->vmat,ret->hmat);
	return ret;
}

inline Hpc* Hpc::rotate(int spacedim,int i,int j,real angle)
{
	assert(i>0 && j>0); //ex (1,2)->(x,y)
	Hpc* ret=new Hpc(spacedim);
	if (!angle) return ret;
	ret->vmat=new real[(spacedim+1)*(spacedim+1)];
	ret->hmat=new real[(spacedim+1)*(spacedim+1)];
	tmat<real>::rotate(spacedim+1,i,j,angle,ret->vmat,ret->hmat);
	return ret;
}

inline Hpc* Hpc::shear(int spacedim,int ref,const real* factor)
{
	assert(ref>0);
	Hpc* ret=new Hpc(spacedim);
	ret->vmat=new real[(spacedim+1)*(spacedim+1)];
	ret->hmat=new real[(spacedim+1)*(spacedim+1)];
	tmat<real>::shear(spacedim+1,ref,factor,ret->vmat,ret->hmat);
	return ret;
}

inline Hpc* Hpc::transform(int spacedim,const real* vmat,const real* hmat)
{
	Hpc* ret=new Hpc(spacedim);
	ret->vmat=new real[(spacedim+1)*(spacedim+1)];
	ret->hmat=new real[(spacedim+1)*(spacedim+1)];
	memcpy(ret->vmat,vmat,sizeof(real)*(spacedim+1)*(spacedim+1));

	if (hmat)
		memcpy(ret->hmat,hmat,sizeof(real)*(spacedim+1)*(spacedim+1));
	else
		tmat<real>::invert(spacedim+1,ret->hmat,ret->vmat);

	return ret;
}


// *********** BOOLEAN OPS***************************

inline Hpc* Hpc::Or	(int n,Hpc** v)		{return booleanOp(bsp_bool::BOOL_OR  ,n,v);}
inline Hpc* Hpc::And	(int n,Hpc** v)		{return booleanOp(bsp_bool::BOOL_AND ,n,v);}
inline Hpc* Hpc::Diff(int n,Hpc** v)		{return booleanOp(bsp_bool::BOOL_DIFF,n,v);}
inline Hpc* Hpc::Xor	(int n,Hpc** v)		{return booleanOp(bsp_bool::BOOL_XOR ,n,v);}

#endif //_hpc_H__




