/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#include "generaldef.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <vector>
#include <math.h>

extern "C" {
#include <qhull.h>
}

#include "plane.h"
#include "bsp.h"




// _______________________________________________________________________________________________

bsp::bsp(uint8 _dim)
	:dim(_dim),nodes(sizeof(node_t)),db(sizeof(real)*(_dim+1))
{
	block_expand=false;
	nin=0,nout=0,ncut=0;
	root=in();
}


bsp::~bsp()
{
	clear();
}


void bsp::visit(key N)
{
	if (!N) N=getRoot();
	if (!isCut(N)) return ; 
	visit(getBelow(N));
	visit(getAbove(N));
}



void bsp::copy(bsp* dst,key Ndst,bsp* src,key Nsrc)
{
	if (src->isIn(Nsrc))
	{dst->setIn(Ndst);return;}
	
	if (src->isOut(Nsrc))
	{dst->setOut(Ndst);return;}

	key Below,Above;
	dst->setCutn(Ndst,src->getPlane(Nsrc),Below=dst->in(),Above=dst->in());
	copy(dst,Below,src,src->getBelow(Nsrc));
	copy(dst,Above,src,src->getAbove(Nsrc));
}



void bsp::embed(int howmuch)
{
	int newdim=this->dim+howmuch;
	db.resize(sizeof(real)*(newdim+1));
	this->dim=newdim;
}

void bsp::more(key N,key More)
{}

void bsp::clear()
{
	nodes.clear();
	db.clear();
	nin=0,nout=0,ncut=0;
	root=in();
}


void bsp::print(key N,char* name,int space,int lvl)
{	
	bsp* b=this;
	int i;
	for (i=0;i<space*2;i++) printf(" ");

	if (b->isIn(N))	{printf(" /* %d.%d.Info(%d)*/ ",(int)N,(int)lvl,(int)b->nodeInfo(N));printf("%s.in()",name);return;}
	if (b->isOut(N))	{printf(" /* %d.%d.Info(%d)*/ ",(int)N,(int)lvl,(int)b->nodeInfo(N));printf("%s.out()",name);return;}

	real* g=b->getPlane(N);
	printf("%s.cut(",name);
	for (i=0;i<=b->getDim();++i) 
		printf("%e,",(float)g[i]);

	printf(" /* %d.%d.Info(%d)*/ ",(int)N,(int)lvl,(int)b->nodeInfo(N));

	printf("\n");
	
	print(b->getBelow(N),name,space+1,lvl+1);printf(",\n");
	print(b->getAbove(N),name,space+1,lvl+1);printf(")");
}


void bsp::print(char* name)
{
	printf("t* %s=create(%d);\n",name?name:"b",dim);
	printf("%s->root=\n",name?name:"b");
	print(getRoot(),name,1,0);
	printf(";\n");
}


// _______________________________________________________________________________________________

bsp_clone::bsp_clone(bsp* src)
	:bsp(src->getDim())
{
	this->src=src;
	setMore(getRoot(),src->getRoot());
}


void bsp_clone::more(key N,key More)
{
	if (src->isIn(More))
		{setOut(N);return;}

	if (src->isOut(More))
		{setIn(N);return;}

	key    Below,Above;
	real   hdst[MAX_GEOMETRIC_DIMENSION+1];
	real  *hsrc=src->getPlane(More);
	for (int i=0;i<=dim;i++) hdst[i]=-hsrc[i];
	setCutn(N,hdst,Below=in(),Above=in());
	setMore(Below,src->getBelow(More));
	setMore(Above,src->getAbove(More));
}

// _______________________________________________________________________________________________

bsp_complement::bsp_complement(bsp* src)
	:bsp(src->getDim())
{
	this->src=src;
	setMore(getRoot(),src->getRoot());
}

void bsp_complement::more(key N,key More)
{
	if (src->isIn(More))
		{setOut(N);return;}

	if (src->isOut(More))
		{setIn(N);return;}
	
	key    Below,Above;
	real   hdst[MAX_GEOMETRIC_DIMENSION+1];
	real  *hsrc=src->getPlane(More);
	for (int i=0;i<=dim;i++) hdst[i]=-hsrc[i];
	setCutn(N,hdst,Below=in(),Above=in());
	setMore(Below,src->getAbove(More));
	setMore(Above,src->getBelow(More));
}


// _______________________________________________________________________________________________


bsp_bool::bsp_bool(int dim,int code):
	bsp(dim),args(sizeof(info))
{
}

bsp_bool::bsp_bool(int code,bsp** src,key* vmax)
	:bsp(src[0]->getDim()),args(sizeof(info))
{
	checkpoint(code!=BOOL_XOR);//todo!

	int n=0;
	while (src[n]) n++;

	key I=block(n);
	info& i=*((info*)args[I]);
	i.num=n;
	i.cur=0;
	i.code=code;
	
	for (int k=0;k<n;k++)
	{
		i.bsps[k]=src[k];
		i.nodes[k]=src[k]->getRoot();
		i.m[k]=i.vmax[k]=(vmax)?vmax[k]:1;
	}

	setMore(getRoot(),I);
}

bsp_bool* bsp_bool::Union			(bsp** src,key* vmax) {return new bsp_bool(BOOL_OR		,src,vmax);}
bsp_bool* bsp_bool::Intersection	(bsp** src,key* vmax) {return new bsp_bool(BOOL_AND	,src,vmax);}
bsp_bool* bsp_bool::Difference	(bsp** src,key* vmax) {return new bsp_bool(BOOL_DIFF	,src,vmax);}
bsp_bool* bsp_bool::Xor          (bsp** src,key* vmax) {return new bsp_bool(BOOL_XOR   ,src,vmax);}


inline key bsp_bool::block(int num)
{
	key I=args.block();
	info& i=*((info*)args[I]);

	//necessary at least to have one element (for memcache)
	num=max2(1,num);

	i.bsps	=new bsp*[num];
	i.nodes	=new key[num];
	i.m		=new key[num];
	i.vmax	=new key[num];

	return I;
}

inline key bsp_bool::block(const info& src)
{
	key I=args.block();
	info& i=*((info*)args[I]);
	i=src;

	int n=max2(1,src.num);

	i.bsps	=new bsp*[n];memcpy(i.bsps		,src.bsps	,sizeof(bsp*) * src.num);
	i.nodes	=new key[n]	;memcpy(i.nodes	,src.nodes	,sizeof(key ) * src.num);
	i.m		=new key[n]	;memcpy(i.m			,src.m		,sizeof(key ) * src.num);
	i.vmax	=new key[n]	;memcpy(i.vmax    ,src.vmax   ,sizeof(key ) * src.num);

	return I;
}

inline void bsp_bool::release(key I)
{
	info& i=*((info*)args[I]);

	delete [] i.bsps;
	delete [] i.nodes;
	delete [] i.m;
	delete [] i.vmax;

	args.release(I);
}


void bsp_bool::more(key N,key More)
{
	info i=*((info*)args[More]);
	args.release(More);

REDO:

	bsp* bsrc		=i.bsps	[i.cur];
	key  Nsrc		=i.nodes	[i.cur];
	int  code      =i.code;

	if (!bsrc->isCut(Nsrc))
	{
		bool isin  = bsrc->isIn (Nsrc);
		bool isout = bsrc->isOut(Nsrc);

		if ((isin && code==BOOL_OR              ) || //OR:[...IN...] -> IN 
			 (isin && code==BOOL_AND  && i.num==1) || //AND:[IN]      -> IN
			 (isin && code==BOOL_DIFF && i.num==1)    //DIFF:[IN]     -> IN
			)
		{	
			setIn(N);
			return;
		}

		if ((isout && code==BOOL_AND)               || //AND:[....OUT...]   -> OUT 
			 (isout && code==BOOL_OR   && i.num==1)  || //OR:[OUT]           -> OUT
			 (isout && code==BOOL_DIFF && i.cur==0)  || //DIFF:[OUT,a,b,...] -> OUT
			 (isin  && code==BOOL_DIFF && i.cur>0 )     //DIFF:[a,....IN...] -> OUT
			 )
		{
			setOut(N);
			return;
		}
		
		//simplify the current argument
		key cont=0;
		for (key k=0;k<i.num;k++)
		{
			if (k!=i.cur || (i.code==BOOL_DIFF && i.cur==0)) //-:[IN,a,b,...] DO NOT simplify the first IN
			{
				i.bsps  [cont]	=i.bsps [k];
				i.nodes [cont]	=i.nodes[k];
				i.m     [cont] =i.m    [k];
				i.vmax  [cont] =i.vmax [k];
				++cont;
			}
		}

		i.num=cont;
		i.cur=(i.cur+1) % i.num;
		goto REDO;
	}

	real hdst[MAX_GEOMETRIC_DIMENSION+1];
	memcpy(hdst,bsrc->getPlane(Nsrc),sizeof(real)*(dim+1));

	key Below,Above;
	setCutn(N,hdst,Below=in(),Above=in());

	//go to the next if I've done n-planes for the current arguments
	key Ibelow=block(i),Iabove=block(i);

	info& ibelow=*((info*)args[Ibelow]);
	info& iabove=*((info*)args[Iabove]);

	ibelow.nodes[ibelow.cur]=bsrc->getBelow(Nsrc);
	iabove.nodes[iabove.cur]=bsrc->getAbove(Nsrc);

	if (ibelow.m[ibelow.cur] && bsrc->isCut(ibelow.nodes[i.cur]) && !(--ibelow.m[ibelow.cur]))
	{
		ibelow.m[ibelow.cur]=ibelow.vmax[ibelow.cur];
		ibelow.cur=(ibelow.cur+1) % (ibelow.num);
	}

	if (iabove.m[iabove.cur] && bsrc->isCut(iabove.nodes[i.cur]) && !(--iabove.m[iabove.cur]))
	{
		iabove.m[iabove.cur]=iabove.vmax[iabove.cur];
		ibelow.cur=(ibelow.cur+1) % (ibelow.num);
	}

	setMore(Below,Ibelow);
	setMore(Above,Iabove);
}




// _______________________________________________________________________________________________

bsp_cube::bsp_cube(int dim,real* limits)
	:bsp(dim)
{
	bsp::cube(root,limits);
}


// _______________________________________________________________________________________________

bsp_grid::bsp_grid(int dim,uint8* v,real* limits)
	:bsp(dim),args(sizeof(info))
{
	for (int k=0;;k++)
	{
		this->v[k]=v[k];
		if (v[k]==(uint8)-1) break;
	}

	key cellin=bsp::cube(root,0);

	key I=args.block();
	info& i=*((info*)args[I]);
	i.num=0;

	if (limits)
		memcpy(i.limits,limits,sizeof(real)*dim*2);
	else
		for (int k=0;k<dim;k++) {i.limits[k*2+0]=0;i.limits[k*2+1]=1;}

	setMore(cellin,I);
}


void bsp_grid::more(key N,key More)
{
	info i=*((info*)args[More]);
	args.release(More);

	if (v[i.num]==(uint8)-1) 
		{setIn(N);return;}

	uint8 ref=v[i.num];

	real from   = i.limits[ref*2+0];
	real to     = i.limits[ref*2+1];
	real middle = (from+to)/2;

	real h[MAX_GEOMETRIC_DIMENSION+1];
	memset(h,0,sizeof(h));
	h[ref+1]=1;h[0]=-middle;

	key Below,Above;
	setCutn(N,h,Below=in(),Above=in());

	key I1=args.block(),I2=args.block();
	info& i1=*((info*)args[I1]);
	info& i2=*((info*)args[I2]);
	++i.num;
	i1=i2=i;
	
	i1.limits[ref*2+0]=from  ;i1.limits[ref*2+1]=middle;
	i2.limits[ref*2+0]=middle;i2.limits[ref*2+1]=to;
	setMore(Below,I1);
	setMore(Above,I2);	
}

// _______________________________________________________________________________________________

key bsp::cube(key N,real* limits)
{
	real h[MAX_GEOMETRIC_DIMENSION+1];
	memset(h,0,sizeof(real)*(dim+1));

	setIn(N);
	key cur=N;
	
	for (int i=1;i<=dim;i++)
	{	
		h[0]=limits?limits[(i-1)*2+0]:+0;h[i]=-1;
		setCutn(cur,h,in(),out());
		cur=getBelow(cur);

		h[0]=limits?-limits[(i-1)*2+1]:-1;h[i]=+1;
		setCutn(cur,h,in(),out());
		cur=getBelow(cur);
		h[i]=0;
	}

	return cur;
}


// _______________________________________________________________________________________________


bsp_circle::bsp_circle(int nsteps)
	:bsp(2),args(sizeof(info))
{
	real h[MAX_GEOMETRIC_DIMENSION+1];
	memset(h,0,sizeof(real)*(dim+1));
	
	real step=(2*PIGRECO/nsteps);
	real angle=0;
	key N=getRoot();

	for (int i=0;i<nsteps;i++,angle+=step)
	{
		setCut2(N,-1,cos(angle),sin(angle),in(),out());
		N=getBelow(N);
	}
}


bsp_circle::bsp_circle(int nrec,int ntg)
	:bsp(2),args(sizeof(info))
{
	this->ntg= ntg;

	key K[4];
	real limits[]={-1,+1,-1,+1};

	key N=cube(root,limits);

	setCut2(N,0,1,0,
		cut2(0,0,1,K[0]=in(),K[1]=in()),
		cut2(0,0,1,K[2]=in(),K[3]=in()));

	key I[]={args.block(),args.block(),args.block(),args.block()};

	info p[]={{nrec,+1.0,+0.5},	{nrec,+0.5,+0.5},	{nrec,-0.5,+0.5},	{nrec,+0.0,+0.5}};

	for (int n=0;n<4;n++)
	{
		*((info*)args[I[n]])=p[n];
		setMore(K[n],I[n]);
	}
}

void bsp_circle::more(key N,key More)
{
	info i=*((info*)args[More]);
	args.release(More);

	if (!(i.nrec--)) 
		return ;

	for (int I=0;I<=ntg;++I)
	{
		real a=(PIGRECO  )*(i.A+i.DA*(real)I/(real)ntg);
		setCut2(N,-1,cos(a),sin(a),in(),out());
		N=getBelow(N);
	}

	real a  =(PIGRECO     )*(i.A+i.DA/2);
	real da =(PIGRECO*i.DA)/2;
	real ca =(real)cos( a);
	real sa =(real)sin( a);
	real cda=(real)cos(da);
	real paa=(real)-1*cda; real pab=(real)+ca; real pac=(real)+sa; 
	real xya=(real)+0    ; real xyb=(real)-sa; real xyc=(real)+ca; 

	key Below,Above;
	setCut2(N,paa,pab,pac,in(),cut2(xya,xyb,xyc,Below=in(),Above=in()));

	key I1=args.block(),I2=args.block();

	info& i1=*((info*)args[I1]);
	info& i2=*((info*)args[I2]);

	i.DA/=2;
	i1=i2=i;
	i1.A=i.A      ; 
	i2.A=i.A+i.DA ; 

	setMore(Below,I1);
	setMore(Above,I2);
}

// ______________________________________________________________________________________________

bsp_sphere::bsp_sphere(int nsteps)
	:bsp(3),args(sizeof(info))
{
	key N=root;
	real step=(2*PIGRECO/nsteps);

	for (real a=-PIGRECO/2 ; a<PIGRECO/2 ; a+=step)
	for (real  b=         0 ; b<2*PIGRECO ; b+=step)
	{
		real ca=(real)cos(a);real sa=(real)sin(a);
		real cb=(real)cos(b);real sb=(real)sin(b);
		setCut3(N,-1,ca*cb,sa*cb,sb,in(),out());
		N=getBelow(N);
	}
}


bsp_sphere::bsp_sphere(int nrec,int ntg)
	:bsp(3),args(sizeof(info))
{
	real limits[]={-1,+1,-1,+1,-1,+1};
	key CellIn=cube(getRoot(),limits);

	this->ntg=ntg;

	int n;
	key I[8],K[8];
		
	setCut3(CellIn,0,1,0,0,
		cut3(0,0,1,0,
			cut3(0,0,0,1,K[0]=in(),K[1]=in()),
			cut3(0,0,0,1,K[2]=in(),K[3]=in())),
		cut3(0,0,1,0,
			cut3(0,0,0,1,K[4]=in(),K[5]=in()),
			cut3(0,0,0,1,K[6]=in(),K[7]=in())));

	for (n=0;n<8;n++) 
		I[n]=args.block(); 

	info p[]=
	{
		{nrec,-1.0,0.5,-1.0,1.0},
		{nrec,-1.0,0.5,+0.0,1.0},
		{nrec,+0.5,0.5,-1.0,1.0},
		{nrec,+0.5,0.5,+0.0,1.0},
		{nrec,-0.5,0.5,-1.0,1.0},
		{nrec,-0.5,0.5,+0.0,1.0},
		{nrec,+0.0,0.5,-1.0,1.0},
		{nrec,+0.0,0.5,+0.0,1.0}
	};

	for (n=0;n<8;n++)
	{
		*((info*)args[I[n]])=p[n];
		setMore(K[n],I[n]);
	}
}


void bsp_sphere::more(key N,key More)
{	
	info i=*((info*)args[More]);
	args.release(More);

	if (!(i.nrec--)) 
	{
		setIn(N);
		return ;
	}

	for (int I=0;I<=ntg;++I)
	for (int J=0;J<=ntg;++J)
	{
		real a  = (PIGRECO  )*(i.A+i.DA*(real)I/(real)ntg);
		real b  = (PIGRECO/2)*(i.B+i.DB*(real)J/(real)ntg);
		real ca = (real)cos(a);real sa=(real)sin(a);
		real cb = (real)cos(b);real sb=(real)sin(b);
		setCut3(N,-1,ca*cb,sa*cb,sb,in(),out());
		N=getBelow(N);
	}
	
	real a  =(PIGRECO   )*(i.A+i.DA/2);real da =(PIGRECO*i.DA)/2;
	real b  =(PIGRECO/2 )*(i.B+i.DB/2);real db =(PIGRECO/2 )*i.DB/2;
	
	real ca =(real)cos( a)   ; real cb =(real)cos( b);
	real sa =(real)sin( a)   ; real sb =(real)sin( b);
	real cda=(real)cos(da)   ; real cdb=(real)cos(db);
	real paa=(real)-1*cda*cdb; real pab=(real)+ca*cb ; real pac=(real)+sa*cb; real pad=+sb*cda;
	real xya=(real)+0        ; real xyb=(real)-sa*cb ; real xyc=(real)+ca*cb; real xyd=+0;
	real zza=(real)-1*sb     ; real zzb=(real)+0     ; real zzc=(real)+0    ; real zzd=+1;

	i.DA/=2;i.DB/=2;
	key NA,NB,NC,ND;

	setCut3(N,paa,pab,pac,pad, 
		in(),
		cut3(xya,xyb,xyc,xyd, 
			cut3(zza,zzb,zzc,zzd,NA=in(),NB=in()),
			cut3(zza,zzb,zzc,zzd,NC=in(),ND=in())));
	
	key I1=args.block(),I2=args.block(),I3=args.block(),I4=args.block();

	info& i1=*((info*)args[I1]);
	info& i2=*((info*)args[I2]);
	info& i3=*((info*)args[I3]);
	info& i4=*((info*)args[I4]);
	
	i1=i2=i3=i4=i;

	i1.A=i.A      ; i1.B=i.B;
	i2.A=i.A      ; i2.B=i.B+i.DB;
	i3.A=i.A+i.DA ; i3.B=i.B;
	i4.A=i.A+i.DA ; i4.B=i.B+i.DB;

	setMore(NA,I1);
	setMore(NB,I2);
	setMore(NC,I3);
	setMore(ND,I4);
}


// _______________________________________________________________________________________________

bsp_extrude::bsp_extrude(bsp* src,real From,real To)
	:bsp(src->getDim()+1)
{
	this->src=src;

	real h0[MAX_GEOMETRIC_DIMENSION+1],h1[MAX_GEOMETRIC_DIMENSION+1];
	memset(h0,0,sizeof(h0));
	memset(h1,0,sizeof(h1));

	h0[dim]= 1;h0[0]=-To ; //z>=To
	h1[dim]=-1;h1[0]=From; //z<=From

	key N;

	this->root=cutn(h0,
		cutn(h1,N=in(),out()),
		out());

	setMore(N,src->getRoot());
}


void bsp_extrude::more(key N,key More)
{
	if (src->isIn(More))
		{setIn(N);return;}

	if (src->isOut(More))
		{setOut(N);return;}

	key    Below,Above;
	real   hdst[MAX_GEOMETRIC_DIMENSION+1];
	memcpy(hdst,src->getPlane(More),sizeof(real)*dim);
	hdst[dim]=0;
	setCutn(N,hdst,Below=in(),Above=in());
	setMore(Below,src->getBelow(More));
	setMore(Above,src->getAbove(More));
}


// _______________________________________________________________________________________________

bsp_transform::bsp_transform(int code,bsp* src,real* v)
	:bsp(src->getDim())
{
	this->src=src;
	this->code=code;

	if      (code==TRANSFORM)	memcpy(this->v,v,sizeof(real)*(dim+1)*(dim+1));
	else if (code==ROTATE)		memcpy(this->v,v,sizeof(real)*3); //ax1[1,n..],ax2[1..n],angle
	else if (code==SHEAR)		memcpy(this->v,v,sizeof(real)*(dim+1)); //[ref shearx,sheary,shearz]
	else								memcpy(this->v,v,sizeof(real)*dim);//[vx,vy,vz] | [sx,sy,sz] | ...

	setMore(getRoot(),src->getRoot());
};


void bsp_transform::more(key N,key More)
{
	if (src->isIn(More))
		{setIn(N);return;}

	if (src->isOut(More))
		{setOut(N);return;}


	key    Below,Above;
	real   hdst[MAX_GEOMETRIC_DIMENSION+1];
	real  *hsrc=src->getPlane(More);
	memcpy(hdst,hsrc,sizeof(real)*(dim+1));

	switch (code)
	{
		case TRANSLATE:
		{
			switch(dim)
			{
			case 0:break;
			case 1:hdst[0]-=hdst[1]*v[0];break;
			case 2:hdst[0]-=hdst[1]*v[0]+hdst[2]*v[1];break;
			case 3:hdst[0]-=hdst[1]*v[0]+hdst[2]*v[1]+hdst[3]*v[2];break;
			default:for (int k=1;k<=dim;k++) hdst[0]-=hdst[k]*v[k-1];break;
			}
			break;
		}
		case SCALE:
		{
			switch(dim)
			{
			case 0:break;
			case 1:hdst[1]/=v[0];break;
			case 2:hdst[1]/=v[0];hdst[2]/=v[1];break;
			case 3:hdst[1]/=v[0];hdst[2]/=v[1];hdst[3]/=v[2];break;
			default:for (int k=1;k<=dim;k++) hdst[k]/=v[k-1];break;
			}
			break;
		}

		case ROTATE:
		{
			int ax1=(int)v[0],ax2=(int)v[1];
			real alpha=v[2];
			real c=(real)cos(alpha);
			real s=(real)sin(alpha);
			hdst[ax1]=+c*hsrc[ax1]+s*hsrc[ax2];
			hdst[ax2]=-s*hsrc[ax1]+c*hsrc[ax2];
			break;
		}

		case SHEAR:
		{
			int ref=(int)v[0];	
			
			switch(dim)
			{
			case 0:
				break;

			case 1:
				if (ref!=1) hdst[1]-=hdst[ref]*v[0]/v[ref-1];
				hdst[ref]/=v[ref-1];
				break;

			case 2:
				if (ref!=1) hdst[1]-=hdst[ref]*v[0]/v[ref-1];
				if (ref!=2) hdst[2]-=hdst[ref]*v[1]/v[ref-1];
				hdst[ref]/=v[ref-1];
				break;

			case 3:
				if (ref!=1) hdst[1]-=hdst[ref]*v[0]/v[ref-1];
				if (ref!=2) hdst[2]-=hdst[ref]*v[1]/v[ref-1];
				if (ref!=3) hdst[3]-=hdst[ref]*v[2]/v[ref-1];
				hdst[ref]/=v[ref-1];
				break;

			default:
				for (int k=1;k<=dim;k++) if (k!=ref) hdst[k]-=hdst[ref]*v[k-1]/v[ref-1];
				hdst[ref]/=v[ref-1];
				break;
			}

			break;
		}

		case TRANSFORM:
		{
			switch(dim)
			{
			case 0:
				hdst[0]=hsrc[0]*v[0];
				break;

			case 1:
				hdst[0]=hsrc[0]*v[0]+hsrc[1]*v[2];
				hdst[1]=hsrc[0]*v[1]+hsrc[1]*v[3];
				break;

			case 2:
				hdst[0]=hsrc[0]*v[0]+hsrc[1]*v[3]+hsrc[2]*v[6];
				hdst[1]=hsrc[0]*v[1]+hsrc[1]*v[4]+hsrc[2]*v[7];
				hdst[2]=hsrc[0]*v[2]+hsrc[1]*v[5]+hsrc[2]*v[8];
				break;

			case 3:
				hdst[0]=hsrc[0]*v[0]+hsrc[1]*v[4]+hsrc[2]*v[ 8]+hsrc[3]*v[12];
				hdst[1]=hsrc[0]*v[1]+hsrc[1]*v[5]+hsrc[2]*v[ 9]+hsrc[3]*v[13];
				hdst[2]=hsrc[0]*v[2]+hsrc[1]*v[6]+hsrc[2]*v[10]+hsrc[3]*v[14];
				hdst[3]=hsrc[0]*v[3]+hsrc[1]*v[7]+hsrc[2]*v[11]+hsrc[3]*v[15];					
				break;

			default:
				{
					memset(hdst,0,sizeof(real)*(dim+1));
						
					for (int I=0;I<=dim;I++)
					for (int J=0;J<=dim;J++)
						hdst[J]+=hsrc[I]*v[I*(dim+1)+J];

					break;
				}
			}
			break;
		}
	}

	normalize_plane(dim,hdst);
	setCutn(N,hdst,Below=in(),Above=in());
	setMore(Below,src->getBelow(More));
	setMore(Above,src->getAbove(More));
}



// _______________________________________________________________________________________________

bsp_hull::bsp_hull(int pointdim,int npoints,double* points,real tolerance)
	:bsp(pointdim)
{
#ifdef WIN32
	static FILE* devnull=fopen("nul","w");
#else
	static FILE* devnull=fopen("/dev/null","w");
#endif

	checkpoint(pointdim>=1 && npoints>=1);

	key cur=getRoot();

	if (pointdim==1)
	{
		real m=points[0],M=points[0];

		for (int i=1;i<npoints;i++)
		{
			m=(real)min2(m,points[i]);
			M=(real)max2(M,points[i]);
		}

		setCut1(getRoot(),m,-1,cut1(-M,1,in(),out()),out());
		return;
	}

	//Pp=does not print warnings, n=print hyperplane normals with offsets, En roudoff, Qs search all points for initial simplex
	char qhull_opts[256];sprintf(qhull_opts,"qhull n Pp E%e Qs",tolerance);
	int error_qhull=qh_new_qhull(pointdim,npoints,points,(boolT)false,qhull_opts, NULL, devnull);

	if (error_qhull)  
	{
		qh_freeqhull(qh_ALL);
		setOut(cur);
		return;
	}

	real g[MAX_GEOMETRIC_DIMENSION+1];
	facetT* facet;
	FORALLfacets 
	{
		g[0]=(real)facet->offset;
		for (int k=1;k<=pointdim;k++) g[k]=(real)facet->normal[k-1];
		setCutn(cur,g,in(),out());
		cur=getBelow(cur);
	}
	qh_freeqhull(qh_ALL);
}



// _______________________________________________________________________________________________

bsp_solver::bsp_solver(bsp* src,real tolerance,int maxdeep,real limit,bool simplify)
	:bsp(src->getDim()),args(sizeof(info))
{
	this->src=src;
	this->tolerance=tolerance;
	this->maxdeep=0;
	this->limit=limit;
	this->simplify=simplify;
	this->g=new ggraph(src->getDim());
	clear();
}

bsp_solver::~bsp_solver()
{
	delete g;
}


void bsp_solver::setSimplify(bool simplify)
{
	this->simplify=simplify;
}

void bsp_solver::clear()
{
	args.clear();

	delete g;
	g=ggraph::cuboid(getDim());
	
	real vt[MAX_GEOMETRIC_DIMENSION+1],vs[MAX_GEOMETRIC_DIMENSION+1];

	for (int u=0;u<getDim();u++)
	{
		vt[u]=-0.5;
		vs[u]=2*limit;
	}
	g->translate(vt);
	g->scale(vs);

	key I=args.block();
	info& i=*((info*)args[I]);
	i.Cell=*g->each(dim);
	i.Node=src->getRoot();
	i.deep=0;
	setMore(root,I);
}


void bsp_solver::split(ggraph* g,real* h,key Cell,key& cminus,key& cequal,key& cplus,real tolerance,int maxnumtry,key Id)
{
	int retcode=g->split(Cell,h,tolerance,cminus,cplus,cequal,Id);

	for (int numtry=0;retcode==ggraph::IMPOSSIBLE && numtry<maxnumtry;numtry++)
	{
		real perplane[MAX_GEOMETRIC_DIMENSION+1];
		printf(" perturbation ");fflush(stdout);

		for (int k=0;k<=g->pointdim;k++) 
			perplane[k]=(real)(h[k]+numtry*1e-4*(rand()/(real)RAND_MAX-0.5));

		retcode=g->split(Cell,perplane,tolerance,cminus,cplus,cequal,Id);
	}

	 //assign to above
	if (retcode==ggraph::FLAT)
	{
		cplus		=Cell;
		cminus	=cequal=0;
		retcode	=ggraph::OK;
	}

	checkpoint(retcode==ggraph::OK);
}


ggraph* bsp_solver::getGraph()
{
	return this->g;
}

void bsp_solver::setMaxDeep(int value)
{
	maxdeep=max2(0,value);
}

void bsp_solver::more(key N,key More)
{
	info i=*((info*)args[More]);
	args.release(More);
	
	if (!i.Cell)
		{setOut(N);return;}

REDO:

	checkpoint(i.Cell);

	if (src->isIn(i.Node))
	{
CASE_IN:
		nodeInfo(N)=i.Cell;
		setIn(N);
		return;
	}

	if (src->isOut(i.Node))
	{
CASE_OUT:
		if (i.Cell) g->remNode(i.Cell); 
		setOut(N);
		return;
	}

	if (maxdeep && i.deep>=maxdeep)
	{
		setIn(N);
		return;
	}

	//special case (simplification)
	key srcbelow=src->getBelow(i.Node);
	key srcabove=src->getAbove(i.Node);

	if (simplify)
	{
		if (src->isIn(srcbelow ) && src->isIn(srcabove )) goto CASE_IN;
		if (src->isOut(srcbelow) && src->isOut(srcabove)) goto CASE_OUT;
	}

	key cminus,cequal,cplus;
	real* h=src->getPlane(i.Node);
	split(g,h,i.Cell,cminus,cequal,cplus,tolerance,100,N); //important: the Id is unique, if i had used the <h> I cannot be sure of it!!!

	//useless cutting plane 
	if (simplify)
	{	
		if ((cminus && !cplus) || (cplus && !cminus))
		{
			checkpoint(cminus==i.Cell || cplus==i.Cell);
			i.Node=cminus?src->getBelow(i.Node):src->getAbove(i.Node);
			goto REDO;
		}
	}

	key Below,Above;
	setCutn(N,h,Below=in(),Above=in());
	
	if (cminus && !src->isOut(src->getBelow(i.Node)))
	{
		key Ibelow=args.block();
		info& ibelow=*((info*)args[Ibelow]);
		ibelow.Cell=cminus;
		ibelow.Node=src->getBelow(i.Node);
		ibelow.deep=i.deep+1;
		setMore(Below,Ibelow);
	}
	else
	{
		if (cminus) g->remNode(cminus);
		setOut(Below);
	}
	
	if (cplus && !src->isOut(src->getAbove(i.Node)))
	{
		key Iabove=args.block();
		info& iabove=*((info*)args[Iabove]);
		iabove.Cell=cplus;
		iabove.Node=src->getAbove(i.Node);
		iabove.deep=i.deep+1;
		setMore(Above,Iabove);
	}
	else
	{
		if (cplus) g->remNode(cplus);
		setOut(Above);
	}
}











// _______________________________________________________________________________________________

bsp_repair::bsp_repair(bsp* src,real tolerance,int maxdeep,real limit,real expand,real toofar)
	:bsp_solver(src,tolerance,maxdeep,limit)
{
	this->expand=expand;
	this->toofar=toofar;
};


void bsp_repair::more(key N,key More)
{	
	bsp_solver::more(N,More);

	if (isIn(N) && nodeInfo(N))
	{	
		key Cell=nodeInfo(N);
		const real* c=g->boundingBall(Cell);

		for (int j=1;j<=g->pointdim;j++)
		{
			//the current cell goes too far
			if (fabs(c[j]*expand)>toofar)
			{
				g->remNode(Cell);
				setOut(N);
				return;
			}
		}
	}
}



// _______________________________________________________________________________________________

bsp_brep::bsp_brep(int dim)
{
	this->g=new ggraph(dim);
	this->b=new bsp(dim);
}

bsp_brep::bsp_brep(ggraph* g)
{
	this->g=g;
	this->b=new bsp(g->pointdim);
}

bsp_brep::~bsp_brep()
{
	if (g) delete this->g;
	if (b) delete this->b;
}

ggraph* bsp_brep::getGraph()
	{return g;}

bsp* bsp_brep::getBsp()
	{return b;}


void bsp_brep::split(ggraph* g,real* h,key F,key& Fminus,key& Fplus,real tolerance)
{
	Fminus =0;
	Fplus  =0;
	uint8 lvl=g->getLevel(F);

	while (g->getNDw(F))
	{
		key CDown=g->getFirstDwNode(F);

		key CDownMinus,CDownPlus,CDownEqual;

		//todo: perturbation in case of classification error
		int retcode=g->split(CDown,h,tolerance,CDownMinus,CDownPlus,CDownEqual);

		switch(retcode)
		{
		case (ggraph::OK):
			{
				if (CDownMinus) 
				{
					if (!Fminus) Fminus =g->addNode(lvl);
					g->addArch(CDownMinus,Fminus);
					g->remArch(CDownMinus,F);
				}

				if (CDownPlus) 
				{
					if (!Fplus) Fplus =g->addNode(lvl);
					g->addArch(CDownPlus ,Fplus);
					g->remArch(CDownPlus, F);
				}
				break;
			} 
		case (ggraph::FLAT):
			{
				g->remArch(CDown,F);
				break;
			}
		default:
			{
				fprintf(stderr,"internal error. impossible to split in file %s line %d!\n",__FILE__,__LINE__);
				abort();
				break;
			}
		}
	}

	g->remNode(F);
}


void bsp_brep::convert(bsp* b,key N,ggraph* g,key F,int nrec)
{
	key lvl=g->getLevel(F);
	checkpoint(lvl==g->pointdim);

	int ndown=g->getNDw(F);

	key CDown=g->getFirstDwNode(F);
	checkpoint(CDown);
	g->remArch(CDown,F);

	real h[MAX_GEOMETRIC_DIMENSION+1];
	memcpy(h,g->getGeom(CDown),sizeof(real)*(g->pointdim+1));
	b->setCutn(N,h,b->in(),b->out());

	key Fminus,Fplus;
	split(g,h,F,Fminus,Fplus);

	if (Fminus) convert(b,b->getBelow(N),g,Fminus,nrec+1);
	if (Fplus ) convert(b,b->getAbove(N),g,Fplus ,nrec+1);
}


void bsp_brep::convert(bool bScaleToUnitBox)
{
	if (bScaleToUnitBox)
		g->scaleToUnitBox();
	
	clock_t t1=clock();
	printf("Converting to bsp...");

	//assume there is only one pointdim-cell
	checkpoint(g->each(g->pointdim).next().end());
	key C=*(g->each(g->pointdim));
	
	convert(b,b->getRoot(),g,C);
	clock_t t2=clock();
	printf("done in %.2f seconds\n",(t2-t1)/(float)CLOCKS_PER_SEC);
}

