/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _BSP_H__
#define _BSP_H__

#include "generaldef.h"
#include "memcache.h"
#include "ggraph.h"


/*
------------------------------------------------------------------------------------
CLASSNAME: BSP 

The principal class which handle Binary Space Partitioning (BSP)
data structures. 

It can support dinamically expand of the tree structure: 
each function set a special param <More> in the current node <N>; 
if this param <More>!=0 then the bsp-core automatically calls a pre-defined function:

	virtual void more(key N,key More);

telling that the node <N> should be expanded, and that the callback param is <More>.

The callback-function can then change the "content" of the node <N> and eventually
setting the <More> param for below,above subspace in case of cutting plane (so
continuing the dynamic expansion of the BSP). 

Example:

//a test bsp which builds a bsp with an IN node only at level=8
//for level<8 it builds cutting planes x>=0

class bsp_test: public bsp
{
	virtual void more(key N,key lvl)
	{
		if (lvl>=8)
			{setIn(N);return;}

		key below,above;
		setCut2(N,0,1,0,below=in(),above=in());
		setMore(below,lvl+1);
		setMore(above,lvl+1);
	}

public:

	bsp_test():bsp(2)
	{
		setMore(getRoot(),1);
	}
};

------------------------------------------------------------------------------------
*/

class bsp
{
protected:

	enum 
	{
		BSP_OUT=0,
		BSP_IN=1
	};


	struct node_t
	{
		uint8 type;				//BSP_IN or BSP_OUT. 
									//A node is IN  iif h==0 && type==BSP_IN
									//A node is OUT iif h==0 && type==BSP_OUT
									//A node is CUT iif h<>0

		key   h;					//if h<> 0 then is a cutting
		key   below,above;	//below and above subspace
		key   Info;				//special information (store here whatever you want)
		key   More;				//special callback-param for dynamically expand
	};

	//dimension of the bsp
	uint8       dim;

	//key for the root node
	key         root;

	//container for bsp nodes
	memcache    nodes;

	//container for planes
	memcache    db;

	//count number of in,out and cut nodes
	key         nin,nout,ncut;



	//for dynamic generated bsp (default: do nothing)
	virtual void more(key N,key More);

	//create the cub in node <N>
	//return the last in-cell which represents the cube
	//limits could be expressed as [x1,x2,y1,y1,...] (default: 0,1,0,1,...)
	key cube(key N,real* limits=0);

	//special function which copy nodes of a bsp to a specific node of a destination-bsp
	static void copy(bsp* dst,key Ndst,bsp* src,key Nsrc);
	
public:


	//stop dynamiccally expand
	bool        block_expand;

	//constructor. dimension should be specified here
	bsp(uint8 _dim);

	//destructor
	~bsp();

	//clear all the nodes (and planes). The root will be a single IN-node
	virtual void clear();

	//access the root node
	key getRoot();

	//get the plane dimension
	int   getDim() const;

	//access the <N> node
	node_t& getNode(key N);

	//the node <N> needs to be processed
	void setMore(key N,key More);

	//access node information
	key   getBelow     (key N);
	key   getAbove     (key N);
	real* getPlane     (key N);
	key&  nodeInfo     (key N);

	//construct in, out, cut nodes in the current bsp
	key in();
	key out();

	//for n-dim, 2, 3 planes
	key cutn(const real* h,						key below,key above);
	key cut1(real a,real b,                key below,key above);
	key cut2(real a,real b,real c,			key below,key above);
	key cut3(real a,real b,real c,real d,	key below,key above);

	//set node N to in,out,cut
	void setIn(key N);
	void setOut(key N);

	void setCutn(key N,const real* h,key below,key above);
	void setCut1 (key N,real a,real b					,key below,key above);
	void setCut2 (key N,real a,real b,real c			,key below,key above);
	void setCut3 (key N,real a,real b,real c,real d ,key below,key above);

	//test the type of the node
	bool isIn (key N) ;
	bool isOut(key N) ;
	bool isCut(key N) ;

	void print(key N,char* name="b",int space=0,int lvl=0);
	void print(char* name="b");

	//visit the bsp. This function could be helpfull to
	//expand all the bsp. Since each node is visited all <more>
	//functions will be invoked
	void visit(key N=0);

	//get number of IN,OUT,CUT nodes
	int getNumberOfIn()  const ;
	int getNumberOfOut() const ;
	int getNumberOfCut() const ;

	// embed a tree (behaviour undefined if the tree is not totally exploded)
	void embed(int howmuch);

};

//_________________________________________________________________________
//a bsp-definition of a cube (not dynamic!)

class bsp_cube:public bsp
{
public:
	//limits [x1,x2,y1,y2,...] (default:0,1,0,1,...)
	bsp_cube(int dim,real* limits=0);
};


//_________________________________________________________________________
//a bsp-definition of a quad-tree, octree, etc (dynamic!)
class bsp_grid:public bsp
{
	typedef struct
	{
		int  num;
		real limits[MAX_GEOMETRIC_DIMENSION*2];
	}
	info;

	memcache args;

	uint8 v[2048]; //should be enough for a grid with 2^2048 cells

	//dinamically expand. The <More> arg is the key for <args> 
	//which will give the actual bounding box (see info.limits) and the
	//pointer to the <v>[<num>] which specify the carthesian axe to use
	virtual void more(key N,key More);

public:

	//constructor
	//Example for a grid which is splitted in the following order (x,y,x,y) 
	//so the grid will contains 2^4=16 cells
	//
	// uint8 v[]={0,1,0,1,-1};
	// real  l[]={x1,x2,y1,y2};
	// bsp_grid* g=new bsp_grid(2,v,l); 
	bsp_grid(int dim,uint8* v,real* limits=0);
};


//_________________________________________________________________________
//bsp-definition for static or progressive 2d-circle
class bsp_circle: public bsp
{
	typedef struct
	{
		uint8 nrec;
		real A,DA;
	}
	info;

	memcache args;
	uint8 ntg;

	//dinamically expansion of progressive circle
	//info.A and info.DA specify the actual circular sector
	//info.nrec the actual "level" of recursion 
	virtual void more(key N,key More);

public:

	//progressive circle
	bsp_circle(int nrec,int ntg);

	//static circle
	bsp_circle(int nsteps);

};


//_________________________________________________________________________
//bsp-definition for static or progressive 3d-sphere
class bsp_sphere : public bsp
{
	typedef struct
	{
		uint8 nrec;
		real A,DA,B,DB;
	}
	info;

	memcache args;
	uint8 ntg;

	//dinamically expansion of progressive sphere
	//info.A, info.DA, info.B, info.DB specify the actual sphere sector
	//info.nrec the actual "level" of recursion 
	virtual void more(key N,key More);

public:

	//progressive sphere
	bsp_sphere(int nrec,int ntg);

	//static sphere
	bsp_sphere(int nsteps);

};


//_________________________________________________________________________
//a special bsp which dynamically clones another bsp
class bsp_clone: public bsp
{
	bsp* src;

	//the <More> argument is only the src-node for the src-bsp (ie the source-node to clone in
	//the actual node <N>.
	virtual void more(key N,key More);

public:

	//constructor
	bsp_clone(bsp* src);

};

//_________________________________________________________________________
//a bsp which dinamically complement a source-bsp
//by inverting IN and OUT nodes (and reverting the planes)
class bsp_complement: public bsp
{
	bsp* src;

	virtual void more(key N,key More);

public:

	//constructor
	bsp_complement(bsp* src);

	//static component
	static void run(bsp* src,key N)
	{

		if (src->isIn(N))
			{src->setOut(N);return;}

		if (src->isOut(N))
			{src->setIn(N);return;}
	
		real  *hsrc=src->getPlane(N);
		for (int i=0;i<=src->getDim();i++) hsrc[i]=-hsrc[i];

		key below=src->getBelow(N);
		key above=src->getAbove(N);

		src->getNode(N).below=above;
		src->getNode(N).above=below;

		run(src,below);
		run(src,above);
	}

}; 


//_________________________________________________________________________
//a bsp which dynamicall extrude a <src> bsp
//the extrusion operator simply embed planes in the new dimension
//and add two new planes at the beginning (xn<=From and xn>=To)
class bsp_extrude: public bsp
{
	bsp* src;

	virtual void more(key N,key More);

public:

	//constructor
	bsp_extrude(bsp* src,real From=0,real To=1);

}; 


//_________________________________________________________________________
// BSP dynamic geometric transformation
enum {TRANSLATE,SCALE,ROTATE,SHEAR,TRANSFORM};

class bsp_transform: public bsp
{
	//see enum for good codes
	int  code; 

	//the bsp to transform
	bsp* src;  

	//the transformation matrix
	real v[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)]; 

	virtual void more(key N,key More);

public:

	//constructor
	// Example:
	//
	// real vt[]={tx,ty,tz};bsp_transform(TRANSLATE,src,vt);
	// real vs[]={sx,sy,sz};bsp_transform(SCALE,src,vt);
	// real vr[]={1,2,PIGRECO/4};bsp_transform(ROTATE,src,vr); -> rotation for axe perpendicular to plane <x,y>
	// real vh[]={1,shx,shy,shz};bsp_transform(SHEAR,src,vh); -> shear <x> with the vector <shx,shy,shz>
	// real mt[]={matrix of transf, with omogeneous components in row-col 0}; bsp_transform(TRANSFORM,src,mat)

	bsp_transform(int code,bsp* src,real* v);
};


//_________________________________________________________________________
//dynamic boolean operations
class bsp_bool:public bsp
{
protected:

	typedef struct
	{
		//see the enum inside the class
		uint8    code; 

		//special counter to use different source-bsps where to "pick" planes
		key      num,cur,*m,*vmax;

		//source-bsp
		bsp  **bsps;

		//source nodes for source bsps
		key  *nodes;
	}
	info;

	memcache args;

	virtual void more(key N,key More);

	key	block		(int n);
	key	block		(const info& src);
	void	release	(key I);

	//hidden contructor
	bsp_bool(int dim,int code);

public:

	enum {BOOL_AND,BOOL_OR,BOOL_DIFF,BOOL_XOR};

	//constructor
	// Example
	//
	// bsp* args[]={b1,b2,b3,0};
	// key  vmax[]={1,1,1}; -> use one plane from b1, then one plane from b2, then one plane from b3 etc
	//                      -> if vmax={0,0,0} then use all the planes from b1 before b2's, all planes from b2 before b3's,...
	//                      -> *** default behavious *** is to interleave {1,1,1,....}
	//new bsp_bool(BOOL_OR,args,vmax)

	bsp_bool(int code,bsp** src,key* vmax=0);

	static bsp_bool* Union			(bsp** src,key* vmax=0);
	static bsp_bool* Intersection	(bsp** src,key* vmax=0);
	static bsp_bool* Difference	(bsp** src,key* vmax=0);
	static bsp_bool* Xor          (bsp** src,key* vmax=0);
};


//_________________________________________________________________________
// build a static bsp which represent the convex hull of points (using qhull library)
class bsp_hull:public bsp
{
public:

	//make the convex hull from all points 
	bsp_hull(int pointdim,int npoints,double* points,real tolerance=1e-6);
};


//_________________________________________________________________________
//dynamically computes the full lattice (ggraph) starting from a bsp representation
class bsp_solver: public bsp
{
protected:

	typedef struct
	{
		//the actual cell of the ggraph
		key Cell;
		//the source-node
		key Node;
		//the level of recursion
		int deep;
	} 
	info;
	
	memcache args;

	//source bsp to solve
	bsp*    src;
	//the ggraph I'm buildinng
	ggraph* g;
	//the tolerance for splitting operation
	real    tolerance;
	//stop the solving at <maxdeep>
	int     maxdeep;
	//the bounding box
	real    limit;
	//do simplification where possible. ex (cut,in,in)->(in)
	bool    simplify;

	virtual void more(key N,key More);

	static void split(ggraph* g,real* h,key Cell,key& cminus,key& cequal,key& cplus,real tolerance,int maxnumtry,key Id);

	void doExportSvg(FILE* file,key N,int width,int height,bool back_to_front,real* observer,real* vmat);

public:

	//constructor
	// src is the source-bsp to solve
	// tolerance is used when (robust) splitting a cell
	// at <maxdeep> level of recursion stops
	// limit is the bounding box limit [-limit,+limit,-limit,+limit...]
	// <simplify> allows the simplification where possible. ex (cut,in,in)->(in)

	bsp_solver(bsp* src,real tolerance=1e-4,int maxdeep=0,real limit=1e4,bool simplify=true);

	~bsp_solver();

	//simplify or not
	void setSimplify(bool simplify);

	//access the graph
	ggraph* getGraph();

	//reset all content
	virtual void clear();

	//reset the max-deep 
	void setMaxDeep(int value);

	//export function (with polygons for each bsp-node)
	void doExportSvg(const char* filename,int width,int height,bool back_to_front,real* observer,real* vmat);

};


//_________________________________________________________________________
// dynamically repair a bsp. "Repair" means
// to convert the bsp->lattice and remove weird cells
// (ie cells with baricenter too far)
class bsp_repair: public bsp_solver
{
protected:

	virtual void more(key N,key More);

	real expand,toofar;

public:

	// the algorithm for repair is the following. For each IN cell (of the lattice) (1) calculate
	// its centroid cx,cy,... (2) expand the centroid cx'=expand*cx,cy'=expand*cy,...
	// (3) if cx' or cy' or... >toofar then the IN cell -> OUT cell
	bsp_repair(bsp* src,real tolerance=1e-4,int maxdeep=0,real limit=1e4,real expand=0.1,real toofar=10);
};


//_________________________________________________________________________
//Builds a full-bsp starting from a boundary representation
//see code for details. Working in arbitrary dimension
class bsp_brep
{
protected:

	ggraph* g;
	bsp*    b;

	
	static void split(ggraph* g,real* h,key F,key& Fminus,key& Fplus,real tolerance=1e-6);

public:

	//constructor
	bsp_brep(int dim);

	//constructor
	bsp_brep(ggraph* g);

	//destructor
	~bsp_brep();

	//access the graph
	ggraph* getGraph();

	//access the full-bsp
	bsp* getBsp();

	//start the convertion
	void convert(bool bScaleToUnitBox=true);

	//for special convertion
	static void convert(bsp* b,key N,ggraph* g,key F,int nrec=0);
};

//_________________________________________________________________________
// INLINES

inline int  bsp::getDim() const	{return this->dim;}
inline key  bsp::getRoot()			{return root;}

inline void bsp::setMore(key N,key More)	{getNode(N).More=More;}

inline key   bsp::getBelow     (key N) {return getNode(N).below;}
inline key   bsp::getAbove     (key N) {return getNode(N).above;}
inline real* bsp::getPlane     (key N) {return (real*)db[getNode(N).h];}
inline key&  bsp::nodeInfo     (key N) {return getNode(N).Info;}

inline int bsp::getNumberOfIn()  const {return nin ;}
inline int bsp::getNumberOfOut() const {return nout;}
inline int bsp::getNumberOfCut() const {return ncut;}

inline bool bsp::isIn (key N)    {return !isCut(N) && getNode(N).type==BSP_IN;}
inline bool bsp::isOut(key N)		{return !isCut(N) && getNode(N).type==BSP_OUT;}
inline bool bsp::isCut(key N)		{return getNode(N).h!=0;}

inline key bsp::cut1(real a,real b              ,key below,key above)				{real h[2]={a,b    };return cutn(h,below,above);}
inline key bsp::cut2(real a,real b,real c       ,key below,key above)				{real h[3]={a,b,c  };return cutn(h,below,above);}
inline key bsp::cut3(real a,real b,real c,real d,key below,key above)				{real h[4]={a,b,c,d};return cutn(h,below,above);}

inline void bsp::setCut1 (key N,real a,real b       ,key below,key above)			{real h[2]={a,b    };setCutn(N,h,below,above);}
inline void bsp::setCut2 (key N,real a,real b,real c,key below,key above)			{real h[3]={a,b,c  };setCutn(N,h,below,above);}
inline void bsp::setCut3 (key N,real a,real b,real c,real d,key below,key above) {real h[4]={a,b,c,d};setCutn(N,h,below,above);}


inline bsp::node_t& bsp::getNode(key N)
{
	node_t* n=(node_t*)nodes[N];
	if (block_expand || !n->More) return *n;
	key Arg=n->More;
	n->More=0; //do not expand again!
	more(N,Arg);
	return *((node_t*)nodes[N]);
}

inline key bsp::in() 
{
	key N=nodes.block();
	node_t* n=(node_t*)nodes[N];
	memset(n,0,sizeof(node_t));
	n->type=BSP_IN;
	++nin;
	return N;
}

inline key bsp::out() 
{
	key N=nodes.block();
	node_t* n=(node_t*)nodes[N];
	memset(n,0,sizeof(node_t));
	n->type=BSP_OUT;
	++nout;
	return N;
}

inline key bsp::cutn(const real* h,key below,key above)
{
	key N=nodes.block();
	node_t* n=(node_t*)nodes[N];
	memset(n,0,sizeof(node_t));
	n->h=db.block();
	memcpy(db[n->h],h,db.size());
	n->below=below;
	n->above=above;
	++ncut;
	return N;
}

inline void bsp::setIn(key N)
{
	node_t& n=getNode(N);
	if (n.type==BSP_IN) return;
	if (n.h) {db.release(n.h);n.h=0;--ncut;} else {assert(n.type==BSP_OUT);--nout;}
	++nin;n.type=BSP_IN;
}

inline void bsp::setOut(key N)
{
	node_t& n=getNode(N);
	if (n.type==BSP_OUT) return;
	if (n.h) {db.release(n.h);n.h=0;--ncut;} else {assert(n.type==BSP_IN);--nin;}
	++nout;n.type=BSP_OUT;
}

inline void bsp::setCutn(key N,const real* h,key below,key above)
{
	node_t& n=getNode(N);

	if (!n.h)
	{
		++ncut;n.h=db.block();
		if (n.type==BSP_IN) --nin; else --nout;
	}

	memcpy(db[n.h],h,db.size());
	n.below=below;n.above=above;
}




#endif //_H__

