/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _GENERALDEF_H__
#define _GENERALDEF_H__


#ifdef WIN32
#pragma warning(disable:4786)

#define isnan(a) _isnan(a)
#define finite(a) _finite(a)
#endif

#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <string.h>

//integer type
typedef char							  int8;
typedef unsigned char   			 uint8;
typedef short           			 int16;
typedef unsigned short  			uint16;
typedef int              			 int32;
typedef unsigned int     			uint32;
//typedef __int64           			 int64;
//typedef unsigned __int64 			uint64;

//decimal type
typedef float							 dec32;
typedef double							 dec64;

//the default decimal type
typedef dec32                      real;

//a key is a number which can be safely converted to/from a pointer
//	   0 means a null pointer (no reference to other items)
//   -1 means an invalid reference (for particular purpouse)

typedef uint32 key; 

#define INVALID_KEY ((key)-1)

#ifndef min2
#define min2(a,b) (((a)<=(b))?(a):(b))
#endif

#ifndef max2
#define max2(a,b) (((a)>=(b))?(a):(b))
#endif

#ifndef min3
#define min3(a,b,c) min2(min2(a,b),min2(b,c))
#endif

#ifndef max3
#define max3(a,b,c) max2(max2(a,b),max2(b,c))
#endif

#ifndef min4
#define min4(a,b,c,d) min2(min3(a,b,c),d)
#endif

#ifndef max4
#define max4(a,b,c,d) max2(max3(a,b,c),d)
#endif

#ifndef Swap
#define Swap(_type,a,b) {_type __tmp__=(a);(a)=(b);(b)=__tmp__;}
#endif

#ifndef PIGRECO
#define PIGRECO ((real)3.1415926535897932)
#endif

#ifdef ENABLE_CHECKPOINT
#define checkpoint(a) if (!(a)) {fprintf(stderr,"FATAL error in file %s line %d\n",__FILE__,__LINE__);fflush(stderr);assert(a);exit(-1);}
#else
#define checkpoint(a)
#endif

#define MAX_GEOMETRIC_DIMENSION 16


inline char* fileExtension(char* filename)
{
	for (int i=strlen(filename)-1;i>=0;i--)
		if (filename[i]=='.') return &filename[i+1];

	return 0;
}

template<class T>
inline void normalize_plane(int pointdim,T* h)
{
  T acc=0;int k;
	switch (pointdim)
	{
	case 0:return;
	case 1:acc=(T)(h[1]*h[1])                           ;if ((acc=(T)sqrt(acc))!=1) {h[0]/=acc;h[1]/=acc;}                    break;
	case 2:acc=(T)(h[1]*h[1]+h[2]*h[2])                 ;if ((acc=(T)sqrt(acc))!=1) {h[0]/=acc;h[1]/=acc;h[2]/=acc;}          break;
	case 3:acc=(T)(h[1]*h[1]+h[2]*h[2]+h[3]*h[3])       ;if ((acc=(T)sqrt(acc))!=1) {h[0]/=acc;h[1]/=acc;h[2]/=acc;h[3]/=acc;}break;
	default:for (k=1;k<=pointdim;k++) acc+=h[k]*h[k];if ((acc=(T)sqrt(acc))!=1) {for (k=0;k<=pointdim;k++)  h[k]/=acc;}   break;
	}
	checkpoint(acc && !isnan(acc) && finite(acc)) ;
}


/*
   Calculate HSL from RGB
   Hue is in degrees
   Lightness is between 0 and 1
   Saturation is between 0 and 1
*/
inline void RGB2HSL(float& hue,float& saturation,float& lightness,float red,float green,float blue)
{
   double themin,themax,delta;

   themin = min3(red,green,blue);
   themax = max3(red,green,blue);
   delta = themax - themin;
   lightness = (float)((themin + themax) / 2);

   saturation = 0;
   if (lightness > 0 && lightness < 1)
      saturation = (float)(delta / (lightness < 0.5 ? (2*lightness) : (2-2*lightness)));

   hue = 0;
   if (delta > 0) {
      if (themax == red && themax != green)
         hue += (float)((green - blue) / delta);

      if (themax == green && themax != blue)
         hue += (float)((2 + (blue - red) / delta));

      if (themax == blue && themax != red)
         hue += (float)((4 + (red - green) / delta));

      hue *= 60;
   }
}

/*
   Calculate RGB from HSL, reverse of RGB2HSL()
   Hue is in degrees
   Lightness is between 0 and 1
   Saturation is between 0 and 1
*/
inline void HSL2RGB(float& red,float& green,float& blue,float hue,float saturation,float lightness)
{
	float _red,_green,_blue;
	float _Red,_Green,_Blue;

   while (hue < 0)
      hue += 360;

   while (hue > 360)
      hue -= 360;

   if (hue < 120) 
	{
      _red = (120 - hue) / 60.0f;
      _green = hue / 60.0f;
      _blue = 0;
   } else if (hue < 240) 
	{
      _red = 0;
      _green = (240 - hue) / 60.0f;
      _blue = (hue - 120) / 60.0f;
   } else 
	{
      _red = (hue - 240) / 60.0f;
      _green = 0;
      _blue = (360 - hue) / 60.0f;
   }

   _red   = min2(_red,1);
   _green = min2(_green,1);
   _blue  = min2(_blue,1);

   _Red   = 2 * saturation * _red   + (1 - saturation);
   _Green = 2 * saturation * _green + (1 - saturation);
   _Blue  = 2 * saturation * _blue  + (1 - saturation);

   if (lightness < 0.5) 
	{
      red   = lightness * _Red;
      green = lightness * _Green;
      blue  = lightness * _Blue;
   } else 
	{
      red = (1 - lightness) * _Red + 2 * lightness - 1;
      green = (1 - lightness) * _Green + 2 * lightness - 1;
      blue = (1 - lightness) * _Blue + 2 * lightness - 1;
   }
}



#endif //_GENERALDEF_H__

