/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _GGRAPH_H__
#define _GGRAPH_H__

#include <math.h>

#include "graph.h"
#include "mat.h"




class ggraph: public graph 
{

	void triangulate_inner(key N);

	ggraph* join(key N,real coord);

public:

	//point dimension (for very particular operations)
	int spacedim,pointdim;

	//transformation matrix (eventually with embedding if spacedim!=pointdim) (for very particular operations)
	real *vmat,*hmat;

	// geometric db for hyperplanes (pointdim-1)-cells,vertices 0-cells, and bounding-balls pointdim-cells
	memcache  db;  
	
	//default contructor
	ggraph(int _pointdim=0);

	//copy constructor
	ggraph(const ggraph& src);

	//destructor
	~ggraph();

	//clone
	ggraph* clone() const;

	//assignment
	ggraph& operator=(const ggraph& src);

	//clear memory
	virtual void clear();

	//print
	virtual void print(key N,char* space="",FILE* file=stdout);
	virtual void print(char* space="",FILE* file=stdout);
	virtual void printStatistics(FILE* file=stdout);

	//get cells at a certain level (speed up for pointdim-cells)
	virtual key   findCells(int level,key Cell,key* dst);
	virtual void  findCells(int level,key Cell,key* nnav,key** nav);

	//remove a node
	virtual void remNode(key Node,bool recursive=true);

	//access the geom of a cell
	inline real* getGeom(key N);

	//copy the geometry from src-cell to dst-cell
	real* setGeom(key D,ggraph* scom=0,key S=0);

	//release
	void releaseGeom(key Node);

	//read from file (ply,obj,3d-files...)
	static ggraph* read(char* filename);

	////////// GEOM UTILS /////////////////

	//spacedim
	void setSpacedim(int spacedim,const real* vmat=0,const real* hmat=0);

	//cuboid
	static ggraph* cuboid(int pointdim);

	//simplex
	static ggraph* simplex(int pointdim);

	//build a grid
	static ggraph* grid(int pointdim,uint8* v);

	//find the overall bounding box
	void getBoundingBox(key npoints,key* points,real* m,real* M,real& limit,real& maxdelta,bool bStartup=true,bool bApplySpacedim=false) ;
	void getBoundingBox(real* m,real* M,real& maxdelta,real& limit);

	//scale to unit box
	void scaleToUnitBox();
	void scaleToUnitBox(tmat<real>* vmat,tmat<real>* hmat);


	//centroid of a pointdim-cell. The bounding ball has the radius (^2) for [0]
	//and the point location for [i] (i>=1), the centroid is cached (->computed only one time)
	real* boundingBall(key C);

	//slow version which returns the hyperplane of face <F> which leave the cell in the below space
	void robustFaceOutside(real* hsplit,key C,key F);

	//find the plane of all points "referenced" by a cell (all points if Cell=0),
	//Operate in pointdim (which can be different from this->pointdim)
	void findPlane(real* h,key npoints,key* points,int pointdim,const real module_too_small=1e-6);

	//fix boundary (pointdim-1)-cells orientation
	void fixPlaneOrientation();

	//transformation
	ggraph& shear     (int ref,const real* v);
	ggraph& translate (real* vt);
	ggraph& scale     (real* vs);
	ggraph& rotate    (int ax1,int ax2,real alpha);
	ggraph& transform (const real* pmat,const real *hmat);

	//embed
	ggraph* embed(real From,real To);

	//permutation of coordinate
	ggraph& permutate(int* perm);

	//embed
	ggraph& embed(int pointdim);

	//triangulate
	ggraph& triangulate ();

	//check a cell
	virtual void check(key N);
	virtual void check() 
	{
		if (vmat)
		{
			checkpoint(tmat<real>::isInvert(spacedim+1,vmat,hmat));
		}
		graph::check();
	}

	//add something
	key addVertex	(const real* v);
	key addEdge		(key N0,key N1);
	key addEdge		(real* v0,real* v1);
	key addFace		(key n,const key* N);
	key addTet		(key N0,key N1,key N2,key N3);

	//make a ggraph representing the convex hull of points in Node (all the complex if Node=0), 
	//eventually returns a matrix for the embedding
	ggraph* mkpol(key npoints,key* points,bool bUseSymbolic=true,real tolerance=1e-6) ;
	static ggraph* mkpol(int pointdim,int npoints,const real* points,bool bUseSymbolic=true,real tolerance=1e-6);


	//PLaSM power
	static ggraph* power(
		const ggraph* _g1,const int sdim1,const real* _g1vmat,const real* _g1hmat,
		const ggraph* _g2,const int sdim2,const real* _g2vmat,const real* _g2hmat,
		real tolerance,bool use_internal_faces,bool use_partition_planes);

	//PLaSM skeleton
	ggraph** skeleton(int level,int& num);

	//robust split! the most important algorithm
	enum 
	{
		OK,FLAT,IMPOSSIBLE
	};
	int split(key C,real* h,real tolerance,key& Cminus,key& Cplus,key& Cequal,key Id=0);

	//friend class XgeContour2;
	//friend class XgeContour3;

};


inline ggraph::ggraph(int _pointdim)
	:pointdim(_pointdim),db(sizeof(real)*(_pointdim+1)),vmat(0),hmat(0)
{
	setSpacedim(_pointdim,0,0);
}

inline ggraph::ggraph(const ggraph& src)	
	:pointdim(src.pointdim),db(src.db),graph(src),vmat(0),hmat(0)
{
	setSpacedim(src.spacedim,src.vmat,src.hmat);
}

inline ggraph::~ggraph()
{
	if (vmat) delete [] vmat;
	if (hmat) delete [] hmat;
}

inline ggraph* ggraph::clone() const
	{return new ggraph(*this);}

inline ggraph& ggraph::operator=(const ggraph& src)
{
	graph::operator =(src);
	this->pointdim=src.pointdim;
	this->db=src.db; 
	setSpacedim(src.spacedim,src.vmat,src.hmat);
	return *this;
}

inline void ggraph::setSpacedim(int _spacedim,const real* _vmat,const real* _hmat)
{
	checkpoint(_spacedim>=pointdim);
	this->spacedim=_spacedim;

	if (vmat) delete [] vmat;vmat=0;
	if (hmat) delete [] hmat;hmat=0;

	if (_vmat)
	{
		this->vmat=new real[(spacedim+1)*(spacedim+1)];
		memcpy(this->vmat,_vmat,sizeof(real)*(spacedim+1)*(spacedim+1));
	}
	if (_hmat)
	{
		this->hmat=new real[(spacedim+1)*(spacedim+1)];
		memcpy(this->hmat,_hmat,sizeof(real)*(spacedim+1)*(spacedim+1));
	}
}

inline void ggraph::clear()
{
	graph::clear();
	this->db.clear();
	spacedim=pointdim;
	if (vmat) delete [] vmat;vmat=0;
	if (hmat) delete [] hmat;hmat=0;
}

inline real* ggraph::getGeom(key N)
{
	assert(nodeData(N));
	return (real*)db[nodeData(N)];
}

//add a vertex
inline key ggraph::addVertex(const real* v)
{
	key V=addNode(0);
	nodeData(V)=db.block();
	real* g=(real*)db[nodeData(V)];
	memcpy(g,v,db.size());
	return V;
}

//copy the geometry from src-cell to dst-cell
inline real* ggraph::setGeom(key D,ggraph* scom,key S)
{
	if (!nodeData(D)) 
		nodeData(D)=db.block();
	real* g=getGeom(D);
	memset(g,0,db.size());
	g[0]=1;
	if (S) 
		memcpy(g,scom->getGeom(S),scom->db.size());
	return g;
}

//release the geometry of a cells
inline void ggraph::releaseGeom(key Node)
{
	key& Info=nodeData(Node);
	if (Info) db.release(Info);
	Info=0;
}


//fast bounding box of points
inline void ggraph::getBoundingBox(real* m,real* M,real& maxdelta,real& limit)
{
	ggraph::list_iterator it;

	maxdelta=0;
	limit=0;

	//checkpoint(!vmat && !hmat);

	switch(pointdim)
	{
	case 0:
		{
			m[0]=(real)+1e18;
			M[0]=(real)-1e18;
			for (it=each(0);!it.end();it++)
			{
				real* vg=getGeom(*it);
				m[0]=min2(m[0],vg[0]);
				M[0]=max2(M[0],vg[0]);
			}
			maxdelta=M[0]-m[0];
			limit=max3(limit,m[0],M[0]);
			break;
		}
	case 1:
		{
			m[0]=m[1]=(real)+1e18;
			M[0]=M[1]=(real)-1e18;
			for (it=each(0);!it.end();it++)
			{
				real* vg=getGeom(*it);
				m[0]=min2(m[0],vg[0]);m[1]=min2(m[1],vg[1]);
				M[0]=max2(M[0],vg[0]);M[1]=max2(M[1],vg[1]);
			}

			maxdelta=M[0]-m[0];
			maxdelta=max2(maxdelta,M[1]-m[1]);
			limit=max3(limit,m[0],M[0]);
			limit=max3(limit,m[1],M[1]);
			break;
		}
	case 2:
		{
			m[0]=m[1]=m[2]=(real)+1e18;
			M[0]=M[1]=M[2]=(real)-1e18;
			for (it=each(0);!it.end();it++)
			{
				real* vg=getGeom(*it);
				m[0]=min2(m[0],vg[0]);m[1]=min2(m[1],vg[1]);m[2]=min2(m[2],vg[2]);
				M[0]=max2(M[0],vg[0]);M[1]=max2(M[1],vg[1]);M[2]=max2(M[2],vg[2]);
			}

			maxdelta=M[0]-m[0];
			maxdelta=max2(maxdelta,M[1]-m[1]);
			maxdelta=max2(maxdelta,M[2]-m[2]);
			limit=max3(limit,m[0],M[0]);
			limit=max3(limit,m[1],M[1]);
			limit=max3(limit,m[2],M[2]);
			break;
		}
	case 3:
		{
			m[0]=m[1]=m[2]=m[3]=(real)+1e18;
			M[0]=M[1]=M[2]=M[3]=(real)-1e18;
			for (it=each(0);!it.end();it++)
			{
				real* vg=getGeom(*it);
				m[0]=min2(m[0],vg[0]);m[1]=min2(m[1],vg[1]);m[2]=min2(m[2],vg[2]);m[3]=min2(m[3],vg[3]);
				M[0]=max2(M[0],vg[0]);M[1]=max2(M[1],vg[1]);M[2]=max2(M[2],vg[2]);M[3]=max2(M[3],vg[3]);
			}

			maxdelta=M[0]-m[0];
			maxdelta=max2(maxdelta,M[1]-m[1]);
			maxdelta=max2(maxdelta,M[2]-m[2]);
			maxdelta=max2(maxdelta,M[3]-m[3]);
			limit=max3(limit,m[0],M[0]);
			limit=max3(limit,m[1],M[1]);
			limit=max3(limit,m[2],M[2]);
			limit=max3(limit,m[3],M[3]);
			break;
		}
	default:
		{
			int i;
			for (i=0;i<=pointdim;i++)
				{m[i]=(real)+1e18;M[i]=(real)-1e18;}
		
			for (it=each(0);!it.end();it++)
			{
				real* vg=getGeom(*it);
				for (int u=0;u<=pointdim;u++)
				{
					m[u]=min2(m[u],vg[u]);
					M[u]=max2(M[u],vg[u]);
				}
			}

			for (i=0;i<=pointdim;i++)
			{
				maxdelta=max2(maxdelta,M[i]-m[i]);
				limit=max3(limit,m[i],M[i]);
			}

			break;
		}
	}

}



#endif //_GGRAPH_H__

