/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#include "generaldef.h"

extern "C"
{
#include <qset.h>
#include <qhull.h>
}


#include <map>
#include <set>

#include "bsp.h"
#include "ggraph.h"
#include "mat.h"


/////////////////////////////////////////////////////////////////////////////////////////

/*
S_V_D decomposition to find null space
REF1: find Null Space of a matrix. 

ROTATE points so that the vector (0,h[1],h[2],..h[pointdim]) goes to (0,0,...1)
REF2: change of basis of matrix, transform vector u to v, use 2 reflections in a plane.

write the matrix:

	M1 = I - 2*u*u' 
	M2 = I - 2*w*w' 
	
where:

	w = (u+v)/ Norm(u+v) (bisector of angle in the plane between u and v), 
	I is the n x n identity matrix,
	u_1 and w are column vectors 
	u_1' and w' are row vectors, 

so u_1*u_1' for example is the outer product that creates and n x n  matrix. 
The way it works is that matrix M1 reflects u into -u and matrix  M2 reflects -u into v 
So you multiply R=M2*M1 to get the rotation matrix.  
The determinant  of M1 and M2 is -1 (because they are reflections) and so the det of 
the product R is 1, as expected for a rotation
*/

/////////////////////////////////////////////////////////////////////////////////////////

ggraph* ggraph::mkpol(key npoints,key* points,bool bUseSymbolic,real tolerance) 
{
#ifdef WIN32
	static FILE* devnull=fopen("nul","w");
#else
	static FILE* devnull=fopen("/dev/null","w");
#endif


	double M1[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
	double M2[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
	real   h[MAX_GEOMETRIC_DIMENSION+1],h0;
	double v[MAX_GEOMETRIC_DIMENSION+1],w[MAX_GEOMETRIC_DIMENSION+1];
	real   m[MAX_GEOMETRIC_DIMENSION+1],M[MAX_GEOMETRIC_DIMENSION+1],limit,maxsize;
	double acc;

	real* mat=0;
	ggraph* g=0;

	if (!npoints)
		return 0;

	//copy the geometric database since it can be changed by embedding transformations
	memcache* db_clone=0; 

	dec64* tmp_points=new dec64[npoints*this->pointdim];

	for (int PDim=this->pointdim;PDim>=0 && !g;PDim--)
	{
		//enough points
		if (npoints>=(PDim+1))
		{
			getBoundingBox(npoints,points,m,M,limit,maxsize); 

			//base case
			if (!PDim)
			{
				g=new ggraph(0);
				real v[]={1};
				g->addVertex(v);
				break;
			}

			//base case
			if (PDim==1)
			{
				if (m[1]!=M[1])
				{
					g=new ggraph(1);
					real v1[]={1,m[1]};key V0=g->addVertex(v1);
					real v2[]={1,M[1]};key V1=g->addVertex(v2);
					key E=g->addNode(1);
					g->addArch(V0,E);g->addArch(E,V0);
					g->addArch(V1,E);g->addArch(E,V1);
					break;
				}
				continue;
			}

			dec64* ptr=tmp_points;

			for (int it=0;it<npoints;it++) 
			{
				real* vg=getGeom(points[it]);
				for (int u=1;u<=PDim;u++) *ptr++=vg[u];
			}

			//use bsp and resolve the system
			if (!bUseSymbolic)
			{
				bsp_hull b(PDim,npoints,tmp_points);
				
				if (!b.isOut(b.getRoot()))
				{		
					bsp_solver s(&b,tolerance,0,limit*1.5); 
					s.visit();
					
					if (!s.getGraph()->each(PDim).end()) 
					{
						g=s.getGraph()->clone(); 
						break;
					}
				}
			}
			//use symbolic ggraph building
			else
			{	
				//OPTIONS: i==vertices incidence Pp=does not print warnings, n=print hyperplane normals with offsets, En max roundoff error, Qs search all points for initial simplex
				char qhull_opts[256];sprintf(qhull_opts,"qhull i Pp E%e Qs",tolerance);
				int error_qhull=qh_new_qhull(PDim,npoints,tmp_points,(boolT)false,qhull_opts, NULL, devnull);

				if (!error_qhull)
				{
					g=new ggraph(PDim); 

					key A,V,B,C;
					std::map<key,key> mapv;
					std::map<std::set<key>,key> mapf;
					std::map<key,key>::iterator mapv_it;
					ggraph::list_iterator Ait;
					ggraph::iterator Vit,Bit,It1,It2,It;
					std::set<key> v_of_face;
					
					key Up=g->addNode(PDim);
					
					facetT* facet;
					FORALLfacets 
					{
						key F=g->addNode(PDim-1);
						real* h=g->setGeom(F);
						h[0]=(real)facet->offset;

						for (int k=1;k<=PDim;k++) 
							h[k]=(real)facet->normal[k-1];
						
						g->addArch(F,Up);
						
						vertexT *vertex, **vertexp;
						FOREACHvertex_(facet->vertices) 
						{
							key id=(key)vertex->id,Id;
							mapv_it=mapv.find(id);
							
							if (mapv_it==mapv.end())
							{
								real coords[MAX_GEOMETRIC_DIMENSION+1];
								coords[0]=1;
								for (int u=1;u<=PDim;u++) 
									coords[u]=vertex->point[u-1];
								
								mapv[id]=Id=(g->addVertex(coords));
								g->addArch(Up,Id);
							}
							else
								Id=mapv_it->second;
							
							g->addArch(Id,F);
						}
					}

					//create other links (by symbolic intersection of faces)
					for (int lvl=PDim-1;lvl>=2;lvl--)
					{
						mapf.clear();
						
						for (A=*(Ait=g->each(lvl));!Ait.end();Ait++,A=*Ait) 
						{	 		
							for (V=*(Vit=g->goDw(A))  ;!Vit.end();Vit++,V=*Vit) 
							{				
								for (B=*(Bit=g->goUp(V))  ;!Bit.end();Bit++,B=*Bit)
								{					
									if (B<=A) 
										continue;
									
									v_of_face.clear();
									
									for (It1=g->goDw(A);!It1.end();It1++)
									{
										for (It2=g->goUp(*It1);!It2.end();It2++)
										{
											if (*It2==B) v_of_face.insert(*It1);
										}
									}
									
									if (v_of_face.size()>=lvl)
									{	
										//if the face already exists...
										if (mapf.find(v_of_face)==mapf.end())
										{
											mapf[v_of_face]=(C=g->addNode(lvl-1));
											for (std::set<key>::iterator u=v_of_face.begin();u!=v_of_face.end();u++) g->addArch(*u,C,DIRECTION_DOWN);
										}
										else
										{
											C=mapf[v_of_face];
										}
										
										if (!g->findArch(C,A,DIRECTION_UP)) g->addArch(C,A,DIRECTION_UP);
										if (!g->findArch(C,B,DIRECTION_UP)) g->addArch(C,B,DIRECTION_UP);
									}
								}
							}
						}
						
						for (Ait=g->each(lvl);!Ait.end();Ait++)
						{
							while (g->getNDw(*Ait)) 
								g->remArch(g->getFirstDwArch(*Ait));
						}
						
						for (Ait=g->each(lvl-1);!Ait.end();Ait++)
						{
							for (It=g->goUp(*Ait);!It.end();It++) 
								g->addArchDirection(It.getArch(),DIRECTION_DOWN);
							
							for (It=g->goDw(*Ait);!It.end();It++) 
								g->addArchDirection(It.getArch(),DIRECTION_UP);
						}
					}
					
					//order 2-faces
					for (ggraph::list_iterator jt=g->each(2);!jt.end();jt++)
					{
						bool bOrdered=g->orderFace(*jt);
						checkpoint(bOrdered);
					}
				}

				qh_freeqhull(qh_ALL);
			}
		}

		//failed in PDim
		if (!g) 
		{
			if (!db_clone)
				db_clone=new memcache(this->db);

			int r,c,i,j;

			findPlane(h,npoints,points,PDim);
			
			//prepare for the translation...
			h0=h[0];
			h[0]=0;
			
			//find the rotation with one orthogonal axe a (and -a) for which the rotation angle is less
			//this is usefull to make sure the bisector vector w doesn't get null
			
			int less_angle=0,nzero=0;
			
			for (i=1;i<=PDim;i++)
			{
				if (!h[i]) ++nzero;
				if (!less_angle || fabs(h[i])>fabs(h[less_angle])) less_angle=i;
			}

			checkpoint(less_angle>=1 && h[less_angle]!=0);
			
			//the target vector
			memset(v,0,sizeof(v));
			v[less_angle]=(h[less_angle]>0)?(+1):(-1);
			
			//find bisector
			for (acc=0,i=0;i<=PDim;i++) 
			{
				w[i]=h[i]+v[i];
				acc+=w[i]*w[i];
			}

			for (acc=sqrt(acc),i=0;i<=PDim;i++) 
				w[i]/=acc;
			
			for (r=0;r<=PDim;r++)
			{
				for (c=0;c<=PDim;c++)
				{
					M1[r*(PDim+1)+c]=((r==c)?1:0)-2*h[r]*h[c];
					M2[r*(PDim+1)+c]=((r==c)?1:0)-2*w[r]*w[c];
				}
			}
			
			//rotation
			tmat<double>::product(PDim+1,M1,M2,M1); 
			
			//apply the translation (first matrix to be applied)
			tmat<double>::identity(PDim+1,M2);
			real* point0=getGeom(points[0]);
			for (i=1;i<=PDim;i++) M2[i*(PDim+1)]=-point0[i];
			tmat<double>::product(PDim+1,M1,M1,M2); 
			
			//needed a permutation of coordinates so the points with the <less_angle> coordinates (==0) goes in the plane xn=0 (n:=PDim)
			tmat<double>::swap_rows(PDim+1,M1,less_angle,PDim);

			//transform points and lower dimension		
			for (int it=0;it<npoints;it++)
			{
				real* vg=getGeom(points[it]);
				productv<real,double>(PDim+1,vg,M1,vg);

				//make sure the last coordinates gets cancelled (== points are in plane xn=0)
				checkpoint(fabs(vg[PDim])<1e-4 && vg[0]); 
				
				if (vg[0]!=1)
				{
					for (j=1;j<=PDim;j++) vg[j]/=vg[0];
					vg[0]=1;
				}
			}
			
			int tot=(PDim+1)*(PDim+1);
			real* local_mat=new real[tot];
			for (i=0;i<tot;i++) local_mat[i]=M1[i];
			tmat<real>::embed(this->pointdim+1,PDim+1,local_mat);

			if (!mat)
				{mat=local_mat;}
			else
				{tmat<real>::product(this->pointdim+1,mat,local_mat,mat);delete [] local_mat;}
		}
	}

	checkpoint(g);
	delete [] tmp_points;

	if (mat)
	{
		real*& Hmat=mat;
		real Vmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
		tmat<real>::invert(this->pointdim+1,Vmat,Hmat);
		g->setSpacedim(pointdim,Vmat,Hmat);
		delete [] mat;
	}

	if (db_clone)
	{
		this->db=*db_clone;
		delete db_clone;
	}

#ifdef _DEBUG
	g->check();
#endif


	
	return g;
}

/////////////////////////////////////////////////////////////////////////////////////////
ggraph* ggraph::mkpol(int pointdim,int npoints,const real* points,bool bUseSymbolic,real tolerance)
{
	if (!npoints) 
		return 0;

	real Tmp[MAX_GEOMETRIC_DIMENSION+1];
	Tmp[0]=1;
	
	ggraph gpoints(pointdim);
	const real* p=points;

	key Id[MAX_CELLS_PER_LEVEL];
	key *pId=Id;

	for (int i=0;i<npoints;i++,p+=pointdim)
	{
		memcpy(&Tmp[1],p,sizeof(real)*pointdim);
		*pId++=gpoints.addVertex(Tmp);
	}

	return gpoints.mkpol(npoints,Id,bUseSymbolic,tolerance);
}

