/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#include "generaldef.h"


#include "bsp.h"
#include "ggraph.h"
#include "mat.h"
#include "gbool.h"

//todo! not thread safe
static real g1vmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
static real g1hmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
static real g2vmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
static real g2hmat[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
static real VmatT [(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
static real HmatT [(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];


ggraph* ggraph::power(const ggraph* _g1,const int sdim1,const real* _g1vmat,const real* _g1hmat,
							 const ggraph* _g2,const int sdim2,const real* _g2vmat,const real* _g2hmat,
							 real tolerance,bool use_internal_faces,bool use_partition_planes)
{

	//note : i Ignore transf matrices of _g1 and _g2


	const int pdim1=_g1->pointdim;
	const int pdim2=_g2->pointdim;
	const int sdimT=sdim1+sdim2;
	const int pdimT=pdim1+pdim2;

	const int matsize=(sdimT+1)*(sdimT+1);
	
	real m[MAX_GEOMETRIC_DIMENSION+1],M[MAX_GEOMETRIC_DIMENSION+1];
	real t[MAX_GEOMETRIC_DIMENSION+1],s[MAX_GEOMETRIC_DIMENSION+1],maxdelta,limit;

	tmat<real>::identity(sdim1+1,g1vmat);tmat<real>::identity(sdim1+1,g1hmat);
	tmat<real>::identity(sdim2+1,g2vmat);tmat<real>::identity(sdim2+1,g2hmat);
	tmat<real>::identity(sdimT+1,VmatT );tmat<real>::identity(sdimT+1,HmatT );

	int i,u,r;

	//first of all apply the power to full dimension (avoid the clone of matrices)
	ggraph* g1=_g1->clone();
	ggraph* g2=_g2->clone();

	//make the power more stable by transforming arguments to unit-box
	if (g1->pointdim)
	{
		g1->getBoundingBox(m,M,maxdelta,limit);

		for (i=0;i<=pdim1;i++) 
		{
			t[i]=(real)(-0.5*(m[i]+M[i]));
			s[i]=(real)(maxdelta?(+2.0f/maxdelta):1);
		}
		g1->translate(t+1);g1->scale(s+1);

		for (i=1;i<=pdim1;i++) 
		{
			g1hmat[i*(sdim1+1)]=+t[i]*s[i];g1hmat[i*(sdim1+1)+i]=1*s[i];
			g1vmat[i*(sdim1+1)]=-t[i]     ;g1vmat[i*(sdim1+1)+i]=1/s[i];
		}
	}

	//make the power more stable by transforming arguments to unit-box
	if (g2->pointdim)
	{
		g2->getBoundingBox(m,M,maxdelta,limit);

		for (i=0;i<=pdim2;i++) 
		{
			t[i]=(real)(-0.5*(m[i]+M[i]));
			s[i]=(real)(maxdelta?(+2.0f/maxdelta):1);
		}
		g2->translate(t+1);g2->scale(s+1);

		for (i=1;i<=pdim2;i++) 
		{
			g2hmat[i*(sdim2+1)]=+t[i]*s[i];g2hmat[i*(sdim2+1)+i]=1*s[i];
			g2vmat[i*(sdim2+1)]=-t[i]     ;g2vmat[i*(sdim2+1)+i]=1/s[i];
		}
	}


	ggraph* gT=0;
	
	//very simple case <g> * point_in_origin_0_dim == <g>....
	if (!g1->pointdim || !g2->pointdim)
	{
		gT=g1->pointdim?g1->clone():g2->clone();
	}
	
	//very simple case, edge(-1,+1) * edge(-1,+1) -> cuboid
	else if (g1->pointdim==1 && g1->getNCells(1)==1 && g2->pointdim==1 && g2->getNCells(1)==1)
	{
		gT=cuboid(2);
		for (int j=0;j<2;j++) {s[j]=2;t[j]=-1;}
		gT->scale(s);gT->translate(t);
	}
	//otherwise...
	else
	{
		const real epsilon=0;

		for (u=0;u<pdim2;u++) 
		{
			ggraph* tmp=g1;
			g1=g1->embed(-1-epsilon,+1+epsilon);
			delete tmp;
		}
		
		for (u=0;u<pdim1;u++) 
		{
			ggraph* tmp=g2;
			g2=g2->embed(-1-epsilon,+1+epsilon);
			delete tmp;
		}

		int permutation2[MAX_GEOMETRIC_DIMENSION+1];
		permutation2[0]=0;
		for (i=1;i<=pdim1;i++)	permutation2[i]=pdim2+i; 
		for (i=1;i<=pdim2;i++)	permutation2[i+pdim1]=i; 
		g2->permutate(permutation2);
		
		GBool* gbool=new GBool(pdimT);
		gbool->useInternalFaces   (use_internal_faces );
		gbool->usePartitionPlanes (use_partition_planes);
		key And=gbool->BoolOp(bsp_bool::BOOL_AND);
		gbool->addArg(gbool->add(g1),And);
		gbool->addArg(gbool->add(g2),And);
		gbool->setRoot(And);
		gbool->visit();
		bsp_solver solver(gbool);
		solver.setSimplify(false);
		solver.visit();
		gT=solver.getGraph()->clone();

		delete gbool;
		gbool=0;
	}

	delete g1;g1=0;
	delete g2;g2=0;

	if (_g1vmat)
	{
		tmat<real>::product(sdim1+1,g1vmat,_g1vmat,g1vmat);
		tmat<real>::product(sdim1+1,g1hmat,g1hmat,_g1hmat);
	}

	for (r=1;r<=sdim1;r++)
	{
		for (int c=0;c<=sdim1;c++)
		{
			VmatT[r*(sdimT+1)+c]=g1vmat[r*(sdim1+1)+c];
			HmatT[r*(sdimT+1)+c]=g1hmat[r*(sdim1+1)+c];
		}
	}
	if (_g2vmat)
	{
		tmat<real>::product(sdim2+1,g2vmat,_g2vmat,g2vmat);
		tmat<real>::product(sdim2+1,g2hmat,g2hmat,_g2hmat);
	}

	for (r=sdim1+1;r<=sdimT;r++)
	{
		for (int c=sdim1+1;c<=sdimT;c++)
		{
			VmatT[r*(sdimT+1)+c]=g2vmat[(r-sdim1)*(sdim2+1)+(c-sdim1)];
			HmatT[r*(sdimT+1)+c]=g2hmat[(r-sdim1)*(sdim2+1)+(c-sdim1)];
		}
		
		VmatT[r*(sdimT+1)]=g2vmat[(r-sdim1)*(sdim2+1)];
		HmatT[r*(sdimT+1)]=g2hmat[(r-sdim1)*(sdim2+1)];
	}

	//permutation
	int permutation[MAX_GEOMETRIC_DIMENSION+1];
	int c=0;
	permutation[c++]=0;

	for (i=1      ;i<=pdim1;i++) permutation[c++]=i              ; //the full space of g1
	for (i=1      ;i<=pdim2;i++) permutation[c++]=sdim1+i; //the full space of g2
	for (i=pdim1+1;i<=sdim1;i++) permutation[c++]=i              ; //the embedding space of g1
	for (i=pdim2+1;i<=sdim2;i++) permutation[c++]=sdim1+i; //the embedding space of g2
	
	tmat<real>::swap_cols(sdimT+1,VmatT,permutation);
	tmat<real>::swap_rows(sdimT+1,HmatT,permutation);

	gT->setSpacedim(sdimT,VmatT,HmatT);

#ifdef _DEBUG
	gT->check();
#endif

	return gT;
}
