/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _GRAPH_H__
#define _GRAPH_H__

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "generaldef.h"
#include "memcache.h"

//directed or undirected graph
enum Direction 
{
	DIRECTION_DOWN       =1,
	DIRECTION_UP         =2,
	DIRECTION_UP_AND_DOWN=3
};


#define MAX_CELLS_PER_LEVEL 65536

//too big. cause stack overflow
//#define MAX_CELLS_PER_LEVEL 524288




//==============================================
//CLASS to store graph
//==============================================

class graph
{
protected:

	static int MainNumGGraph;


	//internal vars

	//list (0,1,n,...n-cells + auxiliary lists)
	//each cell can belong to only one list (I use cell.Prev and cell.Next)
	struct
	{
		key v[256];
		key n[256];
	}
	nodelists;

	memcache arches;
	memcache nodes ;

public:

	int refcount;
	int NumGGraph;

	typedef struct 
	{	
		key   N0,N1;
		key   N0_prev,N0_next;
		key   N1_prev,N1_next;
		key   Info; //reference information
	}
	arch_t;

	typedef struct 
	{
		uint8  Level,Sign;
		key    Dw,Up;
		uint16 Ndw,Nup;
		key    Prev,Next;
		key    Id;
		key    Info; //reference information
		key	 Tmp;  //for dynamic information
	}
	node_t; 


	arch_t* getArch(key A);
	node_t* getNode(key N);	

	class iterator;
	class list_iterator;

	
	//++++++++++++++++++++++++++++++++++++
	//iterator for a list
	//++++++++++++++++++++++++++++++++++++
	friend class list_iterator;

	class list_iterator
	{
		friend class graph;

	protected:
		key    Prev,Cursor;
		graph* g;

	public:
		list_iterator            ();
		list_iterator            (const list_iterator& src);
		bool           end        ();
		key            operator*  ();
		key            getPrev    ();
		list_iterator& operator++ (int);
		list_iterator  next       ();
	};

	//++++++++++++++++++++++++++++++++++++
	//iterator for arches connected to a node
	//++++++++++++++++++++++++++++++++++++
	friend class iterator;

	class iterator
	{
		friend class graph;

	protected:
		
		Direction dir;
		uint16 cont,max;
		key    A;
		graph* g;
	
	public:
		iterator            ();
		iterator            (const iterator& src);
		bool      end         ();
		key       getNode     ();
		key       getArch     ();
		key       operator*   ();
		iterator& operator++  (int);
		iterator   next       ();
	};


public:

	//constructor
	graph();

	//copy constructor
	graph(const graph& src);

	//fast copy
	graph* clone() const;

	//destructor
	~graph();

	//clear the graph
	virtual void clear();

	graph& operator=(const graph& src);

	//find all sub-cells of <Cells> or up-cells of <Cells>. If Cell==0 then returns all the
	//cells at a certain level (better use list_iterator). THe nav[level]/nnav[level] will be filled.
	virtual key  findCells(int level,key Cell,key*  dst);
	virtual void findCells(int level,key Cell,key* nnav,key** nav);

	key* getNavigation(key**& nav);

	//add a node
	key addNode(uint8 Level);

	//remove a node (NOTE: remove before all arches)
	virtual void remNode(key Node);

	//access internal information of node 
	key& nodeData(key Node);

	//access internal information of arch 
	key& archData(key Arch);

	//access internal information of node 
	key& nodeTmp(key Node);

	//add/get/remove an arch to the graph (insert before arch A0 or A1)
	key addArch(key N0,key N1,Direction dir=DIRECTION_UP_AND_DOWN,key A0=0,key A1=0);

	//special function to add a direction after the first addArch (insert before arch A0 or A1)
	void addArchDirection(key A,Direction dir,key A0=0,key A1=0);

	//find the arch <N0,N1>. Return 0 iif not found
	key findArch(key N0,key N1,Direction dir=DIRECTION_UP_AND_DOWN);

	//remove an arch 
	void remArch(key Arch,Direction dir=DIRECTION_UP_AND_DOWN);

	//remove an arch from <N0,N1>
	void remArch(key N0,key N1,Direction dir=DIRECTION_UP_AND_DOWN);

	//get start/end point of an arch
	key getN0(key Arch);
	key getN1(key Arch);

	//find a common node between N and N
	key findCommonNode(key N,key M,Direction dir=DIRECTION_UP_AND_DOWN);

	key findCommonNodes(key N,key M,key* v,Direction dir=DIRECTION_UP_AND_DOWN);

	//get number of down/up link of a node 
	uint16 getNDw(key Node);
	uint16 getNUp(key Node);

	//access arches and nodes
	key getFirstDwArch (key N);
	key getFirstUpArch (key N);
	key getLastDwArch  (key N);
	key getLastUpArch  (key N);

	key getNextDwArch  (key A);
	key getPrevDwArch  (key A);
	key getNextUpArch  (key A);
	key getPrevUpArch  (key A);

	key getFirstDwNode (key N);
	key getFirstUpNode (key N);
	key getLastDwNode  (key N);
	key getLastUpNode  (key N);

	key getNextUpBrother(key A);
	key getNextDwBrother(key A);
	key getPrevUpBrother(key A);
	key getPrevDwBrother(key A);

	key getNumNode();
	key getNumArch();

	key getMaxNode();
	key getMaxArch();

	const uint8& getLevel(key N);

	uint8& getSign (key N);

	//print the graph
	virtual void print(char* space="",FILE* file=stdout);

	//print links of node N
	virtual void print(key N,char* space="",FILE* file=stdout);

	//print statistics
	virtual void printStatistics(FILE* file);

	//access to all the down arch of a node
	iterator goDw(key N);

	//access to all the up arch of a node
	iterator goUp(key N);

	//swap node order
	void swapDwOrder(key A,key B);

	//order a 2-face
	bool orderFace(key face2);

	//access to all the node in the internal list
	list_iterator each(uint8 L);

	//next
	void next(bool bUp,key& ndst,key* dst,const key nsrc,const key* src);

	//get number of cells
	int getNCells(int lvl);

	//debug
	virtual void check(key N);
	virtual void check();

}; //end class


//INLINES

inline graph::arch_t* graph::getArch(key A)                           {return (arch_t*)arches[A];}
inline graph::node_t* graph::getNode(key N)                           {return (node_t*)nodes[N];}
inline uint16         graph::getNDw          (key Node)               {return getNode(Node)->Ndw;}
inline key            graph::getNumNode      ()                       {return nodes.getNBlocked();}
inline key            graph::getNumArch      ()                       {return arches.getNBlocked();}
inline key            graph::getMaxNode      ()                       {return nodes.getMax();}
inline key            graph::getMaxArch      ()                       {return arches.getMax();}
inline uint16         graph::getNUp          (key Node)               {return getNode(Node)->Nup;}
inline key            graph::getFirstDwArch  (key N)                  {return getNode(N)->Dw;}
inline key            graph::getFirstUpArch  (key N)                  {return getNode(N)->Up;}
inline key            graph::getLastDwArch   (key N)                  {return getArch(getFirstDwArch(N))->N1_prev;}
inline key            graph::getLastUpArch   (key N)                  {return getArch(getFirstUpArch(N))->N0_prev;}
inline key            graph::getFirstDwNode  (key N)                  {return getArch(getFirstDwArch(N))->N0;}
inline key            graph::getFirstUpNode  (key N)                  {return getArch(getFirstUpArch(N))->N1;};
inline key            graph::getLastDwNode   (key N)                  {return getArch(getLastDwArch (N))->N0;}
inline key            graph::getLastUpNode   (key N)                  {return getArch(getLastUpArch (N))->N1;}
inline key            graph::getNextDwArch   (key A)                  {return getArch(A)->N1_next;}
inline key            graph::getPrevDwArch   (key A)                  {return getArch(A)->N1_prev;}
inline key            graph::getNextUpArch   (key A)                  {return getArch(A)->N0_next;}
inline key            graph::getPrevUpArch   (key A)                  {return getArch(A)->N0_prev;}
inline const uint8&   graph::getLevel        (key N)                  {return getNode(N)->Level;}
inline uint8&         graph::getSign         (key N)                  {return getNode(N)->Sign;}
inline key            graph::getN0           (key Arch)               {return getArch(Arch)->N0;}
inline key            graph::getN1           (key Arch)               {return getArch(Arch)->N1;}
inline key&           graph::nodeData        (key Node)               {return getNode(Node)->Info;}
inline key&           graph::archData        (key Arch)               {return getArch(Arch)->Info;}
inline key&           graph::nodeTmp         (key Node)               {return getNode(Node)->Tmp;}
inline key            graph::getNextUpBrother(key A)                  {return getN0(getNextDwArch(A));}
inline key            graph::getNextDwBrother(key A)                  {return getN1(getNextUpArch(A));}
inline key            graph::getPrevUpBrother(key A)                  {return getN0(getPrevDwArch(A));}
inline key            graph::getPrevDwBrother(key A)                  {return getN1(getPrevUpArch(A));}
inline void				 graph::remArch(key N0,key N1,Direction dir)     {remArch(findArch(N0,N1,dir),dir);}
inline int            graph::getNCells(int L)								 {return (int)nodelists.n[L];}

inline graph* graph::clone() const
{
	return new graph(*this);
}

inline key graph::addNode(uint8 L)
{
	key N=nodes.block();
	node_t& node=*getNode(N);
	node.Ndw=node.Nup=0;
	node. Dw=node. Up=0;
	node.Info	=0;
	node.Level	=L;
	node.Id		=0;
	if (nodelists.v[L]) getNode(nodelists.v[L])->Prev=N;
	node.Next=nodelists.v[L];
	node.Prev=0;
	nodelists.v[L]=N;
	++nodelists.n[L];
	return N;
}

inline void graph::swapDwOrder(key A,key B)
{
	if (A==B) return;

	arch_t& a=*getArch(A);
	arch_t& b=*getArch(B);
	assert(a.N1==b.N1);

	key Ap=a.N1_prev,An=a.N1_next;
	key Bp=b.N1_prev,Bn=b.N1_next;	

#define link(A,B) {getArch(A)->N1_next=(B);getArch(B)->N1_prev=(A);}

	if (An!=B || Bn!=A)
	{
		if (Ap!=B) {link(Ap,B);}
		if (An!=B) {link(B,An);} else {link(B,A);}
		if (Bp!=A) {link(Bp,A);}
		if (Bn!=A) {link(A,Bn);} else {link(A,B);}
	}

	if (getNode(a.N1)->Dw==A)
		getNode(a.N1)->Dw=B;

#undef link
}

inline			graph::list_iterator::list_iterator()											{}
inline			graph::list_iterator::list_iterator(const graph::list_iterator& src)	{*this=src;}
inline bool		graph::list_iterator::end()														{return Cursor==0;}
inline key		graph::list_iterator::operator*()												{return Cursor;}
inline key		graph::list_iterator::getPrev()													{return Prev;}
inline			graph::list_iterator graph::list_iterator::next()							{return list_iterator(*this)++;}

inline graph::list_iterator graph::each(uint8 L)
{
	list_iterator ret;
	ret.g=this;
	ret.Prev=0;
	ret.Cursor=this->nodelists.v[L];
	return ret;
};

inline graph::list_iterator& graph::list_iterator::operator++(int) 
{
	Prev=Cursor;
	Cursor=g->getNode(Cursor)->Next;
	return *this;
}


inline graph::iterator graph::goDw(key N)
{
	iterator ret;
	ret.g    = this;
	ret.cont = 0;
	ret.dir  = DIRECTION_DOWN;
	ret.max  = (getNode(N)->Ndw);
	ret.A    = (getNode(N)-> Dw);
	return ret;
}

inline graph::iterator graph::goUp(key N)
{
	iterator ret;
	ret.g    = this;
	ret.cont = 0;
	ret.dir  = DIRECTION_UP;
	ret.max  = (getNode(N)->Nup);
	ret.A    = (getNode(N)-> Up);
	return ret;	
}

inline						graph::iterator::iterator()									{}
inline						graph::iterator::iterator(const graph::iterator& src)	{*this=src;}
inline bool					graph::iterator::end()											{return cont>=max;}
inline key					graph::iterator::getNode()										{return (dir==DIRECTION_UP)?(g->getArch(A)->N1):(g->getArch(A)->N0);}
inline key					graph::iterator::getArch()										{return A;}
inline key					graph::iterator::operator*()									{return getNode();}
inline graph::iterator	graph::iterator::next()											{return iterator(*this)++;}

inline graph::iterator& graph::iterator::operator++(int) 
{
	cont++;
	A=(dir==DIRECTION_UP)?(g->getArch(A)->N0_next):(g->getArch(A)->N1_next);
	return *this;
}


////////////////////////////////////////////////////////////////////////////////////////
//TODO TOCHANGE
//////////////////////////////////////////////////////////////////////
inline key* graph::getNavigation(key**& nav)
{
	static key* _nnav=0,*_nav[MAX_GEOMETRIC_DIMENSION+2];

	if (!_nnav)
	{
		_nnav=(key*)malloc(sizeof(real)*(MAX_GEOMETRIC_DIMENSION+2));
			
		for (int i=0;i<(key)(MAX_GEOMETRIC_DIMENSION+2);++i) 
			_nav[i]=(key*)malloc(MAX_CELLS_PER_LEVEL*sizeof(key));
	}

	nav=_nav;
	return _nnav;
}



#endif //_GRAPH_H__

