/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */


#include <time.h>

#include "ggraph.h"
#include "bsp.h"
#include "gviewer.h"

#define OPTION_EXPAND    "(e)xpand"
#define OPTION_VIEWMODE  "(v)iewcells"
#define OPTION_DEEP      "(m)ax deep"
#define OPTION_LINEWIDTH "(l)ine width"
#define OPTION_POINTSIZE "(p)oint size"
#define OPTION_BBALL     "(b)ounding ball"

XgeViewer::XgeViewer()
{
	panel=new Panel("mainpanel");
	options  =new Options();

	options->set(OPTION_EXPAND  ,(float)	0,"expand the current scene");
	options->set(OPTION_VIEWMODE,(int)		3,"view d-cells");
	options->set(OPTION_DEEP    ,(int)		0,"max deep");
	options->set(OPTION_LINEWIDTH,(float)  1,"line width");
	options->set(OPTION_POINTSIZE,(float)  1,"point size");
	options->set(OPTION_BBALL,(bool)   false,"bounding ball");

	add(panel);
	add(options);

	float f[]={-0.5,0,0};
	pack(f);


}

XgeViewer::~XgeViewer()
{
	int i;

	for (i=0;i<blist.size();i++)
	{
		delete blist[i]->getSolver();
		delete blist[i];
	}

	for (i=0;i<glist.size();i++)
	{
		delete glist[i];
	}

	for (i=0;i<dlist.size();i++)
	{
		delete dlist[i];
	}

	delete panel;
	delete options;
}

int XgeViewer::getViewmode()
{
	return options->getInt(OPTION_VIEWMODE);
}

float XgeViewer::getPointSize()
{
	return options->getFloat(OPTION_POINTSIZE);
}

void XgeViewer::setPointSize(float value)
{
	options->set(OPTION_POINTSIZE,(float)value);
}

float XgeViewer::getLineWidth()
{
	return options->getFloat(OPTION_LINEWIDTH);
}

void XgeViewer::setLineWidth(float value)
{
	options->set(OPTION_LINEWIDTH,(float)value);
}

bool XgeViewer::getBBall()
{
	return options->getBool(OPTION_BBALL);
}


void XgeViewer::setMaxDeep(int value)
{
	for (int i=0;i<blist.size();i++)
	{
		BspDraw* bd=blist[i];
		bsp_solver* solver=bd->getSolver();
		solver->clear();
		solver->setMaxDeep(value);
		solver->visit();
		solver->getGraph()->fixPlaneOrientation();
		bd->setGraph(solver->getGraph());
	}
}

float XgeViewer::getExpand()
{
	return options->getFloat(OPTION_EXPAND);
}

void XgeViewer::push(ggraph* g,float R,float G,float B)
{
	XgeDraw* gv=new XgeDraw(g,R,G,B);
	glist.push_back(gv);
	getPanel()->add(gv);
}

void  XgeViewer::push(XgeDraw* gv)
{
	glist.push_back(gv);
	getPanel()->add(gv);
}


void XgeViewer::push(Draw* dv)
{
	dlist.push_back(dv);
	getPanel()->add(dv);
}


ggraph* XgeViewer::push(bsp* src,real tolerance,int maxdeep,real limit)
{
	bsp_solver* solver=new bsp_solver(src,tolerance,maxdeep,limit);
	solver->setSimplify(false);
	solver->setMaxDeep(options->getInt(OPTION_DEEP));

	clock_t t1=clock();
	printf("Pushing bsp...");fflush(stdout);
	solver->visit();
	solver->getGraph()->fixPlaneOrientation();
	printf("finished in %.2f seconds\n",(clock()-t1)/(float)CLOCKS_PER_SEC);

#ifdef _DEBUG
	solver->getGraph()->check();
#endif

	BspDraw* draw=new BspDraw(solver);
	blist.push_back(draw);
	getPanel()->add(draw);

	return solver->getGraph();
}


void XgeViewer::keydown(unsigned char key, GLint x, GLint y)
{
	const char* actual=options->actual();

	switch(key)
	{
	case 'r':case 'R':
		{
			int value=options->getInt(OPTION_DEEP);
			setMaxDeep(value);
			break;
		}

	case 'b':case 'B':
	{
		if (strcmp(actual,OPTION_BBALL))
			options->setActual(OPTION_BBALL);
		else
		{
			bool nvalue=!options->getBool(OPTION_BBALL);
			options->set(OPTION_BBALL,nvalue);
		}
		break;
	}

	case 'l':case 'L':
		{
			if (strcmp(actual,OPTION_LINEWIDTH))
				options->setActual(OPTION_LINEWIDTH);
			else
			{
				float nvalue=options->getFloat(OPTION_LINEWIDTH)+((key=='e')?+0.5:-0.5);
				if (nvalue<0) nvalue=0;
				options->set(OPTION_LINEWIDTH,nvalue);
			}
			break;
		}

	case 'p':case 'P':
		{
			if (strcmp(actual,OPTION_POINTSIZE))
				options->setActual(OPTION_POINTSIZE);
			else
			{
				float nvalue=options->getFloat(OPTION_POINTSIZE)+((key=='e')?+0.5:-0.5);
				if (nvalue<0) nvalue=0;
				options->set(OPTION_POINTSIZE,nvalue);
			}
			break;
		}

	case 'e':case 'E':
		{
			if (strcmp(actual,OPTION_EXPAND))
				options->setActual(OPTION_EXPAND);
			else
			{
				float nvalue=options->getFloat(OPTION_EXPAND)+((key=='e')?+1:-1);
				if (nvalue<0) nvalue=0;
				options->set(OPTION_EXPAND,nvalue);
			}
			break;
		}

	case 'v':case 'V':
		{
			if (strcmp(actual,OPTION_VIEWMODE))
				options->setActual(OPTION_VIEWMODE);
			else
			{
				int nvalue=options->getInt(OPTION_VIEWMODE)+((key=='v')?+1:-1);
				nvalue=nvalue % 4;
				options->set(OPTION_VIEWMODE,nvalue);
			}
			break;
		}

	case 'm':case 'M':
		{
			if (strcmp(actual,OPTION_DEEP))
				options->setActual(OPTION_DEEP);
			else
			{
				int nvalue=options->getInt(OPTION_DEEP)+((key=='v')?+1:-1);
				nvalue=max2(nvalue,0);
				options->set(OPTION_DEEP,nvalue);
				setMaxDeep(nvalue);
			}
			break;
		}
	
	default:
		Win::keydown(key,x,y);
		break;
	}

	redisplay();
}

void XgeViewer::special(int key, GLint x, GLint y)
{
const char* actual=options->actual();

	switch(key)
	{
		case GLUT_KEY_LEFT:
			options->prev();
			break;

		case GLUT_KEY_RIGHT:
			options->next();
			break;

		case GLUT_KEY_UP:
		case GLUT_KEY_DOWN:

			const char* actual=options->actual();

			if (!strcmp(actual,OPTION_EXPAND))
			{
				float nvalue=options->getFloat(OPTION_EXPAND)+(key==GLUT_KEY_UP?0.2:-0.2);
				if (nvalue<0) nvalue=0;
				options->set(OPTION_EXPAND,nvalue);
			}
			else if (!strcmp(actual,OPTION_VIEWMODE))
			{
				int viewmode=options->getInt(OPTION_VIEWMODE);
				viewmode+=key==GLUT_KEY_UP?1:-1;
				viewmode=viewmode%4;
				options->set(OPTION_VIEWMODE,viewmode);
			}
			else if (!strcmp(actual,OPTION_DEEP))
			{
				int nvalue=options->getInt(OPTION_DEEP);
				nvalue+=key==GLUT_KEY_UP?1:-1;
				nvalue=max2(0,nvalue);
				options->set(OPTION_DEEP,nvalue);
				setMaxDeep(nvalue);
			}
			else if (!strcmp(actual,OPTION_POINTSIZE))
			{
				float nvalue=options->getFloat(OPTION_POINTSIZE);
				nvalue+=key==GLUT_KEY_UP?1:-1;
				nvalue=max2(0,nvalue);
				options->set(OPTION_POINTSIZE,nvalue);
			}
			else if (!strcmp(actual,OPTION_LINEWIDTH))
			{
				float nvalue=options->getFloat(OPTION_LINEWIDTH);
				nvalue+=key==GLUT_KEY_UP?1:-1;
				nvalue=max2(0,nvalue);
				options->set(OPTION_LINEWIDTH,nvalue);
			}
			else if (!strcmp(actual,OPTION_BBALL))
			{
				bool nvalue=!(options->getBool(OPTION_BBALL));
				options->set(OPTION_BBALL,nvalue);
			}
			break;
	}

	redisplay();
}

/////////////////////////////////////////////////////////////////////////////////

void drawCircle(GLfloat radius,GLfloat centerx,GLfloat centery,GLfloat step=0)
{
	GLfloat angle;
	glBegin(GL_LINE_LOOP);
	if (step<=0) 
		step=PIGRECO/16;
	for (angle=0;angle<2*PIGRECO;angle+=step)
		glVertex2f(centerx+radius*cos(angle),centery+radius*sin(angle));
	glEnd();
}





/////////////////////////////////////////////////////////////////////

static void drawBallId(ggraph* g,real expandx,real expandy,real expandz)
{
	for (graph::list_iterator i=g->each(g->pointdim);!i.end();i++)
	{
		real* g0=g->boundingBall(*i);

		real x=g->pointdim>=1?g0[1]:0;
		real y=g->pointdim>=2?g0[2]:0;
		real z=g->pointdim>=3?g0[3]:0;

		glRasterPos3f(x+x*expandx,y+y*expandy,z+z*expandz);

		char s[256];
		sprintf(s,"0X%x(%d) %f %f %f",(int)g,(int)*i,g0[0],g0[1],g0[2]);
		for(char* c=s;*c;c++) 
			glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *c);
	}
}



/////////////////////////////////////////////////////////////////////

void drawCellId(ggraph* g,int celllvl,real expandx,real expandy,real expandz)
{
	key Points[64*256];

	for (graph::list_iterator i=g->each(celllvl);!i.end();i++)
	{
		key Cell=*i;

		if (celllvl==g->pointdim)
		{
			real* g0=g->boundingBall(Cell);
			real x=g->pointdim>=1?g0[1]:0;
			real y=g->pointdim>=2?g0[2]:0;
			real z=g->pointdim>=3?g0[3]:0;
			glRasterPos3f(x+x*expandx,y+y*expandy,z+z*expandz);
		}
		else
		{
			key NPoints=g->findCells(0,Cell,Points);
			real x=0,y=0,z=0;
			real coeff=1/(real)NPoints;

			for (int u=0;u<NPoints;u++)
			{
				real* g0=g->getGeom(Points[u]);
				x+=coeff*(g->pointdim>=1?g0[1]:0);
				y+=coeff*(g->pointdim>=2?g0[2]:0);
				z+=coeff*(g->pointdim>=3?g0[3]:0);
			}

			glRasterPos3f(x+x*expandx,y+y*expandy,z+z*expandz);
		}

		char s[256];
		sprintf(s,"%d",(int)Cell);

		for(char* c=s;*c;c++) 
			glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *c);
	}
}


/////////////////////////////////////////////////////////////////////////////////

static void drawBoundingBalls(ggraph* g,real expandx,real expandy,real expandz)
{
	graph::list_iterator i;

	switch(g->pointdim)
	{
	case 2:

		if (expandx || expandy)
		{
			for (i=g->each(2);!i.end();i++)
			{
				real* g0=g->boundingBall(*i);
				real tx=g0[1],ty=g0[2];
				glPushMatrix();
				glTranslatef(tx*expandx,ty*expandy,0);
				drawCircle(sqrt(g0[0]),g0[1],g0[2]);
				glPopMatrix();
			}
		}
		else
		{
			for (i=g->each(2);!i.end();i++)
			{
				real* g0=g->boundingBall(*i);
				drawCircle(sqrt(g0[0]),g0[1],g0[2]);
			}

			
		}
		break;

	case 3:

		/*
		if (expandx || expandy || expandz)
		{
			for (i=g->each(3);!i.end();i++)
			{
				real* g0=g->boundingBall(*i);
				real tx=g0[1],ty=g0[2],tz=g0[3];
				glPushMatrix();
				glTranslatef(tx*expandx,ty*expandy,tz*expandz);
				drawSphere(g0[0],g0[1],g0[2],g0[3]);
				glPopMatrix();
			}
		}
		else
		{
			for (i=g->each(3);!i.end();i++)
			{
				real* g0=g->boundingBall(*i);
				drawSphere(g0[0],g0[1],g0[2],g0[3]);
			}
		}
		*/
		break;
	}

	//drawBallId(g,expandx,expandy,expandz);

}

/////////////////////////////////////////////////////////////////////

void drawVertex(ggraph* g,key V,real expandx,real expandy,real expandz)
{
	real* g0=g->getGeom(V);

	real X = (g->pointdim>=1)? g0[1] : 0;
	real Y = (g->pointdim>=2)? g0[2] : 0;
	real Z = (g->pointdim>=3)? g0[3] : 0;

	glVertex3f(X+expandx*X,Y+expandy*Y,Z+expandz*Z);
}

/////////////////////////////////////////////////////////////////////

static void drawVertices(ggraph* g,real expandx,real expandy,real expandz)
{
	graph::list_iterator i;
	glBegin(GL_POINTS);

	switch(g->pointdim)
	{
	case 0:
		for (i=g->each(0);!i.end();i++)
		{
			glVertex3f(0,0,0);
		}
		break;

	case 1:

		if (expandx)
		{
			for (i=g->each(0);!i.end();i++)
			{
				real* g0=g->getGeom(*i);
				real tx=g0[1];
				glVertex3f(g0[1]+tx*expandx,0,0);
			}
		}
		else
		{
			for (i=g->each(0);!i.end();i++)
			{
				real* g0=g->getGeom(*i);
				glVertex3f(g0[1],0,0);
			}
		}
		break;

	case 2:

		if (expandx || expandy )
		{
			for (i=g->each(0);!i.end();i++)
			{
				real* g0=g->getGeom(*i);
				real tx=g0[1];
				real ty=g0[2];
				glVertex3f(g0[1]+tx*expandx,g0[2]+ty*expandy,0);
			}
		}
		else
		{
			for (i=g->each(0);!i.end();i++)
			{
				real* g0=g->getGeom(*i);
				glVertex3f(g0[1],g0[2],0);
			}
		}
		break;

	default:
		if (expandx || expandy || expandz)
		{
			for (i=g->each(0);!i.end();i++)
			{
				real* g0=g->getGeom(*i);
				real tx=g0[1];
				real ty=g0[2];
				real tz=g0[3];
				glVertex3f(g0[1]+tx*expandx,g0[2]+ty*expandy,g0[3]+tz*expandz);
			}
		}
		else
		{
			for (i=g->each(0);!i.end();i++)
			{
				real* g0=g->getGeom(*i);
				glVertex3f(g0[1],g0[2],g0[3]);
			}
		}
		break;
	}

	glEnd();
}

/////////////////////////////////////////////////////////////////////

void drawLine(ggraph* g,key Line,real expandx,real expandy,real expandz)
{
	key N0=g->getFirstDwNode(Line);real* g0=g->getGeom(N0);
	key N1=g->getLastDwNode (Line);real* g1=g->getGeom(N1);

	real X0 = (g->pointdim>=1)? g0[1] : 0; real X1 = (g->pointdim>=1)? g1[1] : 0;
	real Y0 = (g->pointdim>=2)? g0[2] : 0; real Y1 = (g->pointdim>=2)? g1[2] : 0;
	real Z0 = (g->pointdim>=3)? g0[3] : 0; real Z1 = (g->pointdim>=3)? g1[3] : 0;

	real tx=(g0[1]+g1[1])*0.5;
	real ty=(g0[2]+g1[2])*0.5;
	real tz=(g0[3]+g1[3])*0.5;

	glVertex3f(X0+tx*expandx,Y0+ty*expandy,Z0+tz*expandz);
	glVertex3f(X1+tx*expandx,Y1+ty*expandy,Z1+tz*expandz);
}

/////////////////////////////////////////////////////////////////////

static void drawLines(ggraph* g,real expandx,real expandy,real expandz)
{
	graph::list_iterator i;
	glBegin(GL_LINES);

	switch(g->pointdim)
	{
	case 0:
		break;

	case 1:

		if (expandx)
		{
			for (i=g->each(1);!i.end();i++)
			{
				key N0=g->getFirstDwNode(*i);real* g0=g->getGeom(N0);
				key N1=g->getLastDwNode (*i);real* g1=g->getGeom(N1);
				real tx=(g0[1]+g1[1])*0.5;
				glVertex3f(g0[1]+tx*expandx,0,0);
				glVertex3f(g1[1]+tx*expandx,0,0);
			}
		}
		else
		{
			for (i=g->each(1);!i.end();i++)
			{
				key N0=g->getFirstDwNode(*i);real* g0=g->getGeom(N0);
				key N1=g->getLastDwNode (*i);real* g1=g->getGeom(N1);
				glVertex3f(g0[1],0,0);
				glVertex3f(g1[1],0,0);
			}
		}
		break;

	case 2:

		if (expandx ||expandy)
		{
			for (i=g->each(1);!i.end();i++)
			{
				key N0=g->getFirstDwNode(*i);real* g0=g->getGeom(N0);
				key N1=g->getLastDwNode (*i);real* g1=g->getGeom(N1);
				real tx=(g0[1]+g1[1])*0.5;
				real ty=(g0[2]+g1[2])*0.5;
				glVertex3f(g0[1]+tx*expandx,g0[2]+ty*expandy,0);
				glVertex3f(g1[1]+tx*expandx,g1[2]+ty*expandy,0);
			}
		}
		else
		{
			for (i=g->each(1);!i.end();i++)
			{
				key N0=g->getFirstDwNode(*i);real* g0=g->getGeom(N0);
				key N1=g->getLastDwNode (*i);real* g1=g->getGeom(N1);
				glVertex3f(g0[1],g0[2],0);
				glVertex3f(g1[1],g1[2],0);
			}
		}
		break;

	default:

		if (expandx || expandy || expandz)
		{
			for (i=g->each(1);!i.end();i++)
			{
				key N0=g->getFirstDwNode(*i);real* g0=g->getGeom(N0);
				key N1=g->getLastDwNode (*i);real* g1=g->getGeom(N1);
				real tx=(g0[1]+g1[1])*0.5;
				real ty=(g0[2]+g1[2])*0.5;
				real tz=(g0[3]+g1[3])*0.5;
				glVertex3f(g0[1]+tx*expandx,g0[2]+ty*expandy,g0[3]+tz*expandz);
				glVertex3f(g1[1]+tx*expandx,g1[2]+ty*expandy,g1[3]+tz*expandz);
			}
		}
		else
		{
			for (i=g->each(1);!i.end();i++)
			{
				key N0=g->getFirstDwNode(*i);real* g0=g->getGeom(N0);
				key N1=g->getLastDwNode (*i);real* g1=g->getGeom(N1);
				glVertex3f(g0[1],g0[2],g0[3]);
				glVertex3f(g1[1],g1[2],g1[3]);
			}
		}
		break;
	}

	glEnd();
}

/////////////////////////////////////////////////////////////////////

void drawNonConvexFace(ggraph* g,key F,float linewidth=3)
{
	glDisable(GL_LIGHTING);
	glLineWidth(linewidth);
	key prev0=g->getFirstDwNode(g->getLastDwNode(F)),p1;
	key prev1=g->getLastDwNode (g->getLastDwNode(F)),p2;

	float R=rand()/(float)RAND_MAX;
	float G=rand()/(float)RAND_MAX;
	float B=rand()/(float)RAND_MAX;

	glColor3f(R,G,B);
	
	glBegin(GL_LINES);
	for (graph::iterator i=g->goDw(F);!i.end();i++,prev0=p1,prev1=p2)
	{
		p1=g->getFirstDwNode(*i);real* g1=g->getGeom(p1);
		p2=g->getLastDwNode (*i);real* g2=g->getGeom(p2);
		glVertex3f(g1[1],g1[2],(g->pointdim==2)?(0):(g1[3]));
		glVertex3f(g2[1],g2[2],(g->pointdim==2)?(0):(g2[3]));
	}
	glEnd();
	glLineWidth(1);


#if 0
	{
		glColor3f(0,0,0);	
		glPointSize(13);
		glBegin(GL_POINTS);
		for (graph::iterator i=g->goDw(F);!i.end();i++,prev0=p1,prev1=p2)
		{
			p1=g->getFirstDwNode(*i);real* g1=g->getGeom(p1);
			p2=g->getLastDwNode (*i);real* g2=g->getGeom(p2);
			glVertex3f(g1[1],g1[2],(g->pointdim==2)?(0):(g1[3]));
			glVertex3f(g2[1],g2[2],(g->pointdim==2)?(0):(g2[3]));
		}
		glEnd();
		glPointSize(1);
	}

#endif
}

/////////////////////////////////////////////////////////////////////

void drawFace(ggraph* g,key F)
{
	key prev0=g->getFirstDwNode(g->getLastDwNode(F)),p1;
	key prev1=g->getLastDwNode (g->getLastDwNode(F)),p2;

	glBegin(GL_POLYGON);
	for (graph::iterator i=g->goDw(F);!i.end();i++,prev0=p1,prev1=p2)
	{
		p1=g->getFirstDwNode(*i);
		p2=g->getLastDwNode (*i);
		if (p1!=prev0 && p1!=prev1) Swap(key,p1,p2);
		real* g1=g->getGeom(p1);
		glVertex3f(g1[1],g1[2],(g->pointdim==2)?(0):(g1[3]));
	}

	glEnd();
}

/////////////////////////////////////////////////////////////////////

void drawFaces(ggraph* g,real expandx,real expandy,real expandz,bool nonconvex)
{
	graph::list_iterator i;

	if (nonconvex)
		srand((int)g);
	
	switch(g->pointdim)
	{
	case 0:
		break;
		
	case 1:
		break;
		
	default:

		for (i=g->each(2);!i.end();i++)
		{
			key F=*i;

			if (expandx || expandy || expandz)
			{
				real tx=0,ty=0,tz=0;
				real coeff=0.5/g->getNDw(*i);;

				for (graph::iterator iedge=g->goDw(*i);!iedge.end();iedge++)
				{
					real* g0=g->getGeom(g->getFirstDwNode(*iedge));
					real* g1=g->getGeom(g->getLastDwNode (*iedge));
					tx+=coeff*(g0[1]+g1[1]);
					ty+=coeff*(g0[2]+g1[2]);
					if (g->pointdim==3) 
						tz+=(coeff*(g0[3]+g1[3]));
				}
				glPushMatrix();
				glTranslatef(expandx*tx,expandy*ty,expandz*tz);
			}

			if (g->pointdim==3 && g->nodeData(*i))
			{
				real* n=g->getGeom(*i);
				glNormal3f(n[1],n[2],n[3]);
			}

			if (nonconvex)
			{
				drawNonConvexFace(g,*i);
			}
			else
			{
				drawFace(g,*i);
			}
			
			if (expandx  || expandy || expandz) 
				glPopMatrix();
		}
		break;
	}

}


/////////////////////////////////////////////////////////////////////

static void draw3Cell(ggraph* g,real expandx,real expandy,real expandz)
{
	graph::list_iterator i;
	
	switch(g->pointdim)
	{
	case 0:
		break;
		
	case 1:
		break;

	case 2:
		break;
		
	case 3:

		for (i=g->each(3);!i.end();i++)
		{
			if (expandx || expandy || expandz)
			{
				key Cell=*i;
				real* t=g->boundingBall(*i);
				glPushMatrix();
				glTranslatef(expandx*t[1],expandy*t[2],expandz*t[3]);
			}

			for (graph::iterator j=g->goDw(*i);!j.end();j++)
			{
				if (g->nodeData(*j))
				{
					real* n=g->getGeom(*j);
					glNormal3f(n[1],n[2],n[3]);
				}
				drawFace(g,*j);
			}

			if (expandx || expandy || expandz) 
				glPopMatrix();
		}

		break;

		default:
			break;
	}

	glDisable(GL_LIGHTING);
}




/////////////////////////////////////////////////////////////////////

static void drawVertsId(ggraph* g,real expandx,real expandy,real expandz)
{
	for (graph::list_iterator i=g->each(0);!i.end();i++)
	{
		real* g0=g->getGeom(*i);
		real x=g->pointdim>=1?g0[1]:0;
		real y=g->pointdim>=2?g0[2]:0;
		real z=g->pointdim>=3?g0[3]:0;

		glRasterPos3f(x+x*expandx,y+y*expandy,z+z*expandz);

		char s[256];
		sprintf(s,"%d",(int)*i);
		for(char* c=s;*c;c++) 
			glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *c);
	}
}



/////////////////////////////////////////////////////////////////////
XgeDraw::XgeDraw(ggraph* g,float R,float G,float B,float linewidth)
{
	this->g=g;
	this->R=R;
	this->G=G;
	this->B=B;
	this->linewidth=linewidth;
	this->view_face_normal=false;
	this->view_cell_level=g->pointdim;
	this->nonconvex=false;
	
	//visualizza quello che e' disponibile
	while (view_cell_level>0 && g->each(view_cell_level).end())
		--view_cell_level;
}


/////////////////////////////////////////////////////////////////////

void XgeDraw::display(ggraph* g,int8 view_cell_level,
							 float R,float G,float B,
							 real expandx,real expandy,real expandz,bool nonconvex,
							 float pointsize,float linewidth,bool bball)
{
	graph::list_iterator i;

	switch(view_cell_level)
	{
	case 0:
		if (pointsize!=1) glPointSize(pointsize);
		glColor3f(R,G,B);
		glDisable(GL_LIGHTING);
		drawVertices(g,expandx,expandy,expandz);
		glPointSize(1);
		glEnable(GL_LIGHTING);
		break;

	case 1:
		if (linewidth!=1) glLineWidth(linewidth);
		glColor3f(R,G,B);
		glDisable(GL_LIGHTING);
		drawLines(g,expandx,expandy,expandz);
		glLineWidth(1);
		glEnable(GL_LIGHTING);
		break;

	case 2:

		if (nonconvex)
		{
			drawFaces(g,expandx,expandy,expandz,true);
		}
		else
		{
			glDisable(GL_LIGHTING);
			glPolygonOffset(1.0,2);
			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonMode(GL_FRONT_AND_BACK ,GL_FILL);
			glColor3f(R,G,B);
			drawFaces(g,expandx,expandy,expandz,false);
			glColor3f(0,0,0);
			glDisable(GL_POLYGON_OFFSET_FILL);
			glPolygonMode(GL_FRONT_AND_BACK ,GL_LINE);
			drawFaces(g,expandx,expandy,expandz,false);
			glPolygonMode(GL_FRONT_AND_BACK ,GL_FILL);
		}

		break;

	case 3:


#if 0
		glEnable(GL_LIGHTING);
		glColor3f(R,G,B);
		draw3Cell(g,expandx,expandy,expandz);
#else
		glEnable(GL_LIGHTING);
		glPolygonOffset(0.5,0.5);
		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonMode(GL_FRONT_AND_BACK ,GL_FILL);
		glColor3f(R,G,B);
		draw3Cell(g,expandx,expandy,expandz);
		

		glDisable(GL_LIGHTING);
		glColor3f(0,0,0);
		glDisable(GL_POLYGON_OFFSET_FILL);
		glPolygonMode(GL_FRONT_AND_BACK ,GL_LINE);
		draw3Cell(g,expandx,expandy,expandz);
		glPolygonMode(GL_FRONT_AND_BACK ,GL_FILL);
#endif
		break;


	default:
		break;
	}

	if (bball)
		drawBoundingBalls(g,expandx,expandy,expandz);

	//glColor3f(0,0,0);
	//drawVertsId(g,expandx,expandy,expandz);

	//glColor3f(0,0,0);
	//drawCellId(g,2,expandx,expandy,expandz);

	//glColor3f(1,1,0);
	//drawCellId(g,1,expandx,expandy,expandz);
	
}


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
void XgeDraw::display()
{
	if (g->pointdim>3)
		return;

	XgeViewer* win=(XgeViewer*)this->getPanel()->getWindow();
	real expand=win->getExpand()*0.1;
	int  viewmode=win->getViewmode();
	float  pointsize=win->getPointSize();
	float  linewidth=win->getLineWidth();
	bool   bball=win->getBBall();
	

	if (g->vmat && g->spacedim<=3)
	{
		real* m=g->vmat;
		glPushMatrix();
		
		switch(g->spacedim)
		{
		case 0:
			{
				float aux[16]=
				{
					1,0,0,   0,  
					0,1,0,   0,  
					0,0,1,   0,  
					0,0,0,m[0]
				};

				glMultMatrixf(aux);
				break;
			}
		case 1:
			{
				float aux[16]=
				{
					m[3],0,0,m[1], 
					0   ,1,0,   0, 
					0   ,0,1,   0, 
					m[2],0,0,m[0], 
						
				};
				glMultMatrixf(aux);
				break;
			}
		case 2:
			{
				float aux[16]=
				{
					m[4],m[7],0,m[1],
					m[5],m[8],0,m[2],
					0   ,   0,1,   0,
					m[3],m[6],0,m[0],
						
				};
				glMultMatrixf(aux);
				break;
			}
		case 3:
			{
				
				float aux[16]={
					m[ 5],m[ 9],m[13],m[ 1],
					m[ 6],m[10],m[14],m[ 2],
					m[ 7],m[11],m[15],m[ 3],
					m[ 4],m[ 8],m[12],m[ 0]
				};
				
				glMultMatrixf(aux);
				break;
			}
		}
	}

	display(g,view_cell_level,R,G,B,expand,expand,expand,nonconvex,pointsize,linewidth,bball);

	if (g->vmat && g->spacedim<=3)
	{
		glPopMatrix();
	}
}



////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

XgeDrawDebug::XgeDrawDebug()
{}

void XgeDrawDebug::add(ggraph* g,key Cell,float R,float G,float B)
{
	checkpoint(g && Cell && g->pointdim>=0);
	TCell aux={g,Cell};
	aux.R=R;aux.G=G;aux.B=B;
	cells.push_back(aux);
}

void XgeDrawDebug::add(int dim,real* v0,float R,float G,float B)
{
	TPoint aux;
	memset(aux.v0,0,sizeof(aux.v0));
	memcpy(aux.v0,v0,sizeof(real)*(dim+1));
	aux.R=R;aux.G=G;aux.B=B;
	points.push_back(aux);
}

void XgeDrawDebug::add(int dim,real* v0,real* v1,float R,float G,float B)
{
	TLine aux;
	memset(aux.v0,0,sizeof(aux.v0));memcpy(aux.v0,v0,sizeof(real)*(dim+1));
	memset(aux.v1,0,sizeof(aux.v0));memcpy(aux.v1,v1,sizeof(real)*(dim+1));
	aux.R=R;aux.G=G;aux.B=B;
	lines.push_back(aux);
}

void XgeDrawDebug::display()
{
	int i;
	
	XgeViewer* win=(XgeViewer*)this->getPanel()->getWindow();
	real expand=win->getExpand()*0.1;
	
	glDisable(GL_LIGHTING);
	glPointSize(3);
	glBegin(GL_POINTS);
	for (i=0;i<points.size();i++)
	{
		TPoint aux=points[i];
		glColor3f(aux.R,aux.G,aux.B);
		glVertex3f(aux.v0[1],aux.v0[2],aux.v0[3]);
	}
	glEnd();
	glPointSize(1);
	
	
	glLineWidth(3);
	glBegin(GL_LINES);
	for (i=0;i<lines.size();i++)
	{
		TLine aux=lines[i];
		glColor3f(aux.R,aux.G,aux.B);
		glVertex3f(aux.v0[1],aux.v0[2],aux.v0[3]);
		glVertex3f(aux.v1[1],aux.v1[2],aux.v1[3]);
	}
	glEnd();
	glLineWidth(1);
	
	
	for (i=0;i<cells.size();i++)
	{
		TCell aux=cells[i];
		glColor3f(aux.R,aux.G,aux.B);
		
		switch(aux.g->pointdim)
		{
		case 0:
			glBegin(GL_POINTS);
			drawVertex(aux.g,aux.Cell,expand,expand,expand);
			glEnd();
			break;
			
		case 1:
			glBegin(GL_LINES);
			drawLine(aux.g,aux.Cell,expand,expand,expand);
			glEnd();
			break;
			
		case 2:
			
			if (expand)
			{
				real tx=0,ty=0,tz=0;
				real coeff=0.5/aux.g->getNDw(aux.Cell);;
				
				for (graph::iterator iedge=aux.g->goDw(aux.Cell);!iedge.end();iedge++)
				{
					real* g0=aux.g->getGeom(aux.g->getFirstDwNode(*iedge));
					real* g1=aux.g->getGeom(aux.g->getLastDwNode (*iedge));
					tx+=coeff*(g0[1]+g1[1]);
					ty+=coeff*(g0[2]+g1[2]);
					if (aux.g->pointdim==3) tz+=(coeff*(g0[3]+g1[3]));
				}
				glPushMatrix();
				glTranslatef(expand*tx,expand*ty,expand*tz);
			}
			
			glDisable(GL_LIGHTING);
			glPolygonOffset(1.0,2);
			glEnable(GL_POLYGON_OFFSET_FILL);
			glPolygonMode(GL_FRONT_AND_BACK ,GL_FILL);
			glColor3f(aux.R,aux.G,aux.B);
			drawFace(aux.g,aux.Cell);
			glColor3f(0,0,0);
			glDisable(GL_POLYGON_OFFSET_FILL);
			glPolygonMode(GL_FRONT_AND_BACK ,GL_LINE);
			drawFace(aux.g,aux.Cell);
			glPolygonMode(GL_FRONT_AND_BACK ,GL_FILL);
			
			if (expand)
				glPopMatrix();
			
			break;
		}
	}
	
	glColor3f(0,0,0);
	glEnable(GL_LIGHTING);
}






////////////////////////////////////////////////////////////////////////

inline int8 XgeContour2::vsign(key N)
	{return g->getGeom(N)[2]<isovalue?-1:+1;}

inline int8 XgeContour2::vless(key A,key B)
{
	return (g->getGeom(A)[2]< g->getGeom(B)[2])       ||
			 (g->getGeom(A)[2]==g->getGeom(B)[2] && A<B);
}

XgeContour2::XgeContour2(ggraph* g)
{
	this->g=g;
	this->changing_isovalue=false;
	this->isovalue=0;
}

////////////////////////////////////////////////////////////////////////

void XgeContour2::mouse(GLint button, GLint state, GLint x, GLint y) 
{
	if (button==GLUT_RIGHT_BUTTON)
		changing_isovalue=(state==GLUT_DOWN);
}

////////////////////////////////////////////////////////////////////////

void XgeContour2::display()
{
	glDisable(GL_LIGHTING);
	ggraph c=getContour();
	XgeDraw v(&c);
	v.display();
	glEnable(GL_LIGHTING);
}

////////////////////////////////////////////////////////////////////////

void XgeContour2::motion(GLint x, GLint y)
{
	if (changing_isovalue)
	{
		float minf=-100;
		float maxf=+100;
		this->isovalue=minf+(x-panel->getX())*(maxf-minf)/panel->getW();
		printf("setting isovalue %f\n",(float)isovalue);
		panel->getWindow()->redisplay();
	}
}

////////////////////////////////////////////////////////////////////////

ggraph XgeContour2::getContour()
{
	graph::list_iterator it;
	ggraph   m(3);
	
	// interpolation for edges
	for (it=g->each(1);!it.end();it++)
	{
		key E=*it;
		key N0=g->getFirstDwNode(E);
		key N1=g->getLastDwNode (E) ;
		
		g->nodeTmp(E)=0; //assume no interpolation
		
		if (!vless(N0,N1))
			{Swap(key,N0,N1);}
		
		if (vsign(N0)==-1 && vsign(N1)==+1)
		{
			real* v0= g->getGeom(N0);
			real* v1= g->getGeom(N1);
			
			real beta  = (v0[2]==v1[2])?0.5f:((isovalue-v0[2])/(v1[2]-v0[2]));
			real alpha = (1-beta);
			checkpoint(beta>=0 && beta<=1);
			
			real X =alpha*v0[1]+beta*v1[1];
			real Y =alpha*v0[2]+beta*v1[2];
			real Z =alpha*v0[3]+beta*v1[3];
			
			real c[]={1,X,Y,Z};
			key N=m.addVertex(c);
			g->nodeTmp(E)=N;
		}
	}
	
	//for each triangle build an edge
	for (it=g->each(2);!it.end();it++)
	{
		key F=*it;
		key A0,A1,A2;
		
		key N0=g->getN1(A0=g->getFirstUpArch(F));
		key N1=g->getN1(A1=g->getNextUpArch(A0));
		key N2=g->getN1(A2=g->getNextUpArch(A1));
		
		if (vsign(N0)==vsign(N1) && vsign(N1)==vsign(N2))
			continue;
		
		if (!vless(N0,N1)) {Swap(key,N0,N1);}
		if (!vless(N0,N2)) {Swap(key,N0,N2);}
		if (!vless(N1,N2)) {Swap(key,N1,N2);}
		
		int n=0;
		key edge[2];
		
		key E01 =g->findCommonNode(N0,N1,DIRECTION_UP); if (vsign(N0)==-1 && vsign(N1)==+1) edge[n++]=g->nodeTmp(E01);
		key E02 =g->findCommonNode(N0,N2,DIRECTION_UP); if (vsign(N0)==-1 && vsign(N2)==+1) edge[n++]=g->nodeTmp(E02);
		key E12 =g->findCommonNode(N1,N2,DIRECTION_UP); if (vsign(N1)==-1 && vsign(N2)==+1) edge[n++]=g->nodeTmp(E12);
		
		checkpoint(n==2);
		
		m.addEdge(edge[0],edge[1]);
	}
	return m;
}

////////////////////////////////////////////////////////////////////////////////////////


inline int8 XgeContour3::vsign(key N)
	{return g->getGeom(N)[2]<isovalue?-1:+1;}

inline int8 XgeContour3::vless(key A,key B)
{
	return (g->getGeom(A)[2]< g->getGeom(B)[2])       ||
			 (g->getGeom(A)[2]==g->getGeom(B)[2] && A<B);
}


XgeContour3::XgeContour3(ggraph* g)
{
	this->g=g;
	this->changing_isovalue=false;
	this->isovalue=0;
}


void XgeContour3::mouse(GLint button, GLint state, GLint x, GLint y) 
{
	if (button==GLUT_RIGHT_BUTTON)
		changing_isovalue=(state==GLUT_DOWN);
}

void XgeContour3::display()
{
	//ggraph c=getContour();
	//ggraph_viewer v(&c);
	//v.display(2);
}


void XgeContour3::motion(GLint x, GLint y)
{
	if (changing_isovalue)
	{
		float minf=-100;
		float maxf=+100;
		this->isovalue=minf+(x-panel->getX())*(maxf-minf)/panel->getW();
		printf("setting isovalue %f\n",(float)isovalue);
		panel->getWindow()->redisplay();
	}
}


ggraph XgeContour3::getContour()
{
	graph::list_iterator it;
	ggraph m(3);
	
	// interpolation for edges
	for (it=g->each(1);!it.end();it++)
	{
		key E=*it;
		key N0=g->getFirstDwNode(E);
		key N1=g->getLastDwNode (E) ;
		
		g->nodeTmp(E)=0; //assume no interpolation
		
		if (!vless(N0,N1))
		{Swap(key,N0,N1);}
		
		if (vsign(N0)==-1 && vsign(N1)==+1)
		{
			real* v0=g->getGeom(N0);
			real* v1=g->getGeom(N1);
			
			real beta  = (v0[2]==v1[2])?0.5f:((isovalue-v0[2])/(v1[2]-v0[2]));
			real alpha = (1-beta);
			
			checkpoint(beta>=0 && beta<=1);
			
			float X =alpha*v0[1]+beta*v1[1];
			float Y =alpha*v0[2]+beta*v1[2];
			float Z =alpha*v0[3]+beta*v1[3];
			
			//float NX=alpha*v0[3]+beta*v1[3];
			//float NY=alpha*v0[4]+beta*v1[4];
			//float NZ=alpha*v0[5]+beta*v1[5];
			//float len=sqrt(NX*NX+NY*NY+NZ*NZ);
			//if (len) {NX/=len;NY/=len;NZ/=len;}
			//float F=isovalue;
			
			float v[]={1,X,Y,Z};
			key N=m.addVertex(v);
			g->nodeTmp(E)=N;
		}
	}
	
	//for each tet build a quad
	for (it=g->each(3);!it.end();it++)
	{
		key T=*it;
		key A0,A1,A2,A3;
		key N0=g->getN1(A0=g->getFirstUpArch(T));
		key N1=g->getN1(A1=g->getNextUpArch(A0));
		key N2=g->getN1(A2=g->getNextUpArch(A1));
		key N3=g->getN1(A3=g->getNextUpArch(A2));
		
		if (vsign(N0)==vsign(N1) && vsign(N1)==vsign(N2) && vsign(N2)==vsign(N3))
			continue;
		
		//find the 4 edges which forms a quad
		key N[]={N0,N1,N2,N3};
		
		key n=0;
		key quad[4];
		key edges[8];
		
		//for each edge of the tet...
		
		int i;
		for (    i=  0;i<3;i++)
			for (int j=i+1;j<4;j++)
			{
				key E =g->findCommonNode(N[i],N[j],DIRECTION_UP);
				
				key N0=N[i];
				key N1=N[j];
				
				if (!vless(N0,N1)) 
				{Swap(key,N0,N1);}
				
				if (vsign(N0)==-1 && vsign(N1)==+1)
				{
					key N=(key)(g->nodeTmp(E));
					quad[n]=N;
					edges[n*2+0]=N0;
					edges[n*2+1]=N1;
					n++;
				}
			}
			
			if (!n) continue;
			checkpoint(n==3 || n==4);
			
			if (n==3)
			{
				key t1[3]={quad[0],quad[1],quad[2]};
				m.addFace(3,t1);
				
				{
					real* g0=m.getGeom(quad[0]);
					real* g1=m.getGeom(quad[1]);
					real* g2=m.getGeom(quad[2]);
					//printf("(%.1f %.1f %.1f) (%.1f %.1f %.1f) (%.1f %.1f %.1f)\n",g0[1],g0[2],g0[3],g1[1],g1[2],g1[3],g2[1],g2[2],g2[3]);
				}
				
			}
			else if (n==4)
			{
				//could be necessary a reorder
				for (int i=0;i<6;i+=2)
				{
				  int j;
					for (j=i+2;j<8;j+=2)
					{
						if (edges[i+0]==edges[j] || edges[i+0]==edges[j+1] ||
							edges[i+1]==edges[j] || edges[i+1]==edges[j+1])
						{
							if (j!=(i+2))
							{
								Swap(key,edges[(i+2)+0],edges[j+0]);
								Swap(key,edges[(i+2)+1],edges[j+1]);
								Swap(key,quad[(i+2)>>1],quad[j>>1]);
							}
							break;
						}
					}
					checkpoint(j!=8);
				}
				
				key t1[3]={quad[0],quad[1],quad[2]};
				m.addFace(3,t1);
				
				key t2[3]={quad[2],quad[3],quad[0]};
				m.addFace(3,t2);
			}
	}
	
	//if (m.nodelists.n[2]) 
	//	printf("Contour %d triangles\n",m.nodelists.n[2]);
	//m2->setNormals();
	return m;
}


//-------------------------------------------------------------
//-------------------------------------------------------------

/*
Grid2Draw::Grid2Draw(int istepx,int istepy,
		  float minx,float maxx,
		  float miny,float maxy,
		  float R,float G,float B,float linewidth)
{
	this->istepx=istepx;
	this->istepy=istepy;
	this->minx=minx;
	this->miny=miny;
	this->maxx=maxx;
	this->maxy=maxy;
	this->R=R;
	this->G=G;
	this->B=B;
	this->linewidth=linewidth;
}


//-------------------------------------------------------------
//-------------------------------------------------------------
 void Grid2Draw::display()
{
	glColor3f(0.7f,0.7f,0.7f);
	float fstepx=(maxx-minx)/istepx;
	float fstepy=(maxy-miny)/istepy;

	glLineWidth(linewidth);
	glBegin(GL_LINES);
	for (int i=0;i<=istepx;i++)
	{
		float X=minx+i*fstepx;
		glVertex2f(X,miny);
		glVertex2f(X,maxy);
	}

	for (int j=0;j<=istepy;j++)
	{
		float Y=miny+j*fstepy;
		glVertex2f(minx,Y);
		glVertex2f(maxx,Y);
	}
	glEnd();
	glLineWidth(1);
}


*/



#if 0


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
#define normalize(v) { \
                       float len; \
                       len = sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]); \
                       if (len != 0.0) { \
                          v[0] /= len; \
                          v[1] /= len; \
                          v[2] /= len; \
                       } \
                     }


inline void Mesh3d::setGrad(void)
{
   float u[4], v[4], w[4], x[4], y[4], z[4], g[4];
   float len;
	graph::list_iterator it;

	for (it=this->g.forEachNodeInList(0);!it.end(); it++) 
	{
		key N=*it;
      float* verts=(float*)(this->vertices[this->g.nodeData(N)]);
		verts[3]=verts[4]=verts[5]=0;
   }



   for (it=this->g.forEachNodeInList(3);!it.end(); it++) 
	{
		key T=*it;
		key A0,A1,A2,A3;
		key N0=this->g.getN1(A0=this->g.getFirstUpArch(T));float* verts0=(float*)(vertices[this->g.nodeData(N0)]);
		key N1=this->g.getN1(A1=this->g.getNextUpArch(A0));float* verts1=(float*)(vertices[this->g.nodeData(N1)]);
		key N2=this->g.getN1(A2=this->g.getNextUpArch(A1));float* verts2=(float*)(vertices[this->g.nodeData(N2)]);
		key N3=this->g.getN1(A3=this->g.getNextUpArch(A2));float* verts3=(float*)(vertices[this->g.nodeData(N3)]);

		u[0] = verts1[0] - verts0[0];
		u[1] = verts1[1] - verts0[1];
		u[2] = verts1[2] - verts0[2];
      u[3] = verts1[6] - verts0[6];
      
      v[0] = verts2[0] - verts0[0];
      v[1] = verts2[1] - verts0[1];
      v[2] = verts2[2] - verts0[2];
      v[3] = verts2[6] - verts0[6];
      
      w[0] = verts3[0] - verts0[0];
      w[1] = verts3[1] - verts0[1];
      w[2] = verts3[2] - verts0[2];
      w[3] = verts3[6] - verts0[6];

      x[0] = verts3[0] - verts1[0];
      x[1] = verts3[1] - verts1[1];
      x[2] = verts3[2] - verts1[2];
      
      y[0] = verts3[0] - verts2[0];
      y[1] = verts3[1] - verts2[1];
      y[2] = verts3[2] - verts2[2];
      
      z[0] = verts2[0] - verts1[0];
      z[1] = verts2[1] - verts1[1];
      z[2] = verts2[2] - verts1[2];

      g[0] =   u[1] * (v[2]*w[3] - v[3]*w[2]) + u[2] * (v[3]*w[1] - v[1]*w[3]) + u[3] * (v[1]*w[2] - v[2]*w[1]);
      g[1] =   u[0] * (v[2]*w[3] - v[3]*w[2]) + u[2] * (v[3]*w[0] - v[0]*w[3]) + u[3] * (v[0]*w[2] - v[2]*w[0]);
      g[2] =   u[0] * (v[1]*w[3] - v[3]*w[1]) + u[1] * (v[3]*w[0] - v[0]*w[3]) + u[3] * (v[0]*w[1] - v[1]*w[0]);

      float weight=1;

		/*
		weight= u[0]*(v[1]*w[2] - v[2]*w[1])- u[1]*(v[2]*w[0] - v[0]*w[2])+ u[2]*(v[0]*w[1] - v[1]*w[0]);
		weight= (-u[0])*(x[1]*z[2] - x[2]*z[1])- (-u[1])*(x[2]*z[0] - x[0]*z[2])+ (-u[2])*(x[0]*z[1] - x[1]*z[0]);
		weight= (-v[0])*((-z[1])*y[2] - (-z[2])*y[1])- (-v[1])*((-z[2])*y[0] - (-z[0])*y[2])+ (-v[2])*((-z[0])*y[1] - (-z[1])*y[0]);
		weight= (-w[0])*((-y[1])*(-x[2]) - (-y[2])*(-x[1]))- (-w[1])*((-y[2])*(-x[0]) - (-y[0])*(-x[2]))+ (-w[2])*((-y[0])*(-x[1]) - (-y[1])*(-x[0]));
		*/

      verts0[3] += g[0] * weight;verts0[4] += g[1] * weight;verts0[5] += g[2] * weight;
      verts1[3] += g[0] * weight;verts1[4] += g[1] * weight;verts1[5] += g[2] * weight;
      verts2[3] += g[0] * weight;verts2[4] += g[1] * weight;verts2[5] += g[2] * weight;
      verts3[3] += g[0] * weight;verts3[4] += g[1] * weight;verts3[5] += g[2] * weight;
	}

	for (it=this->g.forEachNodeInList(0);!it.end(); it++) 
	{
		key N=*it;
		float* verts=(float*)(this->vertices[this->g.nodeData(N)]);

      len = sqrt(verts[3]*verts[3] + verts[4]*verts[4] +verts[5]*verts[5]);
      if (len) 
		{
			verts[3] /= len;
			verts[4] /= len;
			verts[5] /= len;
		}
   }
}




//=======================================================
//=======================================================
void Mesh2d::setNormals(int flip_sign)
{
	float x,y,z;
	float x0,y0,z0;
	float x1,y1,z1;
	float len;
	float recip;

	graph::list_iterator it;

	for (it=g.forEachNodeInList(0);!it.end();it++)
	{
		key N=*it;
		float* v=(float*)vertices[g.nodeData(N)];
		v[3]=0;v[4]=0;v[5]=0;
	}

	for (it=g.forEachNodeInList(2);!it.end();it++)
	{
		key F=*it;

		key N0,N1,N2;
		getTriangle(F,N0,N1,N2);

		float* v0=(float*)vertices[g.nodeData(N0)];
		float* v1=(float*)vertices[g.nodeData(N1)];
		float* v2=(float*)vertices[g.nodeData(N2)];

		x0 = v2[0] - v0[0];
		y0 = v2[1] - v0[1];
		z0 = v2[2] - v0[2];
		
		x1 = v1[0] - v0[0];
		y1 = v1[1] - v0[1];
		z1 = v1[2] - v0[2];
		
		x = y0 * z1 - z0 * y1;
		y = z0 * x1 - x0 * z1;
		z = x0 * y1 - y0 * x1;

		len = x*x + y*y + z*z;
		
		if (len==0) 
			{x=y=z=0;}
		else 
			{recip=1/sqrt (len);x*=recip;y*=recip;z*=recip;}

		v0[3]+=x;v1[3]+=x;v2[3]+=x;
		v0[4]+=y;v1[4]+=y;v2[4]+=y;
		v0[5]+=z;v1[5]+=z;v2[5]+=z;
	}



	for (it=g.forEachNodeInList(0);!it.end();it++)
	{
		key N=*it;

		float* v=(float*)vertices[g.nodeData(N)];

		if ((len = v[3]*v[3]+v[4]*v[4]+v[5]*v[5])==0) 
			{v[3] = v[4] = v[5] = 0;}
		else 
		{
			recip=(1 / sqrt (len))*(flip_sign?-1:1);
			v[3] *= recip;v[4] *= recip;v[5] *= recip;
		}

		v[3]*=-1;
		v[4]*=-1;
		v[5]*=-1;
	}
}

#endif




