/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _MAT_H__
#define _MAT_H__

#include <stdio.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>

#include <assert.h>
#include <string.h>
#include <math.h>

#include "mat.h"






template <class T>
class tmat
{
public:


	/* create an identity matrix */
	static void identity(int dim,T* dst);

	/* create an identity matrix */
	static  void zero(int dim,T* dst);

	/* copy */
	static void copy(int dstdim,T* dst,int srcdim,const T* src);

	/* embed */
	static void embed(int dstdim,int srcdim,T*& mat);

	/* swap */
	static void swap_rows(int dim,T* mat,const int* perm);
	static void swap_cols(int dim,T* mat,const int* perm);

	static void swap_rows(int dim,T* mat,int r1,int r2);
	static void swap_cols(int dim,T* mat,int c1,int c2);

	/* print to <stream> the matrix ((a11 a12 a13...) (a21 a22 a23...) ...)*/
	static void print(int dim,const T* mat,FILE* stream=stdout);

	/* transpose a matrix */
	static void transpose(int dim,T* dst,const T* src);

	/* create a new matrix which is the product of a*b */
	static void product(int dim,T* dst,const T* a,const T* b);

	//add
	static void add(int dim,T* dst,const T* a,const T* b);

	static void subtract(int dim,T* dst,const T* a,const T* b);

	/* scale a matrix */
	static void scalarproduct(int dim,T* dst,const T* src,const float coeff);

	//invert  
	static T invert(int dim,T* dst,const T* src); 

	//submatrix
	static void submatrix(T* dst,int dim,const T* src,int r1,int r2,int c1,int c2);

	//transformations
	static void scale     (int dim,int ref        ,      T  factor,T* vmat=0,T* hmat=0);
	static void translate (int dim,int ref        ,      T  factor,T* vmat=0,T* hmat=0);
	static void rotate    (int dim,int ax1,int ax2,      T  factor,T* vmat=0,T* hmat=0);
	static void shear     (int dim,int ref        ,const T *factor,T* vmat=0,T* hmat=0);

	//verify the invert
	static bool isInvert(int dim,const T* Vmat,const T* Hmat,real tolerance=1e-2);

};

template <class T>
T tmat<T>::invert(int dim,T* dst,const T* a) 
{
	T det;

	switch(dim)
	{
	case 1:
		det=a[0];
		if (dst  && det)
		{
			T a00=1/det;
			dst[0]=a00;
		}
		break;

	case 2:
	{
		det=a[0]*a[3]-a[1]*a[2];
		if (dst  && det)
		{
			T a00=+a[3]/det;
			T a01=-a[1]/det;
			T a10=-a[2]/det;
			T a11=+a[0]/det;

			dst[0]=a00;dst[1]=a01;
			dst[2]=a10;dst[3]=a11;
		}
		break;
	}
	case 3:
	{
		det=a[0]*a[4]*a[8]-a[0]*a[7]*a[5]+a[3]*a[2]*a[7]-a[3]*a[8]*a[1]+a[1]*a[6]*a[5]-a[6]*a[4]*a[2];
		
		if (dst && det)
		{
			T coeff=1/det;
			T a00=+(a[4]*a[8]-a[7]*a[5]); T a01= -(-a[2]*a[7]+a[8]*a[1]); T a02= +(a[1]*a[5]-a[4]*a[2]);
			T a10=-(a[8]*a[3]-a[6]*a[5]); T a11=  +(a[0]*a[8]-a[6]*a[2]); T a12= -(a[5]*a[0]-a[3]*a[2]);
			T a20=+(a[3]*a[7]-a[4]*a[6]); T a21=  -(a[7]*a[0]-a[1]*a[6]); T a22= +(a[0]*a[4]-a[3]*a[1]);

			dst[0]=a00*coeff;dst[1]=a01*coeff;dst[2]=a02*coeff;
			dst[3]=a10*coeff;dst[4]=a11*coeff;dst[5]=a12*coeff;
			dst[6]=a20*coeff;dst[7]=a21*coeff;dst[8]=a22*coeff;
		}
		break;
	}


	case 4:
	{
		 det=
				a[3] * a[6] * a[9]  * a[12]-a[2] * a[7] * a[9]  * a[12]-a[3] * a[5] * a[10] * a[12]+a[1] * a[7]    * a[10] * a[12]+
				a[2] * a[5] * a[11] * a[12]-a[1] * a[6] * a[11] * a[12]-a[3] * a[6] * a[8]  * a[13]+a[2] * a[7]    * a[8]  * a[13]+
				a[3] * a[4] * a[10] * a[13]-a[0] * a[7] * a[10] * a[13]-a[2] * a[4] * a[11] * a[13]+a[0] * a[6]    * a[11] * a[13]+
				a[3] * a[5] * a[8]  * a[14]-a[1] * a[7] * a[8]  * a[14]-a[3] * a[4] * a[9]  * a[14]+a[0] * a[7]    * a[9]  * a[14]+
				a[1] * a[4] * a[11] * a[14]-a[0] * a[5] * a[11] * a[14]-a[2] * a[5] * a[8]  * a[15]+a[1] * a[6]    * a[8]  * a[15]+
				a[2] * a[4] * a[9]  * a[15]-a[0] * a[6] * a[9]  * a[15]-a[1] * a[4] * a[10] * a[15]+a[0] * a[5]    * a[10] * a[15];
			 
		if (dst  && det)
		{
			T a00  = (a[6]*a[11]*a[13] - a[7]*a[10]*a[13] + a[7]*a[9]*a[14] - a[5]*a[11]*a[14] - a[6]*a[9]*a[15] + a[5]*a[10]*a[15]) / det ;
			T a01  = (a[3]*a[10]*a[13] - a[2]*a[11]*a[13] - a[3]*a[9]*a[14] + a[1]*a[11]*a[14] + a[2]*a[9]*a[15] - a[1]*a[10]*a[15]) / det ;
			T a02  = (a[2]*a[7]*a[13]  - a[3]*a[6]*a[13]  + a[3]*a[5]*a[14] - a[1]*a[7]*a[14]  - a[2]*a[5]*a[15] + a[1]*a[6]*a[15])  / det ;
			T a03  = (a[3]*a[6]*a[9]   - a[2]*a[7]*a[9]   - a[3]*a[5]*a[10] + a[1]*a[7]*a[10]  + a[2]*a[5]*a[11] - a[1]*a[6]*a[11])  / det ;
			T a10  = (a[7]*a[10]*a[12] - a[6]*a[11]*a[12] - a[7]*a[8]*a[14] + a[4]*a[11]*a[14] + a[6]*a[8]*a[15] - a[4]*a[10]*a[15]) / det ;
			T a11  = (a[2]*a[11]*a[12] - a[3]*a[10]*a[12] + a[3]*a[8]*a[14] - a[0]*a[11]*a[14] - a[2]*a[8]*a[15] + a[0]*a[10]*a[15]) / det ;
			T a12  = (a[3]*a[6]*a[12]  - a[2]*a[7]*a[12]  - a[3]*a[4]*a[14] + a[0]*a[7]*a[14]  + a[2]*a[4]*a[15] - a[0]*a[6]*a[15])  / det ;
			T a13  = (a[2]*a[7]*a[8]   - a[3]*a[6]*a[8]   + a[3]*a[4]*a[10] - a[0]*a[7]*a[10]  - a[2]*a[4]*a[11] + a[0]*a[6]*a[11])  / det ;
			T a20  = (a[5]*a[11]*a[12] - a[7]*a[9]*a[12]  + a[7]*a[8]*a[13] - a[4]*a[11]*a[13] - a[5]*a[8]*a[15] + a[4]*a[9]*a[15])  / det ;
			T a21  = (a[3]*a[9]*a[12]  - a[1]*a[11]*a[12] - a[3]*a[8]*a[13] + a[0]*a[11]*a[13] + a[1]*a[8]*a[15] - a[0]*a[9]*a[15])  / det ;
			T a22  = (a[1]*a[7]*a[12]  - a[3]*a[5]*a[12]  + a[3]*a[4]*a[13] - a[0]*a[7]*a[13]  - a[1]*a[4]*a[15] + a[0]*a[5]*a[15])  / det ;
			T a23  = (a[3]*a[5]*a[8]   - a[1]*a[7]*a[8]   - a[3]*a[4]*a[9]  + a[0]*a[7]*a[9]   + a[1]*a[4]*a[11] - a[0]*a[5]*a[11])  / det ;
			T a30  = (a[6]*a[9]*a[12]  - a[5]*a[10]*a[12] - a[6]*a[8]*a[13] + a[4]*a[10]*a[13] + a[5]*a[8]*a[14] - a[4]*a[9]*a[14])  / det ;
			T a31  = (a[1]*a[10]*a[12] - a[2]*a[9]*a[12]  + a[2]*a[8]*a[13] - a[0]*a[10]*a[13] - a[1]*a[8]*a[14] + a[0]*a[9]*a[14])  / det ;
			T a32  = (a[2]*a[5]*a[12]  - a[1]*a[6]*a[12]  - a[2]*a[4]*a[13] + a[0]*a[6]*a[13]  + a[1]*a[4]*a[14] - a[0]*a[5]*a[14])  / det ;
			T a33  = (a[1]*a[6]*a[8]   - a[2]*a[5]*a[8]   + a[2]*a[4]*a[9]  - a[0]*a[6]*a[9]   - a[1]*a[4]*a[10] + a[0]*a[5]*a[10])  / det ;

			dst[ 0]=a00;dst[ 1]=a01;dst[ 2]=a02;dst[ 3]=a03;
			dst[ 4]=a10;dst[ 5]=a11;dst[ 6]=a12;dst[ 7]=a13;
			dst[ 8]=a20;dst[ 9]=a21;dst[10]=a22;dst[11]=a23;
			dst[12]=a30;dst[13]=a31;dst[14]=a32;dst[15]=a33; 
		}

		break;
	}

	default:
	{
		int signum,r,c;
		gsl_permutation *p=gsl_permutation_alloc(dim);

		gsl_matrix* lu =gsl_matrix_calloc(dim,dim);
		gsl_matrix* inv=gsl_matrix_calloc(dim,dim);

		for (r=0;r<dim;r++)
		for (c=0;c<dim;c++)
			gsl_matrix_set(lu,r,c,a[r*(dim)+c]);

		gsl_linalg_LU_decomp (lu, p, &signum);
		det=(T)gsl_linalg_LU_det(lu,signum);
		gsl_linalg_LU_invert (lu, p, inv);
		gsl_permutation_free(p);

		if (dst  && det)
		{
			for (r=0;r<dim;r++)
			for (c=0;c<dim;c++)
				dst[r*(dim)+c]=(T)gsl_matrix_get(inv,r,c);
		}

		gsl_matrix_free(lu );
		gsl_matrix_free(inv);

		break;
	}
	}

	
	return det;
}

template <class T>
inline void tmat<T>::swap_rows(int dim,T* mat,const int* perm)
{
	T tmp[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];

	for (int r=0;r<dim;r++)
	{
		for (int c=0;c<dim;c++)
		{
			tmp[r*dim+c]=mat[perm[r]*dim+c];
		}
	}

	memcpy(mat,tmp,sizeof(T)*dim*dim);
}

template <class T>
inline void tmat<T>::swap_rows(int dim,T* mat,int r1,int r2)
{
	if (r1==r2) return;

	T* row1=mat+r1*dim;
	T* row2=mat+r2*dim;

	for (int c=0;c<dim;c++)
		Swap(T,row1[c],row2[c]);
}


template <class T>
inline void tmat<T>::swap_cols(int dim,T* mat,const int* perm)
{
	T tmp[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];

	for (int r=0;r<dim;r++)
	{
		for (int c=0;c<dim;c++)
		{
			tmp[r*dim+c]=mat[r*dim+perm[c]];
		}
	}

	memcpy(mat,tmp,sizeof(T)*dim*dim);
}

template <class T>
inline void tmat<T>::swap_cols(int dim,T* mat,int c1,int c2)
{
	if (c1==c2) return;
	T* col1=mat+c1;
	T* col2=mat+c2;

	for (int r=0;r<dim;r++,col1+=dim,col2+=dim)
		Swap(T,*col1,*col2);
}



template <class T>
inline void tmat<T>::copy(int dstdim,T* dst,int srcdim,const T* src)
{
	T tmp[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];

	tmat<T>::identity(dstdim,tmp);

	for (int r=0;r<srcdim;r++)
	{
		for (int c=0;c<srcdim;c++)
		{
			tmp[r*dstdim+c]=src[r*srcdim+c];
		}
	}

	memcpy(dst,tmp,sizeof(T)*dstdim*dstdim);
}


template <class T>
void tmat<T>::embed(int dstdim,int srcdim,T*& mat)
{
	if (dstdim==srcdim)
		return;

	checkpoint(dstdim>srcdim);

	T* tmp=new real[dstdim*dstdim];

	tmat<T>::identity(dstdim,tmp);

	for (int r=0;r<srcdim;r++)
	{
		for (int c=0;c<srcdim;c++)
		{
			tmp[r*dstdim+c]=mat[r*srcdim+c];
		}
	}

	Swap(T*,tmp,mat);
	delete [] tmp;
}


template <class T>
inline void tmat<T>::identity(int dim,T* ret)
{
	assert(dim);
	memset(ret,0,sizeof(T)*dim*dim);

	switch(dim)
	{
	case 1:
		ret[0]=1;
		break;
	case 2:
		ret[0]=ret[3]=1;
		break;
	case 3:
		ret[0]=ret[4]=ret[8]=1;
		break;
	case 4:
		ret[0]=ret[5]=ret[10]=ret[15]=1;
		break;

	default:
		for (int i=0;i<dim;i++) 
			ret[i*dim+i]=(T)1.0;
		break;
	}
}

template <class T>
inline void tmat<T>::scale(int dim,int ref,T factor,T* vmat,T* hmat)
{
	if (vmat)
	{
		identity(dim,vmat);
		vmat[ref*dim+ref]=factor;
	}

	if (hmat)
	{
		identity(dim,hmat);
		hmat[ref*dim+ref]=1/factor;
	}
}

template <class T>
inline void tmat<T>::translate(int dim,int ref,T factor,T* vmat,T* hmat)
{
	if (vmat)
	{
		identity(dim,vmat);
		vmat[ref*dim]=+factor;
	}

	if (hmat)
	{
		identity(dim,hmat);
		hmat[ref*dim]=-factor;
	}	
}

template <class T>
inline void tmat<T>::rotate(int dim,int i,int j,T angle,T* vmat,T* hmat)
{
	T c=(T)cos(angle),s=(T)sin(angle);

	if (vmat)
	{
		identity(dim,vmat);
		vmat[i*dim+i]=+c;vmat[i*dim+j]=+s;
		vmat[j*dim+i]=-s;vmat[j*dim+j]=+c;
	}

	if (hmat)
	{
		identity(dim,hmat);
		hmat[i*dim+i]=+c;hmat[i*dim+j]=-s; 
		hmat[j*dim+j]=+c;hmat[j*dim+i]=+s;	
	}	
}

template <class T>
inline void tmat<T>::shear(int dim,int ref,const T* factor,T* vmat,T* hmat)
{
	if (vmat)
	{
		identity(dim,vmat);
		for (int i=1;i<dim;i++) 
			vmat[ref*dim+i]=factor[i-1];
	}


	if (hmat)
	{
		identity(dim,hmat);
		for (int i=1;i<dim;i++) 
			hmat[ref*dim+i]=(i==ref)?(1/factor[ref-1]):(-factor[i-1]/factor[ref-1]);
	}	
}
	


template <class T>
inline void tmat<T>::submatrix(T* dst,int dim,const T* src,int r1,int r2,int c1,int c2)
{
	T _ret[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];

	int n=0;

	for (int row=r1;row<=r2;row++)
	for (int col=c1;col<=c2;col++)
		_ret[n++]=src[row*dim+col];

	memcpy(dst,_ret,sizeof(T)*n);
}




template <class T>
inline void tmat<T>::zero(int dim,T* ret)
{
	memset(ret,0,sizeof(T)*dim*dim);
}




template <class T>
inline void tmat<T>::scalarproduct(int dim,T* dest,const T* src,const float coeff)
{
	switch (dim)
	{
	case 1:
		dest[0]=src[0]*coeff;
		break;
	case 2:
		dest[0]=src[0]*coeff;dest[1]=src[1]*coeff;
		dest[2]=src[2]*coeff;dest[3]=src[3]*coeff;
		break;
	case 3:
		dest[0]=src[0]*coeff;dest[1]=src[1]*coeff;dest[2]=src[2]*coeff;
		dest[3]=src[3]*coeff;dest[4]=src[4]*coeff;dest[5]=src[5]*coeff;
		dest[6]=src[6]*coeff;dest[7]=src[7]*coeff;dest[8]=src[8]*coeff;
		break;
	case 4:
		dest[ 0]=src[ 0]*coeff;dest[ 1]=src[ 1]*coeff;dest[ 2]=src[ 2]*coeff;dest[ 3]=src[ 3]*coeff;
		dest[ 4]=src[ 4]*coeff;dest[ 5]=src[ 5]*coeff;dest[ 6]=src[ 6]*coeff;dest[ 7]=src[ 7]*coeff;
		dest[ 8]=src[ 8]*coeff;dest[ 9]=src[ 9]*coeff;dest[10]=src[10]*coeff;dest[11]=src[11]*coeff;
		dest[12]=src[12]*coeff;dest[13]=src[13]*coeff;dest[14]=src[14]*coeff;dest[15]=src[15]*coeff;
		break;
	default:
		for (int i=0;i<dim;++i)
		for (int j=0;j<dim;++j) 
			dest[i*(dim)+j]=src[i*(dim)+j]*coeff;
		break;
	}
}


template <class T>
inline void tmat<T>::transpose(int dim,T* dest,const T* src)
{
	int i,j;

	if (dest!=src)
		memcpy(dest,src,sizeof(T)*dim*dim);

	switch(dim)
	{
	case 1:
		break;
	case 2:
		Swap(T,dest[1],dest[2]);
		break;
	case 3:
		Swap(T,dest[1],dest[3]);
		Swap(T,dest[2],dest[6]);
		Swap(T,dest[5],dest[7]);
		break;
	case 4:
		Swap(T,dest[ 1],dest[ 4]);
		Swap(T,dest[ 2],dest[ 8]);
		Swap(T,dest[ 6],dest[ 9]);
		Swap(T,dest[ 3],dest[12]);
		Swap(T,dest[ 7],dest[13]);
		Swap(T,dest[11],dest[14]);
		break;

	default:
		for (i=0;i<dim;++i)
		for (j=0;j<dim;++j)
		{
			Swap(T,dest[j*(dim)+i],dest[i*(dim)+j]);
		}

		break;
	}
}


template <class Tv,class Tm>
inline void productv(int dim,Tv* dst,const Tm* a,const Tv* v)
{
	switch(dim)
	{
	case 1:
		{
			Tv A0=(Tv)(a[0]*v[0]);
			dst[0]=A0;
			return;
		}

	case 2:
		{
			Tv A0=(Tv)(a[0]*v[0]+a[1]*v[1]);
			Tv A1=(Tv)(a[2]*v[0]+a[3]*v[1]);
			dst[0]=A0;dst[1]=A1;
			return;
		}

	case 3:
		{
			Tv A0=(Tv)(a[0]*v[0]+a[1]*v[1]+a[2]*v[2]);
			Tv A1=(Tv)(a[3]*v[0]+a[4]*v[1]+a[5]*v[2]);
			Tv A2=(Tv)(a[6]*v[0]+a[7]*v[1]+a[8]*v[2]);
			dst[0]=A0;dst[1]=A1;dst[2]=A2;
			return;
		}

	case 4:
		{
			Tv A0=(Tv)(a[ 0]*v[0]+a[ 1]*v[1]+a[ 2]*v[2]+a[ 3]*v[3]);
			Tv A1=(Tv)(a[ 4]*v[0]+a[ 5]*v[1]+a[ 6]*v[2]+a[ 7]*v[3]);
			Tv A2=(Tv)(a[ 8]*v[0]+a[ 9]*v[1]+a[10]*v[2]+a[11]*v[3]);
			Tv A3=(Tv)(a[12]*v[0]+a[13]*v[1]+a[14]*v[2]+a[15]*v[3]);
			dst[0]=A0;dst[1]=A1;dst[2]=A2;dst[3]=A3;
			return;
		}

	default:
		{
			Tv _ret[(MAX_GEOMETRIC_DIMENSION+1)];
			memset(_ret,0,sizeof(_ret));
			const Tm* p=a;
			
			for (int r=0;r<dim;r++)
			for (int c=0;c<dim;c++)
				_ret[r]+=(Tv)((*p++)*v[c]);
				
			memcpy(dst,_ret,sizeof(Tv)*dim);
		}
	}
}


template <class T>
inline void tmat<T>::add(int dim,T* dst,const T* a,const T *b)
{
	for (int i=0;i<dim;i++)
	for (int j=0;j<dim;j++)
	{
		dst[i*dim+j]=a[i*dim+j]+b[i*dim+j];
	}
}

template <class T>
inline void tmat<T>::subtract(int dim,T* dst,const T* a,const T *b)
{
	for (int i=0;i<dim;i++)
	for (int j=0;j<dim;j++)
	{
		dst[i*dim+j]=a[i*dim+j]-b[i*dim+j];
	}
}

template <class T>
inline void tmat<T>::product(int dim,T* ret,const T* a,const T* b)
{
	switch(dim)
	{
	case 1:
	{
		ret[0]=a[0]*b[0];
		break;
	}
	case 2:
	{
		T a00 = a[0]*b[0]+a[1]*b[2]; 
		T a01 = a[0]*b[1]+a[1]*b[3]; 
		T a10 = a[2]*b[0]+a[3]*b[2]; 
		T a11 = a[2]*b[1]+a[3]*b[3];
		ret[0]=a00;ret[1]=a01;
		ret[2]=a10;ret[3]=a11;
		break;
	}
	case 3:
	{
		T a00= a[0]*b[0]+a[1]*b[3]+a[2]*b[6]; 
		T a01= a[0]*b[1]+a[1]*b[4]+a[2]*b[7]; 
		T a02= a[0]*b[2]+a[1]*b[5]+a[2]*b[8];
		T a10= a[3]*b[0]+a[4]*b[3]+a[5]*b[6]; 
		T a11= a[3]*b[1]+a[4]*b[4]+a[5]*b[7]; 
		T a12= a[3]*b[2]+a[4]*b[5]+a[5]*b[8]; 
		T a20= a[6]*b[0]+a[7]*b[3]+a[8]*b[6]; 
		T a21= a[6]*b[1]+a[7]*b[4]+a[8]*b[7]; 
		T a22= a[6]*b[2]+a[7]*b[5]+a[8]*b[8];
		ret[0]=a00;ret[1]=a01;ret[2]=a02;
		ret[3]=a10;ret[4]=a11;ret[5]=a12;
		ret[6]=a20;ret[7]=a21;ret[8]=a22;
		break;
	}
	case 4:
	{
		T a00=a[ 0]*b[ 0]+a[ 1]*b[ 4]+a[ 2]*b[ 8]+a[ 3]*b[12]; 
		T a01=a[ 0]*b[ 1]+a[ 1]*b[ 5]+a[ 2]*b[ 9]+a[ 3]*b[13]; 
		T a02=a[ 0]*b[ 2]+a[ 1]*b[ 6]+a[ 2]*b[10]+a[ 3]*b[14]; 
		T a03=a[ 0]*b[ 3]+a[ 1]*b[ 7]+a[ 2]*b[11]+a[ 3]*b[15]; 
		T a10=a[ 4]*b[ 0]+a[ 5]*b[ 4]+a[ 6]*b[ 8]+a[ 7]*b[12]; 
		T a11=a[ 4]*b[ 1]+a[ 5]*b[ 5]+a[ 6]*b[ 9]+a[ 7]*b[13]; 
		T a12=a[ 4]*b[ 2]+a[ 5]*b[ 6]+a[ 6]*b[10]+a[ 7]*b[14]; 
		T a13=a[ 4]*b[ 3]+a[ 5]*b[ 7]+a[ 6]*b[11]+a[ 7]*b[15]; 
		T a20=a[ 8]*b[ 0]+a[ 9]*b[ 4]+a[10]*b[ 8]+a[11]*b[12]; 
		T a21=a[ 8]*b[ 1]+a[ 9]*b[ 5]+a[10]*b[ 9]+a[11]*b[13]; 
		T a22=a[ 8]*b[ 2]+a[ 9]*b[ 6]+a[10]*b[10]+a[11]*b[14]; 
		T a23=a[ 8]*b[ 3]+a[ 9]*b[ 7]+a[10]*b[11]+a[11]*b[15]; 
		T a30=a[12]*b[ 0]+a[13]*b[ 4]+a[14]*b[ 8]+a[15]*b[12]; 
		T a31=a[12]*b[ 1]+a[13]*b[ 5]+a[14]*b[ 9]+a[15]*b[13]; 
		T a32=a[12]*b[ 2]+a[13]*b[ 6]+a[14]*b[10]+a[15]*b[14]; 
		T a33=a[12]*b[ 3]+a[13]*b[ 7]+a[14]*b[11]+a[15]*b[15];
		ret[ 0]=a00;ret[ 1]=a01;ret[ 2]=a02;ret[ 3]=a03;
		ret[ 4]=a10;ret[ 5]=a11;ret[ 6]=a12;ret[ 7]=a13;
		ret[ 8]=a20;ret[ 9]=a21;ret[10]=a22;ret[11]=a23;
		ret[12]=a30;ret[13]=a31;ret[14]=a32;ret[15]=a33;
		break;
	}

	default:
	{
		int i,j,k;

		T _ret[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
		memset(_ret,0,sizeof(T)*dim*dim);

		for (i=0;i<dim;++i)
		for (j=0;j<dim;++j)
		for (k=0;k<dim;++k)
			_ret[i*(dim)+j]+=a[i*(dim)+k]*b[k*(dim)+j];
				
		memcpy(ret,_ret,sizeof(T)*dim*dim);
	}
	}
}



template <class T>
inline void tmat<T>::print(int dim,const T* mat,FILE* stream)
{
	int i,j;
	fprintf(stream,"array([");
	for (i=0;i<dim;++i)
	{
		if (i) fprintf(stream," , ");
		fprintf(stream,"[");
		for (j=0;j<dim;++j)
		{
			if (j) printf(",");
			//fprintf(stream,"%le ",(dec64)mat[i*(dim)+j]);
			fprintf(stream,"%f ",(dec64)mat[i*(dim)+j]);
		}
		fprintf(stream,"]");
		
	}
	fprintf(stream,"])\n");
}


template <class T>
inline bool tmat<T>::isInvert(int dim,const T* Vmat,const T* Hmat,real tolerance)
{
	real tmp[(MAX_GEOMETRIC_DIMENSION+1)*(MAX_GEOMETRIC_DIMENSION+1)];
	tmat<real>::product(dim,tmp,Vmat,Hmat);

	for (int i=0;i<dim;i++)
	for (int j=0;j<dim;j++)
	{
		real val=tmp[i*dim+j];

		if (i!=j && fabs(val)>tolerance)
			return false;

		if (i==j && fabs(1-val)>tolerance)
			return false;
	}

	return true;
}


#endif /* _MAT_H__ */


