#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <string.h>

#include "pipeline3d.h"

#define LOG(msg)          do {fprintf(stdout,"%s"  ,msg);fflush(stdout);} while (0);
#define LOG2(format,arg)  do {fprintf(stdout,format,arg);fflush(stdout);} while (0);

#define DIE(msg)          do {fprintf(stderr,"%s"  ,msg);exit(-1);} while (0);
#define DIE2(format,arg)  do {fprintf(stderr,format,arg);fflush(stderr);exit(-1);} while (0);


#define X 0
#define Y 1
#define Z 2
#define W 3

#define CUT_A 0
#define CUT_B 1
#define CUT_C 2
#define CUT_D 3

#define TRUE 1
#define FALSE 0

double vrp1z;
double zmin;

#define VIEWPORT_ZMIN -1600
#define VIEWPORT_ZMAX 1600

typedef struct
{
    double m[4][4];
} 
p3d_matrix;

typedef struct 
{
    double v[4];
} 
p3d_vertex ;

typedef struct
{
    double v[4];
} 
p3d_vector;



typedef struct
{
 	double xmin;
 	double xmax;
 	double ymin;
 	double ymax;
 	double zmin;
 	double zmax;
} 
p3d_viewPort;


typedef struct
{
 	double umin;
 	double umax;
 	double vmin;
 	double vmax;
} 
p3d_box;

typedef struct
{
	p3d_vertex* vrp;
	p3d_vector* vuv;
	p3d_vector* vpn;
	p3d_vector* dop;
	p3d_vertex* cop;
	p3d_vertex* prp;
	double front;
	double back;
	p3d_box    mybox;
} 
p3d_viewModel;

static p3d_vector* create3Vector(double x, double y, double z);
static p3d_vertex* create3Vertex(double x, double y, double z);

static void freeVertex(p3d_vertex* v);
static void freeVector(p3d_vector* v);
static void freeMatrix(p3d_matrix* m);

static p3d_vertex* matVertProductReturn(p3d_matrix* mat,p3d_vertex* v); // mat*v rows per columns product
static p3d_matrix* matMatProductReturn(p3d_matrix* left,p3d_matrix* right); // left*right rows per columns product

static p3d_matrix* parallelViewMapping    (p3d_viewModel* view);
static p3d_matrix* perspectiveViewMapping (p3d_viewModel* view);
static p3d_matrix* viewOrientation        (p3d_viewModel* view);
static p3d_matrix* deviceMapping          (p3d_viewPort* viewport, p3d_viewModel* view);
static p3d_matrix* perspTransf            (p3d_viewModel* view, double vrp_z);

//_________________________________________________________________
//_________________________________________________________________
int makePipeline3d
(
	float Tv[16],float observer[3],
	const char* type,
	const double vrp[3],const double vuv[3],const double vpn[3],const double cop[3],
	const double front,const double back,
	const double range_u[2],const double range_v[2],const double range_x[2],const double range_y[2]
)
{
	p3d_matrix* m;
	p3d_viewModel* vm = (p3d_viewModel*)malloc(sizeof(p3d_viewModel));
	p3d_box*       bb = (p3d_box*)      malloc(sizeof(p3d_box));
	p3d_viewPort*  vp = (p3d_viewPort*) malloc(sizeof(p3d_viewPort));
	
	vm->vrp = create3Vertex(vrp[0],vrp[1],vrp[2]);
	vm->vuv = create3Vector(vuv[0],vuv[1],vuv[2]);
	vm->vpn = create3Vector(vpn[0],vpn[1],vpn[2]);
	vm->prp = NULL;
	vm->dop = NULL;
	vm->cop = create3Vertex(cop[0],cop[1],cop[2]);
	vm->front = front;
	vm->back = back;
	
	bb->umin = range_u[0];bb->umax = range_u[1];
	bb->vmin = range_v[0];bb->vmax = range_v[1];
	vm->mybox = *bb;	
	
	vp->xmin = range_x[0];vp->xmax = range_x[1];
	vp->ymin = range_y[0];vp->ymax = range_y[1];
	vp->zmin = VIEWPORT_ZMIN;
	vp->zmax = VIEWPORT_ZMAX;	
	
	/* Viene avviata la pipeline parallela */
	if(type && !strcmp(PARALLEL,type))
	{
		// Calcolo il prp
		p3d_vertex* prp = matVertProductReturn(viewOrientation(vm),vm->cop);
		vm->prp=prp;
		// Calcolo matrice di trasformazione WC3 -> NDC
		p3d_matrix* m1 = matMatProductReturn(parallelViewMapping(vm),viewOrientation(vm));
		// Calcolo matrice di trasformazione WC3 -> DC3
		m=matMatProductReturn(deviceMapping(vp, vm),m1);
		freeMatrix(m1);
		freeVertex(prp);
	}		
	/* Viene avviata la pipeline prospettica */
	else
	{
		p3d_vertex* prp=matVertProductReturn(viewOrientation(vm),vm->cop);
		vm->prp=prp;
		p3d_matrix* m1 = matMatProductReturn(perspectiveViewMapping(vm),viewOrientation(vm));
		p3d_matrix* m2 = matMatProductReturn(perspTransf(vm, vrp1z),m1);
		m=matMatProductReturn(deviceMapping(vp, vm),m2);
		freeMatrix(m1);
		freeMatrix(m2);
		freeVertex(prp);
	}

	for (int riga=0;riga<4;riga++)
	for (int colonna=0;colonna<4;colonna++)
		Tv[riga*4+colonna]=(float)m->m[riga][colonna];

	observer[0]=(float)vm->cop->v[0];
	observer[1]=(float)vm->cop->v[1];
	observer[2]=(float)vm->cop->v[2];

	freeMatrix(m);
	return 1; // OK
}
//_________________________________________________________________
//_________________________________________________________________


/* Funzioni geometriche di base */

/**
Vector constructor.
Create a p3d_vector. It accepts x,y and z components, w component is set to 1.
@param x the x component of the p3d_vector
@param y the y component of the p3d_vector
@param z the z component of the p3d_vector
@return a pointer to the new p3d_vector structure
 */
static p3d_vector* create3Vector(double x, double y, double z)
{
        p3d_vector* out = (p3d_vector*)malloc(sizeof(p3d_vector));
        if(!out) { DIE("Cannot allocate p3d_vector\n"); }
        out->v[X] = x;
        out->v[Y] = y;
        out->v[Z] = z;
        out->v[W] = 0.0;
        return out;
}

/**
Vector constructor.
Create a degenerate p3d_vector. All components are set to 0, except w which is set to 1.
@return a pointer to the new p3d_vector structure
 */
static p3d_vector* createEmptyVector()
{
        p3d_vector* out = (p3d_vector*)malloc(sizeof(p3d_vector));
        if(!out) { DIE("Cannot allocate p3d_vector\n"); }
        out->v[X] = 0;
        out->v[Y] = 0;
        out->v[Z] = 0;
        out->v[W] = 0.0;
        return out;
}

/**
Vector constructor.
Create a p3d_vector which is a fresh copy of the p3d_vector passed as parameter.
@param v the p3d_vector to copy
@return a pointer to the new p3d_vector structure
 */
static p3d_vector* cloneVector(p3d_vector* v)
{
        p3d_vector* out = createEmptyVector();
        memcpy(out,v,sizeof(p3d_vector));
        return out;
}

/**
Vector destructor.
Destroy a p3d_vector. If v is not NULL, it call free() on it.
@param v the p3d_vector to deallocate
*/
static void freeVector(p3d_vector* v)
{
     if(v) free(v);
}

/**
Vertex constructor.
Create a p3d_vertex. It accepts x,y,z and w components.
@param x the x component of the p3d_vertex
@param y the y component of the p3d_vertex
@param z the z component of the p3d_vertex
@param w the w component of the p3d_vertex
@return a pointer to the new p3d_vertex structure
*/
static p3d_vertex* create4Vertex(double x, double y, double z, double w)
{
        p3d_vertex* out = (p3d_vertex*)malloc(sizeof(p3d_vertex));
        if(!out) { DIE("Cannot allocate p3d_vertex\n"); }
        out->v[X] = x/w;
        out->v[Y] = y/w;
        out->v[Z] = z/w;
        out->v[W] = 1.0;
        return out;
}

/**
Vertex constructor.
Create a p3d_vertex. It accepts x,y, and z components, w is set to 1.
@param x the x component of the p3d_vertex
@param y the y component of the p3d_vertex
@param z the z component of the p3d_vertex
@return a pointer to the new p3d_vertex structure
*/
p3d_vertex* create3Vertex(double x, double y, double z)
{
        return create4Vertex(x,y,z,1.0);
}

/**
Vertex constructor.
Create a p3d_vertex in the origin (at 0,0,0,1).
@return a pointer to the new p3d_vertex structure
*/
static p3d_vertex* createEmptyVertex()
{
        return create4Vertex(0.0,0.0,0.0,1.0);
}

/**
Vertex constructor.
Create a p3d_vertex which is a fresh copy of the parameter p3d_vertex.
@param v p3d_vertex to copy
@return a pointer to the new p3d_vertex structure
*/
static p3d_vertex* cloneVertex(p3d_vertex* v)
{
        p3d_vertex* out = createEmptyVertex();
        memcpy(out,v,sizeof(p3d_vertex));
        return out;
}

/**
Vertex destructor.
Deallocate a p3d_vertex. Call free on it if v is not NULL.
@param v the p3d_vertex to deallocate
*/
void freeVertex(p3d_vertex* v)
{
     if(v) free(v);
}




/**
Empty-p3d_matrix constructor.
Create an all-zero p3d_matrix.
@return a pointer to the new p3d_matrix
*/
static p3d_matrix* createEmptyMatrix()
{
        p3d_matrix* out = (p3d_matrix*)malloc(sizeof(p3d_matrix));
        if(!out) { DIE("Cannot allocare p3d_matrix\n"); }
        memset(out,0,sizeof(p3d_matrix));
        return out;
}

/**
Identity p3d_matrix constructor.
Create an identity p3d_matrix.
@return a pointer to the new identity p3d_matrix
*/
static p3d_matrix* createIdentityMatrix()
{
        p3d_matrix* out = createEmptyMatrix();
        out->m[0][0] = 1;
        out->m[1][1] = 1;
        out->m[2][2] = 1;
        out->m[3][3] = 1;
        return out;
}
/**
Scale transformation p3d_matrix constructor.
Create a p3d_matrix which perform a scaling transformation.
@param x_scale scale factor for x axis
@param y_scale scale factor for y axis
@param z_scale scale factor for z axis
@return a pointer to the new scale p3d_matrix
*/
static p3d_matrix* createScalingMatrix(double x_scale,double y_scale, double z_scale)
{
        p3d_matrix* out = createIdentityMatrix();
        out->m[0][0] = x_scale;
        out->m[1][1] = y_scale;
        out->m[2][2] = z_scale;
        return out;
}

/**
Shearing transformation p3d_matrix constructor.
Create a p3d_matrix which perform a shearing transformation.
@param axis the axis normal to the shearing plane
@param a_shear first shear factor
@param b_shear second shear factor
@return a pointer to the new shearing p3d_matrix
*/
static p3d_matrix* createShearingMatrix(int axis, double a_shear, double b_shear)
{
        p3d_matrix* out = createIdentityMatrix();
        if(axis==X)
        {
               out->m[1][0] = a_shear;
               out->m[2][0] = b_shear;
        }
        else if(axis==Y)
        {
               out->m[0][1] = a_shear;
               out->m[2][1] = b_shear;
        } else if(axis==Z)
        {
               out->m[0][2] = a_shear;
               out->m[1][2] = b_shear;
        } else {
               LOG2("createShearingMatrix wrong axis - %d",axis);
        }
        return out;
}

/**
Rotation on X-axis p3d_matrix constructor.
Create a p3d_matrix which perform a rotation on the x axis
@param angle rotation angle
@return a pointer to the new rotation p3d_matrix
*/
static p3d_matrix* createXRotationMatrix(double angle)
{
     p3d_matrix* out = createIdentityMatrix();
     out->m[1][1] = cos(angle);
     out->m[1][2] = - sin(angle);
     out->m[2][1] = sin(angle);
     out->m[2][2] = cos(angle);
     return out;
}

/**
Rotation on Y-axis p3d_matrix constructor.
Create a p3d_matrix which perform a rotation on the x axis
@param angle rotation angle
@return a pointer to the new rotation p3d_matrix
*/
static p3d_matrix* createYRotationMatrix(double angle)
{
     p3d_matrix* out = createIdentityMatrix();
     out->m[0][0] = cos(angle);
     out->m[0][2] = sin(angle);
     out->m[2][0] = - sin(angle);
     out->m[2][2] = cos(angle);
     return out;
}

/**
Rotation on Z-axis p3d_matrix constructor.
Create a p3d_matrix which perform a rotation on the z axis
@param angle rotation angle
@return a pointer to the new rotation p3d_matrix
*/
static p3d_matrix* createZRotationMatrix(double angle)
{
     p3d_matrix* out = createIdentityMatrix();
     out->m[0][0] = cos(angle);
     out->m[0][1] = - sin(angle);
     out->m[1][0] = sin(angle);
     out->m[1][1] = cos(angle);
     return out;
}

/**
Translation transformation p3d_matrix constructor.
Create a p3d_matrix which perform a translation transformation specified by a p3d_vector
@param translation_vector the translation p3d_vector
@return a pointer to the new translation p3d_matrix
*/
static p3d_matrix* createTranslationMatrix(p3d_vector* translation_vector)
{
        p3d_matrix* out = createIdentityMatrix();
        out->m[0][3] = translation_vector->v[X];
        out->m[1][3] = translation_vector->v[Y];
        out->m[2][3] = translation_vector->v[Z];
        return out;
}

/**
Translation transformation p3d_matrix constructor.
Create a p3d_matrix which perform a translation transformation specified by x,y and z.
@param x x component of the translation
@param y y component of the translation
@param z z component of the translation
@return a pointer to the new translation p3d_matrix
*/
static p3d_matrix* createTranslationMatrixf(double x, double y, double z)
{
        p3d_matrix* out = createIdentityMatrix();
        out->m[0][3] = x;
        out->m[1][3] = y;
        out->m[2][3] = z;
        return out;
}

/**
Matrix constructor.
Build a p3d_matrix. It creates a fresh copy of the parameter p3d_matrix.
@param m the p3d_matrix to copy
@return a pointer to the new p3d_matrix
*/
static p3d_matrix* cloneMatrix(p3d_matrix* m)
{
        p3d_matrix* out = createEmptyMatrix();
        memcpy(out,m,sizeof(p3d_matrix));
        return out;
}

/**
Matrix destructors.
Deallocate a p3d_matrix. Call free on it if m is not NULL.
@param m the p3d_matrix to deallocate
*/
static void freeMatrix(p3d_matrix* m)
{
        if(m) free(m);
}

/**
Matrix copy function.
Copy p3d_matrix values into an other.
@param source the source of the copy action
@param dest the destination of the copy action
*/
static void p3d_matrixCopy(p3d_matrix* source, p3d_matrix* dest)
{
        memcpy(dest,source,sizeof(p3d_matrix));
}

/**
Vector copy function.
Copy p3d_vector values into an other.
@param source the source of the copy action
@param dest the destination of the copy action
*/
static void vectorCopy(p3d_vector* source, p3d_vector* dest)
{
      memcpy(dest,source,sizeof(p3d_vector));
}

/**
Vertex copy function.
Copy p3d_vertex values into an other.
@param source the source of the copy action
@param dest the destination of the copy action
*/
static void vertexCopy(p3d_vertex* source, p3d_vertex* dest)
{
      memcpy(dest,source,sizeof(p3d_vertex));
}



/* Basic operation */
/* per le funzioni in_*, ricordarsi che il parametro di uscita out potrebbe essere un puntatore
 * ad uno stesso parametro di ingresso. Conviene usare le funzioni di utility per:
 * - istanziare una nuova struct temporanea
 * - usarla per i calcoli
 * - copiare i risultati nel puntatore d'uscita
 * - deallocare la struct temporanea usata
 */

/**
 * Vertex normalization function.
 * Normalize the p3d_vertex (the component W has to be==1)
 * @param vert a pointer to the p3d_vertex to normalize
 * @return a pointer to a fresh normalized p3d_vertex
 */
static p3d_vertex* normalizeVertexClone(p3d_vertex* v)
{
        p3d_vertex* out = createEmptyVertex();
        out->v[X] = v->v[X] / v->v[W];
        out->v[Y] = v->v[Y] / v->v[W];
        out->v[Z] = v->v[Z] / v->v[W];
        return out;
}

/**
 * Vertex normalization function.
 * Normalize the p3d_vertex (the component W has to be==1)
 * @param vert a pointer to the p3d_vertex to normalize
 * @return a pointer to a fresh normalized p3d_vertex
 */
static void normalizeVertex(p3d_vertex* v)
{
        v->v[X] = v->v[X] / v->v[W];
        v->v[Y] = v->v[Y] / v->v[W];
        v->v[Z] = v->v[Z] / v->v[W];
        v->v[W] = 1;
}

/**
 * Calculate the norm of a p3d_vector.
 * @param v a pointer to the p3d_vector of which we have to calculate the norm
 * @return the p3d_vector's norm
 */
static double norm(p3d_vector* v)
{
 	 return sqrt(pow(v->v[X],2.0)+pow(v->v[Y],2.0)+pow(v->v[Z],2.0));
}

/**
 * Normalize the p3d_vector.
 * @param v a pointer to the p3d_vector to normalize
 */
static void normalizeVector(p3d_vector* v)
{
 	 double n=norm(v);
 	 v->v[X]=v->v[X]/n;
 	 v->v[Y]=v->v[Y]/n;
 	 v->v[Z]=v->v[Z]/n;
}

/**
 * Normalize the p3d_vector.
 * @param v a pointer to the p3d_vector to normalize
 * @return the pointer to a fresh normalized p3d_vector
 */
static p3d_vector* normalizeVectorClone(p3d_vector* v)
{
 	double n=norm(v);
 	return create3Vector(v->v[X]/n,v->v[Y]/n,v->v[Z]/n);
}

/**
 *
 * Calculate the p3d_vector that derives from the difference between two points.
 * @param left pointer to the left p3d_vertex
 * @param right pointer to the right p3d_vertex
 * @param out pointer to the resultant p3d_vector
 */
static void sub(p3d_vertex* left, p3d_vertex* right, p3d_vector* out)
{
 	out->v[X]=left->v[CUT_A]-right->v[CUT_A];
 	out->v[Y]=left->v[CUT_B]-right->v[CUT_B];
 	out->v[Z]=left->v[CUT_C]-right->v[CUT_C];
 	out->v[W]=0.0;
}

/**
 * Calculate the p3d_vector that derives from the differnce between two points.
 * @param left pointer to one p3d_vertex
 * @param right pointer to the other p3d_vertex
 * @return the pointer to the resultant p3d_vector
 */
static p3d_vector* subReturn(p3d_vertex* left, p3d_vertex* right)
{
 	return create3Vector(left->v[CUT_A]-right->v[CUT_A],left->v[CUT_B]-right->v[CUT_B],left->v[CUT_C]-right->v[CUT_C]);
}

/**
 * Calculate the p3d_vertex that derives from the sum between a p3d_vertex and a p3d_vector.
 * @param origin pointer to the p3d_vertex
 * @param v pointer to the p3d_vector
 * @param out the pointer to the resultant p3d_vertex
 */
static void add(p3d_vertex* origin, p3d_vector* v, p3d_vertex* out)
{
 	out->v[CUT_A]=origin->v[CUT_A]+v->v[X];
 	out->v[CUT_B]=origin->v[CUT_B]+v->v[Y];
 	out->v[CUT_C]=origin->v[CUT_C]+v->v[Z];
 	out->v[CUT_D]=origin->v[CUT_D];
}

/**
 * Calculate the p3d_vertex that derives from the sum between a p3d_vertex and a p3d_vector.
 * @param origin pointer to one p3d_vertex
 * @param v pointer to the p3d_vector
 * @retun the pointer to the resultant p3d_vertex
 */
static p3d_vertex* addReturn(p3d_vertex* origin, p3d_vector* v)
{
 	return create4Vertex(origin->v[CUT_A]+v->v[X],origin->v[CUT_B]+v->v[Y],origin->v[CUT_C]+v->v[Z],origin->v[CUT_D]);
}

/**
 * Calculate the dot product between two vectors.
 * @param left pointer to a p3d_vector
 * @param right pointer to a p3d_vector
 * @return the dot product between the two vectors
 */
static double dot(p3d_vector* left, p3d_vector* right)
{
 	return ((left->v[X])*(right->v[X]))+((left->v[Y])*(right->v[Y]))+((left->v[Z])*(right->v[Z]));
}

/**
 * Calculate the cross product between two vectors.
 * @param left pointer to the left p3d_vector
 * @param right pointer to the right p3d_vector
 * @param out pointer to the resultant p3d_vector
 */
static void cross(p3d_vector* left, p3d_vector* right, p3d_vector* out)
{
    p3d_vector* temp = createEmptyVector();
 	temp->v[X]=(left->v[Y])*(right->v[Z])-(left->v[Z])*(right->v[Y]);
 	temp->v[Y]=(left->v[Z])*(right->v[X])-(left->v[X])*(right->v[Z]);
 	temp->v[Z]=(left->v[X])*(right->v[Y])-(left->v[Y])*(right->v[X]);
 	vectorCopy(temp,out);
 	freeVector(temp);
}

/**
 * Calculate the cross product between two vectors.
 * @param left a pointer to a p3d_vector
 * @param right a pointer to the other p3d_vector
 * @return the pointer to the p3d_vector resultant
 */
static p3d_vector* crossReturn(p3d_vector* left, p3d_vector* right)
{
 	return create3Vector((left->v[Y])*(right->v[Z])-(left->v[Z])*(right->v[Y]),(left->v[Z])*(right->v[X])-(left->v[X])*(right->v[Z]),(left->v[X])*(right->v[Y])-(left->v[Y])*(right->v[X]));
}

/**
 * Calculate the rows*columns product between two matrices.
 * @param left a pointer to the first p3d_matrix
 * @param right a pointer to the second p3d_matrix
 * @return the pointer to the resultant p3d_matrix
 */
static p3d_matrix* matMatProductReturn(p3d_matrix* left, p3d_matrix* right)
{
 	p3d_matrix* mat= createEmptyMatrix();
 	int i, j, k;
 	for(i=0;i<4;i++)
 	{
	 	for(j=0;j<4;j++)
	 	{
		 	mat->m[i][j]=0;
		 	for(k=0;k<4;k++)
		 	{
			 	mat->m[i][j]+=(left->m[i][k])*(right->m[k][j]);
			}
		}
	}
	return mat;
}

/**
 * Calculate the rows*columns product between two matrices.
 * @param left a pointer to the first p3d_matrix
 * @param right a pointer to the second p3d_matrix
 * @param out the pointer to the resultant p3d_matrix
 */
static void matMatProduct(p3d_matrix* left, p3d_matrix* right, p3d_matrix* out)
{
 	p3d_matrix* m = matMatProductReturn(left,right);
 	p3d_matrixCopy(m,out);
 	freeMatrix(m);
}


/**
 * Calculate the rows*columns product between a p3d_matrix and a p3d_vector.
 * @param mat a pointer to the p3d_matrix
 * @param v a pointer to the p3d_vector
 * @return the pointer to the resultant p3d_vector
 */
static p3d_vector* matVectProductReturn(p3d_matrix* mat, p3d_vector* v)
{
 	p3d_vector* out= createEmptyVector();
 	int i, k;
 	for(i=0;i<4;i++)
	{
	 	out->v[i]=0;
	 	for(k=0;k<4;k++)
	 	{
		 	out->v[i]+=(mat->m[i][k])*(v->v[k]);
		}
	}
 	return out;
}

/**
 * Calculate the rows*columns product between a p3d_matrix and a p3d_vector.
 * @param mat pointer to the p3d_matrix
 * @param v pointer to the p3d_vector
 * @param out the pointer to the resultant p3d_vector
 */
static void matVectProduct(p3d_matrix* mat, p3d_vector* v, p3d_vector* out)
{
 	p3d_vector* temp = matVectProductReturn(mat,v);
 	vectorCopy(temp,out);
 	freeVector(temp);
}

/**
 * Calculate the rows*columns product between a p3d_matrix and a p3d_vertex.
 * @param mat pointer to the p3d_matrix
 * @param v pointer to the p3d_vertex
 * @param out the pointer to the resultant p3d_vertex
 */
static p3d_vertex* matVertProductReturn(p3d_matrix* mat,p3d_vertex* v)
{
    p3d_vertex* out= createEmptyVertex();
 	int i, k;
 	for(i=0;i<4;i++)
	{
	 	out->v[i]=0;
	 	for(k=0;k<4;k++)
	 	{
		 	out->v[i]+=(mat->m[i][k])*(v->v[k]);
		}
	}
	normalizeVertex(out);
 	return out;
}

/**
 * Calculate the rows*columns product between a p3d_matrix and a p3d_vertex.
 * @param mat pointer to the p3d_matrix
 * @param v pointer to the p3d_vertex
 * @param out the pointer to the resultant p3d_vertex
 */
static void matVertProduct(p3d_matrix* mat,p3d_vertex* v, p3d_vertex* out)
{
    p3d_vertex* temp = matVertProductReturn(mat,v);
 	vertexCopy(temp,out);
 	freeVertex(temp);
}

/**
Rotation on a generic axis p3d_matrix constructor.
Create a p3d_matrix which perform a rotation around a generic axis
@param axis rotation axis
@param angle rotation angle
@return a pointer to the new rotation p3d_matrix
*/


static p3d_matrix* createRotationMatrix(p3d_vector* axis,double angle)
{
     p3d_matrix* rotX;
     p3d_matrix* rotY;
     p3d_matrix* rotZ;
     p3d_matrix* out;
     double beta = atanf((float)(axis->v[Y] / axis->v[Z]));
     rotX = createXRotationMatrix(beta);
     p3d_vector* nfirst = matVectProductReturn(rotX,axis);
     double gamma = (double)-atanf((float)(nfirst->v[X]/nfirst->v[Z]));

     rotY = createYRotationMatrix(gamma);
     out = matMatProductReturn(rotY,rotX);

     rotZ = createZRotationMatrix(angle);
     matMatProduct(rotZ,out,out);

     freeMatrix(rotY);
     freeMatrix(rotX);
     rotY = createYRotationMatrix(-gamma);
     rotX = createXRotationMatrix(-beta);
     matMatProduct(rotY,out,out);
     matMatProduct(rotX,out,out);

     freeMatrix(rotY);
     freeMatrix(rotX);
     freeMatrix(rotZ);
     freeVector(nfirst);

     return out;
}

static int isZeroVect(p3d_vector* vet)
{
 	if(vet->v[0]==0 && vet->v[1]==0 && vet->v[2]==0)
	{
	 	return TRUE;
	}
	else
	{
	 	return FALSE;
	}

}



static p3d_matrix* viewOrientation(p3d_viewModel* view)
{
	p3d_matrix* transMatrix= createTranslationMatrixf(-1.0*((view->vrp)->v[0]),-1.0*((view->vrp)->v[1]),-1.0*((view->vrp)->v[2]));
	p3d_vector* rn=normalizeVectorClone(view->vpn);
	p3d_vector* vet= crossReturn(view->vuv,rn);
	p3d_vector* ru;
	if(isZeroVect(vet))
	{
		ru=create3Vector(0.0, -1.0, 0.0);
	}
	else
	{
		ru=normalizeVectorClone(vet);
	}
	p3d_vector* rv= crossReturn(rn,ru);
	p3d_matrix* rotationMatrix=createIdentityMatrix();

	rotationMatrix->m[0][0]=ru->v[0];
	rotationMatrix->m[0][1]=ru->v[1];
	rotationMatrix->m[0][2]=ru->v[2];

	rotationMatrix->m[1][0]=rv->v[0];
	rotationMatrix->m[1][1]=rv->v[1];
	rotationMatrix->m[1][2]=rv->v[2];

	rotationMatrix->m[2][0]=rn->v[0];
	rotationMatrix->m[2][1]=rn->v[1];
	rotationMatrix->m[2][2]=rn->v[2];

	return matMatProductReturn(rotationMatrix,transMatrix);
}

static p3d_matrix* parallelViewMapping(p3d_viewModel* view)
{
	// dop = CW - prp

	p3d_vertex* cw= create4Vertex((view->mybox.umax+view->mybox.umin)/2,(view->mybox.vmax+view->mybox.vmin)/2,0.0,1.0);
	p3d_vector* dop= subReturn(cw,view->prp);
	view->dop=dop;
	double shx=-1.0*(view->dop->v[0]/view->dop->v[2]);
	double shy=-1.0*(view->dop->v[1]/view->dop->v[2]);
	p3d_matrix* shearMat=createShearingMatrix(2,shx,shy);
	p3d_matrix* traslMat=createTranslationMatrixf(-1.0*cw->v[0],-1.0*cw->v[1],-1.0*view->front);
	p3d_matrix* scalMat=createScalingMatrix(2.0/(view->mybox.umax-view->mybox.umin),2.0/(view->mybox.vmax-view->mybox.vmin),1.0/(view->front-view->back));
	p3d_matrix* result = matMatProductReturn(traslMat,shearMat);
	return matMatProductReturn(scalMat, result);
}

static p3d_matrix* perspectiveViewMapping(p3d_viewModel* view)
{
	p3d_matrix* traslMat=createTranslationMatrixf(-1*(view->prp->v[0]),-1*(view->prp->v[1]),-1*(view->prp->v[2]));

	p3d_vertex* cw= create4Vertex((view->mybox.umax+view->mybox.umin)/2,(view->mybox.vmax+view->mybox.vmin)/2,0.0,1.0);
	p3d_vector* dop= subReturn(cw,view->prp);
	view->dop=dop;
	double shx=-1*(view->dop->v[0]/view->dop->v[2]);
	double shy=-1*(view->dop->v[1]/view->dop->v[2]);
	p3d_matrix* shearMat=createShearingMatrix(2,shx,shy);
	p3d_matrix* result=matMatProductReturn(shearMat,traslMat);

	vrp1z = (matVertProductReturn(result, create4Vertex(0.0,0.0,0.0,1.0)))->v[2];

	p3d_matrix* scalMat = createScalingMatrix((2.0*vrp1z)/((view->mybox.umax-view->mybox.umin)*(vrp1z+view->back)),(2.0*vrp1z)/((view->mybox.vmax-view->mybox.vmin)*(vrp1z+view->back)),(-1/(vrp1z+view->back)));
	return matMatProductReturn(scalMat,result);
}

static p3d_matrix* perspTransf(p3d_viewModel* view, double vrp_z)
{
	zmin=-1.0*(vrp_z+view->front)/(vrp_z+view->back);
	p3d_matrix* result=createIdentityMatrix();
	result->m[3][2]=-1.0;
	result->m[3][3]=0.0;
	result->m[2][2]=1.0/(1.0+zmin);
	result->m[2][3]=-1.0*zmin/(1.0+zmin);
	return result;
}

static p3d_matrix* deviceMapping(p3d_viewPort* viewport, p3d_viewModel* view)
{

	//p3d_matrix* firstTrasl=createTranslationMatrixf(-1.0*(view->mybox.umin), -1.0*(view->mybox.vmin), 1.0);
	p3d_matrix* firstTrasl=createTranslationMatrixf(1.0, 1.0, 1.0);
	p3d_matrix* secondTrasl=createTranslationMatrixf(viewport->xmin,viewport->ymin,viewport->zmin);

	//double a= (viewport->xmax-viewport->xmin) / (view->mybox.umax - view->mybox.umin);
	double a= (viewport->xmax-viewport->xmin) / 2.0;
	//double b = (viewport->ymax-viewport->ymin) / (view->mybox.vmax - view->mybox.vmin);
	double b = (viewport->ymax-viewport->ymin) / 2.0;
	double c = (viewport->zmax-viewport->zmin);

	p3d_matrix* scaling= createScalingMatrix(a,b,c);

	p3d_matrix* result=matMatProductReturn(scaling,firstTrasl);
	return matMatProductReturn(secondTrasl,result);
}

