#include "generaldef.h"

#include <time.h>

#include "svg3d.h"
#include "scanline.h"
#include "bsp.h"



#include <list>
#include <deque>

class Svg3dCache
{
	//////////////////////
	// typedefs
	//////////////////////

	typedef struct{float x,y,z;}  Point;

	typedef struct{float R,G,B;} Color;

	//actual path
	std::list<Point> path; 

	//svg screen viewport 
	int width,height; 

	//actual normal
	float nx,ny,nz; 

	//reverse normal
	bool bReverse; 

	//centroid
	float cx,cy,cz; 

	//viewpoint
	float vx,vy,vz;
	
	//global transformation
	float* Tv;

	//actual color
	float  R,G,B;

	//front to back or back to front
	bool  bBackToFront;

	//polygons total,exported, back face culling, out of screen, not visibile (front-to-back)
	int   npolygon,nbackculling,noutofscene,nnotvisible;

	//list of path to export
	std::deque< std::pair<std::list<Point> , Color> > list_to_export;

	//function
	void transform(Point& p);

	void dump(FILE* file,std::list<Point> path,float R,float G,float B);

public:

	//constructor
	Svg3dCache(float vx,float vy,float vz,  // view point for bsp traversal
			   float* Tv,                   // global transformation for projection
			   int width,int height,        //final svg resolution in pixel
			   bool bBackToFront           //how to export (front to back or back to front)
	);

	//export to file
	void dump(char* filename="temp.svg");

	/* to add projected face */
	void normal(bool bReverse,float nx,float ny,float nz);

	void add(float x,float y,float z);

	void flush();

};


//------------------------------------------------------------------------
Svg3dCache::Svg3dCache(
		float vx,float vy,float vz,
		float* Tv,
		int width,int height,
		bool bBackToFront
	)
		:npolygon(0),nbackculling(0),noutofscene(0),nnotvisible(0)
{
	this->Tv=Tv;
	this->vx=vx;
	this->vy=vy;
	this->vz=vz;
	this->width=width;
	this->height=height;
	this->bBackToFront=bBackToFront;

	this->cx=0;
	this->cy=0;
	this->cz=0;

	if (!bBackToFront)
		scan_init(width,height,1,1,1);
}


//------------------------------------------------------------------------
void Svg3dCache::add(float x,float y,float z)
{
	Point p;
	p.x=x;cx+=x;
	p.y=y;cy+=y;
	p.z=z;cz+=z;
	path.push_back(p);
}


//------------------------------------------------------------------------
void Svg3dCache::normal(bool bReverse,float nx,float ny,float nz)
{
	this->nx=nx;
	this->ny=ny;
	this->nz=nz;
	this->bReverse=bReverse;
}


//------------------------------------------------------------------------
inline void Svg3dCache::transform(Point& p)
{
	real H=Tv[ 0]*1+Tv[ 1]*p.x+Tv[ 2]*p.y+Tv[ 3]*p.z;
	real X=Tv[ 4]*1+Tv[ 5]*p.x+Tv[ 6]*p.y+Tv[ 7]*p.z;
	real Y=Tv[ 8]*1+Tv[ 9]*p.x+Tv[10]*p.y+Tv[11]*p.z;
	real Z=Tv[12]*1+Tv[13]*p.x+Tv[14]*p.y+Tv[15]*p.z;

	p.x=X/H;
	p.y=Y/H;
	p.z=Z/H;
}

//------------------------------------------------------------------------
void Svg3dCache::dump(char* filename)
{
	clock_t t1=clock();

	FILE* file;
	
	if (!strcmp(filename,"stdout"))
		file=stdout;
	else
		file=fopen(filename,"wt");

	fprintf(file,"<svg width='%d' height='%d' viewBox='0 0 %d %d'>\n",width,height,width,height);
	fprintf(file,"<g stroke='black' stroke-width='1'>\n"); 

	std::deque< std::pair<std::list<Point> , Color> >::iterator it;
	for (it=list_to_export.begin();it!=list_to_export.end();it++)
	{
		std::pair< std::list<Point> , Color > item=*it;
		std::list<Point> path=item.first;
		Color clr=item.second;
		dump(file,path,clr.R,clr.G,clr.B);
	}

	fprintf(file,"</g>\n");
	fprintf(file,"</svg>\n");
	
	if (!strcmp(filename,"stdout"))
		;
	else
		fclose(file);

	printf("Done SVG Export in %.2f seconds\n",(clock()-t1)/(float)CLOCKS_PER_SEC);
	printf("  output file       %s\n",filename);
	printf("  export_method     %s\n",bBackToFront?"back-to-front":"front-to-back");
	printf("  viewpoint         %f %f %f\n",vx,vy,vz);
	printf("  dimension         W=%d H=%d\n",width,height);
	printf("  total_polygon     %d\n",(int)npolygon);
	printf("  exported_faces    %d (%d perc)\n",(int)list_to_export.size(),(int)(100.0f*list_to_export.size()/(float)npolygon));
	printf("  back_face_culled  %d (%d perc)\n",(int)nbackculling         ,(int)(100.0f*nbackculling         /(float)npolygon));
	printf("  out_of_scene      %d (%d perc)\n",(int)noutofscene          ,(int)(100.0f*noutofscene          /(float)npolygon));
	printf("  not_visible       %d (%d perc)\n",(int)nnotvisible          ,(int)(100.0f*nnotvisible          /(float)npolygon));
	printf("\n");
}


//------------------------------------------------------------------------
void Svg3dCache::dump(FILE* file,std::list<Point> path,float R,float G,float B)
{
	fprintf(file,"<path fill='rgb(%d,%d,%d)' d='m ",(int)(R*255),(int)(G*255),(int)(B*255));
	float curx=0,cury=0;
	for (std::list<Point>::iterator it=path.begin();it!=path.end();it++)
	{
		Point p=*it;
		transform(p);
		p.y=(-1*p.y+height);
		fprintf(file,"%f %f ",p.x-curx,p.y-cury);
		curx=p.x;cury=p.y;
	}

	Point p=*path.begin();
	transform(p);
	p.y=(-1*p.y+height);
	fprintf(file,"%f %f ",p.x-curx,p.y-cury);
	fprintf(file,"z' />\n");
}


//------------------------------------------------------------------------
void Svg3dCache::flush()
{
	++npolygon;
	cx/=path.size();
	cy/=path.size();
	cz/=path.size();

	float dx=vx-cx; float nx=(bReverse?-1:+1)*this->nx;
	float dy=vy-cy; float ny=(bReverse?-1:+1)*this->ny;
	float dz=vz-cz; float nz=(bReverse?-1:+1)*this->nz;
	float mdir=sqrt(dx*dx+dy*dy+dz*dz);
	dx/=mdir;dy/=mdir;dz/=mdir;
	float cangle=dx*nx+dy*ny+dz*nz;

	//cull back faces
#if 1
	if (cangle<0)
	{
		++nbackculling;
		path.clear();
		return;
	}
#endif

	float R=0.5f,G=0.5f,B=0.5f;
	float hue,saturation,lightness;
	RGB2HSL(hue,saturation,lightness,R,G,B);
	lightness=fabs(cangle);
	HSL2RGB(R,G,B,hue,saturation,lightness);

	//out of scene 
#if 1
	{
		bool bInScene=false;
		for (std::list<Point>::iterator it=path.begin();it!=path.end();it++)
		{
			Point p=*it;
			transform(p);

			if (p.x>=0 && p.x<width && p.y>=0  && p.y<height)
				{bInScene=true;break;}
		}

		if (!bInScene)
		{
			noutofscene++;
			path.clear();
			return;
		}
	}
#endif

	if (!bBackToFront)
	{
		float* v=new float[path.size()*3];
		float *t=v;

		for (std::list<Point>::iterator it=path.begin();it!=path.end();it++)
		{
			Point p=*it;
			transform(p);
			*t++=p.x;*t++=p.y;*t++=p.z; //probably a bug in z (I need to reverse Z coordinates to draw on screen)
		}

		if (!scan_add_polygon(path.size(),v,R,G,B))
		{
			delete [] v;
			++nnotvisible;
			path.clear();
			return;
		}

		delete [] v;
	}

	Color clr={R,G,B};
	std::pair< std::list<Point> ,Color > item(path,clr);

	if (!bBackToFront)
		list_to_export.push_front( item) ;
	else
		list_to_export.push_back( item) ;

	path.clear();
}



////////////////////////////////////////////////////////////////////////////////////////


static void doExportSvg(Svg3dCache* sc,bsp_solver* solver,key N,int width,int height,bool back_to_front,real* observer,real* vmat)
{
	int i;
	checkpoint(solver->getDim()==3);

	ggraph::list_iterator it;
	ggraph* g=solver->getGraph();
	int pointdim=g->pointdim;


	if (!N)
		N=solver->getRoot();

	if (solver->isIn(N))
	{
		//fprintf(file,"<!-- NODE_IN -->\n");
		return;
	}
	
	if (solver->isOut(N))
	{
		//fprintf(file,"<!-- NODE_OUT -->\n");
		return;
	}

	checkpoint(solver->isCut(N));
	
	real* h=solver->getPlane(N);

	//fprintf(file,"<!-- NODE_CUT plane{");
	//for (i=0;i<=getDim();++i) fprintf(file,"%e ",(float)h[i]);
	//fprintf(file,"--> \n");

	real acc=0;
	for (i=0;i<=pointdim;i++) acc+=h[i]*observer[i];

	key Below=solver->getBelow(N);
	key Above=solver->getAbove(N);

	if ((back_to_front && acc<0) || (!back_to_front && acc>=0)) 
		{Swap(key,Below,Above)};

	//**** BELOW ****
	doExportSvg(sc,solver,Below,width,height,back_to_front,observer,vmat);

	/*** EXPORT POLYGONS ***/
	for (it=g->each(g->pointdim-1);!it.end();it++)
	{
		key Face=*it;

		//only for boundary face,only boundary face which correspond to current node
		if (g->getNUp(Face)!=1 || g->getNode(Face)->Id!=N) 
			continue;

		//fix vertex order
#if 1
		real* bball=g->boundingBall(g->getFirstUpNode(Face)); //i know there is only one up-cell
		bool bReverse=false;

		if (bball[0]>0)
		{
			real acc=h[0];
			for (int J=1;J<=g->pointdim;J++)  acc+=h[J]*bball[J];
			bReverse=(acc>=0);
		}
#endif

		sc->normal(bReverse,h[1],h[2],h[3]); //setup normal
		
		key prev0=g->getFirstDwNode(g->getLastDwNode(Face)),p1;
		key prev1=g->getLastDwNode (g->getLastDwNode(Face)),p2;

		for (graph::iterator i=g->goDw(Face);!i.end();i++,prev0=p1,prev1=p2)
		{
			p1=g->getFirstDwNode(*i);
			p2=g->getLastDwNode (*i);
			if (p1!=prev0 && p1!=prev1) Swap(key,p1,p2);
			real* coords=g->getGeom(p1);
			sc->add(coords[1],coords[2],coords[3]); //add point
		}

		sc->flush(); //flush polygon
	}

	//*** ABOVE ****
	doExportSvg(sc,solver,Above,width,height,back_to_front,observer,vmat);
}


void svg3d(bsp_solver* solver,const char* filename,int width,int height,bool back_to_front,real* observer,real* vmat)
{
	if (solver->getDim()!=3)
		return;

	Svg3dCache sc(observer[1],observer[2],observer[3],vmat,width,height,back_to_front);
	doExportSvg(&sc,solver,solver->getRoot(),width,height,back_to_front,observer,vmat);
	sc.dump((char*)filename);
}