/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */




#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_eigen.h>
#include <math.h>

#include "tensor.h"
#include "mat.h"

///////////////////////////////////////////////////////////////////////////////////////

InertiaTensor::InertiaTensor(int dim)
{
	this->dim=dim;
}

///////////////////////////////////////////////////////////////////////////////////////

static void inertia_bball(int dim,real coeff,real* A,real radius,real* C)
{
	checkpoint(dim==2 || dim==3);

	real PI=(real)3.1415926535;
	real radius2=radius * radius;
	
	if (dim==2) //3x3 matrix
	{
		coeff*=(PI) * radius2;

		A[ 0]+=coeff*(radius2/4+C[0]*C[0]); A[ 1]+=coeff*(          C[0]*C[1]); A[ 2]+=coeff*C[0];
		A[ 3]+=coeff*(          C[1]*C[0]); A[ 4]+=coeff*(radius2/4+C[1]*C[1]); A[ 5]+=coeff*C[1];
		A[ 6]+=coeff*C[0]                 ; A[ 7]+=coeff*C[1]                 ; A[ 8]+=coeff;
	}
	else //4x4 matrix
	{
		coeff*=(4*PI/3) * radius * radius * radius;

		A[ 0]+=coeff*(radius2/5+C[0]*C[0]); A[ 1]+=coeff*(          C[0]*C[1]); A[ 2]+=coeff*(          C[0]*C[2]); A[ 3]+=coeff*C[0];
		A[ 4]+=coeff*(          C[1]*C[0]); A[ 5]+=coeff*(radius2/5+C[1]*C[1]); A[ 6]+=coeff*(          C[1]*C[2]); A[ 7]+=coeff*C[1];
		A[ 8]+=coeff*(          C[2]*C[0]); A[ 9]+=coeff*(          C[2]*C[1]); A[10]+=coeff*(radius2/5+C[2]*C[2]); A[11]+=coeff*C[2];
		A[12]+=coeff*C[0]                 ; A[13]+=coeff*C[1]                 ; A[14]+=coeff*C[2];                  A[15]+=coeff;
	}
}

///////////////////////////////////////////////////////////////////////////////////////

static void findEigenVectors(int dim,real* Eigvec,real* Eigval,real* M)
{
	gsl_matrix*                eigvec  = gsl_matrix_calloc(dim,dim);
	gsl_eigen_symmv_workspace* w       = gsl_eigen_symmv_alloc(dim);
	gsl_vector*                eigval  = gsl_vector_calloc(dim);
	gsl_matrix*                mat     = gsl_matrix_calloc(dim,dim);
	
	//load matrix (note! should be symmetric)
	int r,c;
	for (r=0;r<dim;r++)
	{
		for (c=0;c<dim;c++)
		{
			gsl_matrix_set(mat,r,c,M[r*dim+c]);
			checkpoint(fabs(M[r*dim+c]-M[c*dim+r])<=1e-2); //symmetric!
		}
	}
	
	gsl_eigen_symmv(mat,eigval,eigvec,w);
	gsl_eigen_symmv_sort(eigval,eigvec,GSL_EIGEN_SORT_ABS_DESC); //order by eigenvalue
	
	for (int i=0;i<dim;i++)
		Eigval[i]=(real)gsl_vector_get(eigval,i);
	
	//normalize
	for (r=0;r<dim;r++)
	{
		real acc=0;
		for (c=0;c<dim;c++)
		{
			Eigvec[r*dim+c]=(real)gsl_matrix_get(eigvec,r,c);
			acc+=Eigvec[r*dim+c]*Eigvec[r*dim+c];
		}
		
		
		for (acc=(real)sqrt(acc),c=0;c<dim;c++) 
			Eigvec[r*dim+c]/=acc;
	}
	
	gsl_matrix_free(mat);
	gsl_matrix_free(eigvec);
	gsl_vector_free(eigval);
	gsl_eigen_symmv_free(w);
	
	tmat<real>::transpose(dim,Eigvec,Eigvec); //the first row is the first eigenvec
}

///////////////////////////////////////////////////////////////////////////////////////

void InertiaTensor::solve()
	{
	  int i;
		for (i=0;i<dim;i++)
			Bar[i]=A[i*(dim+1)+dim]/A[(dim+1)*(dim+1)-1];
		
		//translate of the baricenter and find the eigenvectors (which will be the principal axes)
		tmat<real>::identity(dim+1,G);
		
		for (i=0;i<dim;i++) 
			G[i*(dim+1)+dim]=-Bar[i];
		
		tmat<real>::transpose(dim+1,GT,G);
		tmat<real>::product(dim+1,A,A,GT);
		tmat<real>::product(dim+1,A,G,A);
		tmat<real>::submatrix(A,dim+1,A,0,dim-1,0,dim-1);
		
		findEigenVectors(dim,Eigvec,Eigval,A);
	}


///////////////////////////////////////////////////////////////////////////////////////

void InertiaTensor::addBall(real R,real* C)
{
	real coeff=1;
	inertia_bball(dim,coeff,A,R,C);
}
