/* -------------------------------------------------------------------------- *\
   Project..........:   eXtreme Geometric Environment for PLaSM
   Copyright........:   Universita' degli studi di ROMA TRE
                        Dipartimento di Informatica e Automazione 
                        Via Vasca Navale 84, 00146 ROMA
	Development by...:	Giorgio Scorzelli
	E-Mail...........:	scorzell.uniroma3.it
	Environment......:	Win32, UNIX
\* -------------------------------------------------------------------------- */



#ifndef _VIEWER_H__
#define _VIEWER_H__

#include "generaldef.h"

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include <stdio.h>

#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/freeglut.h>
#endif

#include <string.h>

class Win;
class Panel;
class Draw;
class Options;


//___________________________________________________________________________

class Win
{
	friend class Panel;

protected:

	int       reshaping;
	int       loading;
	int       freeze;
	int       W,H;

#ifdef __APPLE__
volatile int       close_flag;
#endif

	//windows handle (platform dependent)
	int       winhandle;

	//number of views,actual views and pointer to views
	int       nview,aview;
	float     pview[256];
	Panel     *view[256];
	Options   *options;

public:

	//default constructor
	Win();

	//draw text in screen
	void text2d(GLint x, GLint y, char* s,GLfloat r=0, GLfloat g=0, GLfloat b=0);

	//draw text in world
	static void text3d(GLfloat x,GLfloat y,GLfloat z,char* s);

	//draw box
	static void box(GLfloat x1,GLfloat y1,GLfloat z1,GLfloat x2,
		GLfloat y2,GLfloat z2,GLfloat r=0,GLfloat g=0,GLfloat b=0);

	//draw background
	static void background(int x,int y,int winx,int winy,
		int r1,int g1,int b1,int r2,int g2,int b2,int r3,int g3,int b3,int r4,int g4,int b4);

	inline static bool isAltPressed()     
		{return (glutGetModifiers() & GLUT_ACTIVE_ALT)!=0; }

	inline static bool isShiftPressed()   
		{return (glutGetModifiers() & GLUT_ACTIVE_SHIFT)!=0;}

	inline static bool isCtrlPressed()    
		{return (glutGetModifiers() & GLUT_ACTIVE_CTRL)!=0;}

	//position the window
	void move(int x1,int y1,int x2,int y2);

	//perform a full redisplay
	void redisplay() {need_redisplay=1;}

	//run glut main loop
	void run();

	//static callback for GLUT
#ifdef __APPLE__
	virtual void close   ();
#endif

	virtual void motion   (GLint x, GLint y);
	virtual void special  (int key, int x, int y);
	virtual void reshape  (GLint winx,GLint winy);
	virtual void mouse    (GLint button, GLint state, GLint x, GLint y);
	virtual void display  ();
	virtual void keydown  (unsigned char key, GLint x, GLint y);

	virtual int save(int version,FILE* f);
	virtual int open(int version,FILE* f);

	int isFreeze() const
		{return freeze;}

	int need_redisplay;

	//view ports
	void add(Panel* view);

	//pack
	void _pack(float* pview,int& n,int& v,int X,int Y,int W,int H);

	void pack(float* pview) ;

	int repack(float* pview,int& n,int pickX,int pickY,int X,int Y,int W,int H);

	// Return window operating system handle
	inline int getWindowHandle() {return winhandle;}

	//return height of the window
	inline int getWindowHeight()	{return H;}

	//return width of the window
	inline int getWindowWidth()	{return W;}

	//access the selected viewport (the actual if no arguments are given)
	Panel* getViewport(int n=-1)	{return view[n==-1?aview:n];}




	static void main(int argn,char** args);

};


//___________________________________________________________________________

class Panel
{
	friend class Win;
	friend class Options;
	friend class PovRay;


	int ndraw;
	Draw* draw[64*1024];

	// actual operation
	bool scaling;
	int  moving;
	bool panning;

	//starting "moving" coordinates
	int beginx,beginy;

	typedef struct
	{
		double model [16]; /* OpenGL MODEL matrix        */
		float  ortho [6] ; /* Ortho view                 */
		double proj  [16]; /* OpenGL PROJECTION matrix   */
		GLint    view  [4] ; /* OpenGL VIEWPORT            */
		double eye   [4] ; /* OpenGL Eye position        */

	} 
	GLConf;

	typedef struct 
	{
		int           newmat;
		GLConf        glconf;
		GLfloat       mat[16];
		unsigned int  viewpoint ;
	} 
	SavedMat;

	int options;
	float bbox[6];
	float X,Y,W,H;

	char name[256];

	//the window that owns the view
	Win* win;

	//current transformation matrix
	float  curmat[4][4];

	//current coordinate system
	float  vx[3],vy[3],vz[3];

	//current scale factor
	float  scalefactor;

	//all viee information saved
	SavedMat savedmat;

	//set a new "point of view"
	void setMatrix(GLfloat t[4][4]);



	//reset ortho for viewing
	void resetOrtho ();

	//communicate that the ortho has changed
	void ortoChanged();

	//build the cartesian reference
	void mkvectors();


public:
	
	//default constructor
	Panel(const char* name,int options=0);

	//display function
	virtual void display();

	//save matrix for point coordinate invertion
	void grabMatrices(bool bForce=false);

	//help function
	virtual void help(FILE * out=stderr);

	//add a draw to the panel
	void add(Draw* draw);

	//set the bounding box
	void setbbox(float x1,float x2,float y1,float y2,float z1,float z2) ;

	//set the bounding box
	void setbbox(float* bbox);

	//access the window
	inline Win* getWindow() {return win;}

	inline float getX() {return X;}
	inline float getY() {return Y;}
	inline float getW() {return W;}
	inline float getH() {return H;}

	//reset "point of view" matrix
	void resetMatrix();

	//set one rotation (rot=0,1,2... 3 different kind of rotation)
	void setNewRotation(int rot);

	//mouse callback
	virtual void mouse(GLint button, GLint state, GLint x, GLint y);

	//motion callback
	virtual void motion(GLint x, GLint y);
	
	//keydown callback
	virtual void keydown(unsigned char key, GLint x, GLint y);

	//special callback
	virtual void special(int, int x, int y);

	//reshape callback
	virtual void reshape();

	// methods useful for navigation
	void scale(GLfloat d);

	void rotate(GLfloat beginx,GLfloat beginy,GLfloat x,GLfloat y);
	
	void ortho(GLfloat dx,GLfloat dy);

	//convert screen coordinate -> world coordinate
	void convertScreenPoint(const GLint& px,const GLint& py,
		GLdouble& x1,GLdouble& y1, GLdouble& z1,
		GLdouble& x2,GLdouble& y2, GLdouble& z2);

	virtual int save(int version,FILE* f);
	virtual int open(int version,FILE* f);

	//transform a point
	void transform(float& X,float& Y,float& Z);
	
		//getviewpoint
	void getViewpoint(float& x,float &y,float& z);
};

//___________________________________________________________________________

class Draw
{
	friend class Panel;

protected:

	Panel* panel;

	virtual void display()=0;

	virtual int save(int version,FILE* f) {return 0;};
	virtual int open(int version,FILE* f) {return 0;};

public:

	inline Panel* getPanel()
		{return this->panel;}

	//mouse callback
	virtual void mouse(GLint button, GLint state, GLint x, GLint y) {}

	//motion callback
	virtual void motion(GLint x, GLint y) {}

	//keydown callback
	void keydown(unsigned char key, GLint x, GLint y);
};


//___________________________________________________________________________

class Options : public Panel
{
protected:

	enum{BOOL,INT,FLOAT,STRING};

	struct
	{
		int   format;
		char  name[256];
		char  explanation[256];
		union {bool b;int i;unsigned int ui;float f;char s[256];} value;
	}
	options[256];
	
	int n,a;

	int find(char* name);

public:

	//constructor
	Options();

	//set actual
	void setActual(char* name);

	//next and prev
	inline void  next()	 {a=(a+1)%n;}
	inline void  prev()	 {a=a==0?n-1:a-1;}
	inline char* actual() {return this->options[a].name;}

	inline bool exist(char* name)	
		{return find(name)!=-1;}

	inline int   getInt   (char* name) {return options[find(name)].value.i;};
	inline bool  getBool  (char* name) {return options[find(name)].value.b;};
	inline float getFloat (char* name) {return options[find(name)].value.f;};
	inline char* getString(char* name) {return options[find(name)].value.s;};

	void set(char* name,bool value ,char* explanation=0);
	void set(char* name,int  value ,char* explanation=0);
	void set(char* name,float value,char* explanation=0);
	void set(char* name,char* value,char* explanation=0);

	virtual void display();
};

//___________________________________________________________________________

class DrawSphere:public Draw
{public: virtual void display()	{glColor3f(0,1,0);glutSolidSphere(0.7,8,8);}};


class DrawCube:public Draw
{public: virtual void display()	{glColor3ub(224,218,218); glutSolidTeapot(3);}};


void win_test();


#endif 


